package com.pengyifan.biocreativev;

import com.google.common.collect.Lists;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.commons.collections.Pairs;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.NamedEntity;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.commons.collections.dependencygraph.Position;
import com.pengyifan.nlp.process.ExtractLemma;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import org.javatuples.Pair;

public class DependencyGraphVertexMerger {

  private static final ExtractLemma LEMMA = new ExtractLemma();
  private final DependencyGraph graph;
  private final Collection<BratEntity> allProteinList;

  public DependencyGraphVertexMerger(DependencyGraph graph, Collection<BratEntity> allProteinList) {
    this.graph = graph;
    this.allProteinList = allProteinList;
  }

  public DependencyGraph merge() {
    // check if an entity covers others
    List<BratEntity> mergedProteins = Lists.newArrayList(allProteinList);
    List<BratEntity> addedProteins = Lists.newArrayList();
    Collection<Pair<BratEntity, BratEntity>> pairs = Pairs.getPairs(mergedProteins);
    for (Pair<BratEntity, BratEntity> pair : pairs) {
      if (pair.getValue0().totalSpan().encloses(pair.getValue1().totalSpan())) {
        mergedProteins.remove(pair.getValue0());
        addedProteins.add(pair.getValue0());
      }
      if (pair.getValue1().totalSpan().encloses(pair.getValue0().totalSpan())) {
        mergedProteins.remove(pair.getValue1());
        addedProteins.add(pair.getValue1());
      }
    }

    for (BratEntity e : mergedProteins) {
      merge(e);
    }
    for (BratEntity e : addedProteins) {
      add(e);
    }
    return graph;
  }

  /**
   * Add new entity
   */
  private void add(BratEntity entity) {
    List<DependencyGraphVertex> nodeList = graph.getSortedVertexList();
    LinkedList<DependencyGraphVertex> subNodeList = Lists.newLinkedList();
    for (DependencyGraphVertex vertex : nodeList) {
      if (contains(entity, vertex)) {
        subNodeList.add(vertex);
      }
    }

    if (subNodeList.size() <= 1) {
      return;
    }
    if (intersect(entity, subNodeList)) {
      return;
    }

    DependencyGraphVertex newVertex = createVertex(graph.getMaxIndex() + 1, entity,
        subNodeList.get(subNodeList.size() - 1).getTag());
    graph.addVertex(newVertex);

    for (DependencyGraphEdge edge : graph.edgeSet()) {
      DependencyGraphVertex source = graph.getEdgeSource(edge);
      DependencyGraphVertex target = graph.getEdgeTarget(edge);
      if (subNodeList.contains(target) && !subNodeList.contains(source)) {
        if (!graph.containsEdge(source, newVertex)) {
          DependencyGraphEdge newEdge = createEdge(graph.getMaxIndex() + 1, edge);
          graph.addEdge(source, newVertex, newEdge);
        }
      } else if (!subNodeList.contains(target)
          && subNodeList.contains(source)) {
        if (!graph.containsEdge(newVertex, target)) {
          DependencyGraphEdge newEdge = createEdge(graph.getMaxIndex() + 1, edge);
          graph.addEdge(newVertex, target, newEdge);
        }
      }
    }
  }

  /**
   * Merge multiple edges to the new vertex.
   */
  private void merge(BratEntity entity) {
    List<DependencyGraphVertex> nodeList = graph.getSortedVertexList();
    LinkedList<DependencyGraphVertex> subNodeList = Lists.newLinkedList();
    for (DependencyGraphVertex vertex : nodeList) {
      if (contains(entity, vertex)) {
        subNodeList.add(vertex);
      }
    }

    if (subNodeList.size() <= 1) {
      return;
    }
    if (intersect(entity, subNodeList)) {
      return;
    }

    DependencyGraphVertex newVertex = createVertex(graph.getMaxIndex() + 1, entity,
        subNodeList.get(subNodeList.size() - 1).getTag());
    graph.addVertex(newVertex);

    for (DependencyGraphEdge edge : graph.edgeSet()) {
      DependencyGraphVertex source = graph.getEdgeSource(edge);
      DependencyGraphVertex target = graph.getEdgeTarget(edge);
      if (subNodeList.contains(target) && !subNodeList.contains(source)) {
        graph.removeEdge(edge);
        if (!graph.containsEdge(source, newVertex)) {
          graph.addEdge(source, newVertex, edge);
        }
      } else if (!subNodeList.contains(target)
          && subNodeList.contains(source)) {
        graph.removeEdge(edge);
        if (!graph.containsEdge(newVertex, target)) {
          graph.addEdge(newVertex, target, edge);
        }
      }
    }
    // remove vertex
    for (DependencyGraphVertex vertex : subNodeList) {
      graph.removeVertex(vertex);
    }
  }

  private boolean intersect(BratEntity entity,
      LinkedList<DependencyGraphVertex> nodeList) {
    return entity.beginPosition()
        != nodeList.getFirst().getBeginPosition()
        || entity.endPosition() != nodeList.getLast().getEndPosition();
  }

  private boolean contains(BratEntity entity, DependencyGraphVertex vertex) {
    return entity.beginPosition() <= vertex.getBeginPosition()
        && vertex.getEndPosition() <= entity.endPosition();
  }

  private DependencyGraphEdge createEdge(int index, DependencyGraphEdge edge) {
    return new DependencyGraphEdge(index, edge);
  }

  private DependencyGraphVertex createVertex(int index, BratEntity entity, PartOfSpeech tag) {
    DependencyGraphVertex vertex = new DependencyGraphVertex(
        index,
        entity.getText(),
        tag,
        LEMMA.lemmatize(entity.getText(), "NN"),
        NamedEntity.PROTEIN_X,
        Position.UNKNOWN,
        entity.beginPosition(),
        entity.endPosition());
    return vertex;
  }
}
