package com.pengyifan.biocreativev;

import com.google.common.collect.Maps;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.util.BratUtils;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.kernel.allpathgraph.singleprocess.DependencyGraphVertexSpliter;
import com.pengyifan.kernel.allpathgraph.singleprocess.LemmaSetter;
import com.pengyifan.nlp.ssdg.SSDGBuilder;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import static com.google.common.base.Preconditions.checkNotNull;

public class EDGCreater {

  private static final Logger LOGGER = Logger.getAnonymousLogger();

  private String text;
  private DependencyGraph dependencyGraph;
  private DependencyGraph ssdg;
  private BratDocument corefDoc;
  private BratDocument abbrDoc;

  private Collection<BratEntity> allProteinList;
  private Set<DependencyGraphVertex> proteinVertexSet;
  private Map<BratEntity, List<DependencyGraphVertex>> allProteinVertexList;

  public EDGCreater() {
  }

  public EDGCreater setA1Doc(BratDocument a1Doc) {
    allProteinList = BratUtils.filtEntities(a1Doc.getEntities(), "Protein");
    return this;
  }

  public EDGCreater setText(String text) {
    this.text = text;
    return this;
  }

  public EDGCreater setCorefDoc(BratDocument corefDoc) {
    this.corefDoc = corefDoc;
    return this;
  }

  public EDGCreater setAbbrDoc(BratDocument abbrDoc) {
    this.abbrDoc = abbrDoc;
    return this;
  }

  public DependencyGraph create() {
    splitDependencyGraphVertex();
    mergeDependencyGraphVertex();
    findAllDependencyGraphVertexList();
    setLemma();
    setProtein();
    setLemma();
    buildSSDG();
    return ssdg;
  }

  private void buildSSDG() {
    LOGGER.finer("Start getting dependency all-path graph");
    checkNotNull(text, "Text is not set");
    checkNotNull(dependencyGraph, "Dependency graph is not set");

    SSDGBuilder b = new SSDGBuilder(text, dependencyGraph, corefDoc, abbrDoc);
    b.build();

    ssdg = b.getSSDG();
  }

  private void setLemma() {
    LOGGER.finer("Start setting lemma");
    checkNotNull(dependencyGraph, "Dependency graph is not set");
    new LemmaSetter(dependencyGraph).set();
  }

  /**
   * Need to change for BioNLP task.
   */
  private void setProtein() {
    LOGGER.finer("Start setting protein");

    checkNotNull(dependencyGraph, "Dependency graph is not set");
    checkNotNull(proteinVertexSet, "Entity is not set");

    new ProteinSetter(dependencyGraph, proteinVertexSet).set();
  }

  private void findAllDependencyGraphVertexList() {
    LOGGER.finer("Start finding dependency graph node");

    checkNotNull(dependencyGraph, "DependencyGraph is null");

    allProteinVertexList = Maps.newHashMap();

    for (BratEntity entity : allProteinList) {
      List<DependencyGraphVertex> vertexList = new DependencyGraphVertexFinder(
          dependencyGraph, entity.totalSpan()).find();
      if (vertexList.size() == 1) {
        allProteinVertexList.put(entity, vertexList);
      }
    }

    proteinVertexSet = allProteinVertexList
        .values()
        .stream()
        .flatMap(list -> list.stream())
        .collect(Collectors.toSet());
  }

  /**
   * "word / word" or "word - word"
   */
  private void splitDependencyGraphVertex() {
    LOGGER.finer("Start splitting dependency graph vertex");
    checkNotNull(dependencyGraph, "Dependency graph is not set");
    checkNotNull(allProteinList, "Proteins are not set");
    dependencyGraph = new DependencyGraphVertexSpliter(dependencyGraph, allProteinList).split();
  }

  public void mergeDependencyGraphVertex() {
    LOGGER.finer("Start merging dependency graph vertex");
    checkNotNull(dependencyGraph, "Dependency graph is not set");
    checkNotNull(allProteinList, "Protein list has not been set");
    new DependencyGraphVertexMerger(dependencyGraph, allProteinList).merge();
  }

  public EDGCreater setDependencyGraph(DependencyGraph dependencyGraph) {
    this.dependencyGraph = dependencyGraph;
    return this;
  }
}
