package com.pengyifan.biocreativev;

import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.commons.collections.dependencygraph.*;
import com.pengyifan.nlp.ssdg.PartWholeLinkAdder;
import edu.emory.mathcs.backport.java.util.Collections;

public class PPIExtractorAnnotation extends PPIExtractorBasic {

  public PPIExtractorAnnotation(DependencyGraph graph, BratDocument bratdoc) {
    super(graph, bratdoc, Collections.emptyList());
  }

  public void extract() {
    /**
     * form complex
     */
    final DependencyGraphPattern formPattern = DependencyGraphPattern
        .of()
            // form
        .addVertex(0, v -> v.getLemma().equals("form"))
            // complex
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        .addEdge(0, 1, Dependency.ofPredicate(DependencyName.ARG0, DependencyName.ARG1));
    extractAnnotation(formPattern);
    /**
     * isolate complex
     */
    final DependencyGraphPattern isolatePattern = DependencyGraphPattern
        .of()
            // isolate
        .addVertex(0, v -> v.getLemma().equals("isolate")
            || v.getLemma().equals("identify")
            || v.getLemma().equals("obtain"))
            // complex
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        .addEdge(0, 1, Dependency.ofPredicate(DependencyName.ARG0, DependencyName.ARG1));
    extractAnnotation(isolatePattern);

    /**
     * interact with each other
     */
    final DependencyGraphPattern eachOtherPattern = DependencyGraphPattern
        .of()
            // trigger
        .addVertex(1, v -> v.getLemma().equals("interact")
            || v.getLemma().equals("bind"))
            // each
        .addVertex(3, v -> v.getWord().equals("each"))
            // other
        .addVertex(4, v -> v.getWord().equals("other"))
        .addEdge(1, 4, d -> true)
        .addEdge(4, 3, Dependency.ofPredicate(DependencyName.DET));
    extractAnnotation(eachOtherPattern);

    /**
     * identification of interaction
     */
    final DependencyGraphPattern identificationPattern = DependencyGraphPattern
        .of()
            // trigger
        .addVertex(1, v -> v.getLemma().equals("identification"))
            // each
        .addVertex(2, v -> v.getLemma().equals("interaction") || v.getLemma().equals("binding"))
        .addEdge(1, 2, Dependency.ofPrepPredicate("of"));
    extractAnnotation(identificationPattern);
  }

  /**
   * 0 and 2 are two arguments, 1 is trigger
   *
   * @param pattern input toUnderstandPattern
   */
  private void extractAnnotation(DependencyGraphPattern pattern) {

    int annIndex = bratdoc.getEntities().size();

    DependencyGraphMatcher matcher = pattern.matcher(graph);
    if (matcher.find()) {
      BratEntity e = new BratEntity();
      e.setId("T" + annIndex);
      e.setType("Interaction-Sentence");
      Range<Integer> span = getRange();
      e.addSpan(span);
      e.setText(bratdoc.getText().substring(span.lowerEndpoint(), span.upperEndpoint()));
      bratdoc.addAnnotation(e);
    }
  }

  private Range<Integer> getRange() {
    RangeSet<Integer> rs = TreeRangeSet.create();
    for(DependencyGraphVertex v: graph.vertexSet()) {
      rs.add(Range.closedOpen(v.getBeginPosition(), v.getEndPosition()));
    }
    return rs.span();
  }
}
