package com.pengyifan.biocreativev;

import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.bionlpst.trigger.Trigger;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.commons.math.RangeUtils;
import java.util.Collection;
import java.util.Optional;
import java.util.function.Predicate;

public class PPIExtractorBasic {

  private static final Predicate<DependencyGraphVertex> bindingPred =
      v -> v.getLemma().equals("binding")
          || v.getLemma().equals("interaction");

  private static final Predicate<DependencyGraphVertex> indirectPred =
      v -> v.getLemma().equals("block")
          || v.getLemma().equals("decrease")
          || v.getLemma().equals("impair")
          || v.getLemma().equals("induce")
          || v.getLemma().equals("mediate")
          || v.getLemma().equals("regulate")
          || v.getLemma().equals("repress")
          || v.getLemma().equals("require")
          || v.getLemma().equals("restrict")
          || v.getLemma().equals("stimulate")
          || v.getLemma().equals("trigger");


  final Collection<Trigger> triggers;
  final DependencyGraph graph;
  final BratDocument bratdoc;

  public PPIExtractorBasic(DependencyGraph graph, BratDocument bratdoc, Collection<Trigger> triggers) {
    this.graph = graph;
    this.bratdoc = bratdoc;
    this.triggers = triggers;
  }

  public void extract() {
    extractDirect();
    extractIndirect();
    extractReceptor();
  }

  private void extractDirect() {
    /**
     * e1 < arg0 < trigger > arg1 > e2
     */
    final DependencyGraphPattern basicPattern = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, this::isBasicTrigger)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1));
    extractPredicateArgument(basicPattern, null, "Basic");

    //    /**
    //     * binding of e1 and e2
    //     * <p>
    //     * e1 and e2 interaction
    //     */
    //    final DependencyGraphPattern bindingPattern = DependencyGraphPattern
    //        .of()
    //            // arg0
    //        .addVertex(0, v -> v.getNer().isProtein())
    //            // trigger
    //        .addVertex(1, bindingPred)
    //            // arg1
    //        .addVertex(2, v -> v.getNer().isProtein())
    //        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
    //        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
    //        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ));
    //    extractPredicateArgument(bindingPattern, DependencyName.ARG1);

    /**
     * interaction between e1 and e2
     */
    final Predicate<Dependency> predE =
        Dependency.ofPredicate(DependencyName.ARG0, "between")
            .or(Dependency.ofPredicate(DependencyName.ARG1, "between"))
            .or(Dependency.ofPredicate(DependencyName.ARG0, "null_involve"))
            .or(Dependency.ofPredicate(DependencyName.ARG1, "null_involve"))
            .or(Dependency.ofPredicate(DependencyName.ARG0, "null_occur"))
            .or(Dependency.ofPredicate(DependencyName.ARG1, "null_occur"));

    final DependencyGraphPattern betweenPattern = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, bindingPred)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, predE)
        .addEdge(1, 2, predE);
    extractPredicateArgument(betweenPattern, null, "Basic");

    //    /**
    //     * complex of e1 and e2
    //     */
    //    final DependencyGraphPattern complexPattern = DependencyGraphPattern
    //        .of()
    //            // arg0
    //        .addVertex(0, v -> v.getNer().isProtein())
    //            // trigger
    //        .addVertex(1, PartWholeLinkAdder.complexPredicate)
    //            // arg1
    //        .addVertex(2, v -> v.getNer().isProtein())
    //        .addEdge(0, 1, Dependency.ofPredicate(DependencyName.PART_WHOLE))
    //        .addEdge(2, 1, Dependency.ofPredicate(DependencyName.PART_WHOLE));
    // extractPredicateArgument(complexPattern, null, "Basic-Noun");

    /**
     * interact with each other
     */
    final DependencyGraphPattern eachOtherPattern = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, this::isBasicTrigger)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        // each
        .addVertex(3, v -> v.getWord().equals("each"))
        // other
        .addVertex(4, v -> v.getWord().equals("other"))
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 4, d -> true)
        .addEdge(4, 3, Dependency.ofPredicate(DependencyName.DET));
    extractPredicateArgument(eachOtherPattern, DependencyName.ARG1, "Basic");

    final DependencyGraphPattern directlyPattern = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, this::isBasicTrigger)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        // directly
        .addVertex(3, v -> v.getWord().equals("directly"))
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 3, d -> true);
    extractPredicateArgument(directlyPattern, DependencyName.ARG1, "Basic");
    /**
     * e1 < arg0 < trigger > arg0 > e2
     */
    final DependencyGraphPattern sameArgPattern = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, this::isBasicTrigger)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ, "dash"));
    extractPredicateArgument(sameArgPattern, DependencyName.ARG1, "Basic");
    /**
     * e1 < arg0 < trigger > arg0 > e2
     */
    final DependencyGraphPattern sameArgPattern2 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, this::isBasicTrigger)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ, "dash"));
    extractPredicateArgument(sameArgPattern2, DependencyName.ARG0, "Basic");
    /**
     * e1 < arg0 < trigger > arg0 > e2
     */
    final DependencyGraphPattern sameArgPattern3 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("binding") || v.getLemma().equals("interaction"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0));
    extractPredicateArgument(sameArgPattern3, DependencyName.ARG1, "Basic");
    /**
     * e1 < arg0 < trigger > arg0 > e2
     */
    final DependencyGraphPattern sameArgPattern4 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("associate"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1));
    extractPredicateArgument(sameArgPattern4, DependencyName.ARG0, "Basic");
  }

  private void extractReceptor() {
    /**
     * e1 < arg1 < ligand > arg1 > e2
     */
    final DependencyGraphPattern p1 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("ligand") || v.getLemma().equals("receptor")
            || v.getLemma().equals("agonist"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPrepPredicate("for"));
    extractPredicateArgument(p1, null, "ligand");

    /**
     * e1 < is_a < ligand > arg1 > e2
     */
    final DependencyGraphPattern p2 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("receptor"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1));
    extractPredicateArgument(p2, DependencyName.ARG0, "receptor");
  }

  private void extractIndirect() {

    final Predicate<DependencyGraphVertex> processPredicate =
        v -> v.getLemma().equals("activity")
            || v.getLemma().equals("ability")
            //|| v.getLemma().equals("binding")
            || (v.getLemma().endsWith("ion") && (this.endsWithTrigger(v)));
    /**
     * N1 < arg1 < process < arg0 < indirect trigger > arg1 > process > arg1 > N2
     */
    final DependencyGraphPattern blockPattern = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(3, processPredicate)
        // trigger
        .addVertex(1, indirectPred)
        // x
        .addVertex(4, processPredicate)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 4, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(3, 0, Dependency.ofPrepPredicate("of"))
        .addEdge(4, 2, Dependency.ofPrepPredicate("of"));
    extractPredicateArgument(blockPattern, null, "Indirect");
    /**
     * N1 < arg1 < process < arg0 < indirect trigger > arg1 > process > arg1 > N2
     */
    final DependencyGraphPattern blockPattern6 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(3, v -> v.getLemma().equals("activation"))
        // trigger
        .addVertex(1, indirectPred)
        // x
        .addVertex(4, processPredicate)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 4, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(3, 0, Dependency.ofPrepPredicate("of"))
        .addEdge(4, 2, Dependency.ofPrepPredicate("of"));
    extractPredicateArgument(blockPattern6, null, "Indirect");
    /**
     * N1 < arg1 < process < arg0 < indirect trigger > arg1 > N2
     */
    final DependencyGraphPattern blockPattern2 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(3, processPredicate)
        // trigger
        .addVertex(1, indirectPred)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(3, 0, Dependency.ofPrepPredicate("of"));
    extractPredicateArgument(blockPattern2, null, "Indirect");
    //    /**
    //     * N1 < arg1 < process < arg0 < indirect trigger > arg1 > N2
    //     */
    //    final DependencyGraphPattern blockPattern9 = DependencyGraphPattern
    //        .of()
    //            // n1
    //        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
    //            // x
    //        .addVertex(3, processPredicate)
    //            // trigger
    //        .addVertex(1, indirectPred)
    //            // n2
    //        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
    //        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG0))
    //        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1))
    //        .addEdge(3, 0, Dependency.ofPrepPredicate("of"));
    //    extractPredicateArgument(blockPattern9, null, "Indirect");
    /**
     * N1 < arg0 < indirect trigger > arg1 > process > arg1 > N2
     */
    final DependencyGraphPattern blockPattern5 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, indirectPred)
        // x
        .addVertex(4, processPredicate)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 4, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(4, 2, Dependency.ofPrepPredicate("of"));
    //Dependency.ofPredicate(DependencyName.ARG0)));
    extractPredicateArgument(blockPattern5, null, "Indirect");
    /**
     * N1 < arg1 < v < nsubj < blocked > arg0 > ? > arg1 > N2
     */
    final DependencyGraphPattern blockPattern3 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(3, processPredicate)
        // trigger
        .addVertex(1, v -> indirectPred.test(v) && v.getTag().equals(PartOfSpeech.VBN))
        // x
        .addVertex(4, processPredicate)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.NSUBJ))
        .addEdge(1, 4, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(3, 0, Dependency.ofPrepPredicate("of"))
        .addEdge(4, 2, Dependency.ofPrepPredicate("of"));
    extractPredicateArgument(blockPattern3, null, "Indirect");
    /**
     * N1 < arg1 < v < nsubj < blocked > arg0 > N2
     */
    final DependencyGraphPattern blockPattern4 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(3, processPredicate)
        // trigger
        .addVertex(1, v -> indirectPred.test(v) && v.getTag().equals(PartOfSpeech.VBN))
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.NSUBJ))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(3, 0, Dependency.ofPrepPredicate("of"));
    extractPredicateArgument(blockPattern4, null, "Indirect");
    /**
     * N1 < arg1 < v < nsubj < blocked > arg0 > N2
     */
    final DependencyGraphPattern blockPattern7 = DependencyGraphPattern
        .of()
        // n1
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // x
        .addVertex(1, v -> indirectPred.test(v) && v.getTag().equals(PartOfSpeech.VBN))
        // x
        .addVertex(4, processPredicate)
        // n2
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 4, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(4, 2, Dependency.ofPredicate(DependencyName.NN));
    // extractPredicateArgument(blockPattern7, null, "Indirect");
  }

  private boolean isBasicTrigger(DependencyGraphVertex v) {
    for (Trigger trigger : triggers) {
      if (trigger.getType().equals("Basic")) {
        if (v.getLemma().equalsIgnoreCase(trigger.getLemma().get())) {
          return true;
        }
      }
    }
    return false;
  }

  private boolean endsWithTrigger(DependencyGraphVertex v) {
    for (Trigger trigger : triggers) {
      if (trigger.getType().equals("Basic") || trigger.getType().equals("PTM")) {
        if (v.getLemma().endsWith(trigger.getLemma().get())) {
          return true;
        }
      }
    }
    return false;
  }

  private boolean isPTMTrigger(DependencyGraphVertex v) {
    for (Trigger trigger : triggers) {
      if (trigger.getType().equals("PTM")) {
        if (v.getLemma().equalsIgnoreCase(trigger.getLemma().get())) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * 0 and 2 are two arguments, 1 is trigger
   *
   * @param pattern input toUnderstandPattern
   */
  private void extractPredicateArgument_Arg(DependencyGraphPattern pattern,
      DependencyName dependencyName, String subtype) {

    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();

    DependencyGraphMatcher matcher = pattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(0);
      DependencyGraphVertex t = matcher.get(2);
      DependencyGraphVertex trigger = matcher.get(1);

      Optional<DependencyGraphVertex> neg = getNeg(trigger);
      if (neg.isPresent() && between(s, t, neg.get())) {
        continue;
      }

      if (dependencyName != null && has(matcher.getGraph(), trigger, dependencyName)) {
        continue;
      }

      //      if (dependencyName != null &&
      //          !graph.hasOutgoingEdgesOf(trigger, Dependency.ofPredicate(dependencyName))) {
      //        continue;
      //      }

      if (RangeUtils.isOverlapped(Range.closedOpen(s.getBeginPosition(), s.getEndPosition()),
          Range.closedOpen(t.getBeginPosition(), t.getEndPosition()))) {
        continue;
      }

      if (s.getBeginPosition() == t.getBeginPosition()) {
        continue;
      }

      BratEntity e1 = find(s);
      if (e1 == null) {
        e1 = convert(s, annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(t);
      if (e2 == null) {
        e2 = convert(t, annIndex, "Protein");
        bratdoc.addAnnotation(e2);
        annIndex++;
      }

      BratEntity triggerEntity = find(trigger);
      if (triggerEntity == null) {
        triggerEntity = convert(trigger, annIndex, "Basic-Trigger");
        bratdoc.addAnnotation(triggerEntity);
        annIndex++;
      }

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex);
        relation.setType("Interaction-" + subtype);
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        relation.putArgument("Trigger", triggerEntity.getId());
        bratdoc.addAnnotation(relation);
        relIndex++;
      }
    }
  }

  /**
   * 0 and 2 are two arguments, 1 is trigger
   *
   * @param pattern input toUnderstandPattern
   */
  private void extractPredicateArgument(DependencyGraphPattern pattern,
      DependencyName dependencyName, String subtype) {

    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();

    DependencyGraphMatcher matcher = pattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(0);
      DependencyGraphVertex t = matcher.get(2);
      DependencyGraphVertex trigger = matcher.get(1);

      Optional<DependencyGraphVertex> neg = getNeg(trigger);
      if (neg.isPresent() && between(s, t, neg.get())) {
        continue;
      }

      if (dependencyName != null && has(matcher.getGraph(), trigger, dependencyName)) {
        continue;
      }

      if (RangeUtils.isOverlapped(Range.closedOpen(s.getBeginPosition(), s.getEndPosition()),
          Range.closedOpen(t.getBeginPosition(), t.getEndPosition()))) {
        continue;
      }

      if (s.getBeginPosition() == t.getBeginPosition()) {
        continue;
      }

      //      if (s.getLemma().contains("transcription factor")
      //          || t.getLemma().contains("transcription factor")) {
      //        continue;
      //      }

      //      if ((hasComplex(s) && !hasComplex(t)) || (!hasComplex(s) && hasComplex(t))) {
      //        continue;
      //      }

      BratEntity e1 = find(s);
      if (e1 == null) {
        e1 = convert(s, annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(t);
      if (e2 == null) {
        e2 = convert(t, annIndex, "Protein");
        bratdoc.addAnnotation(e2);
        annIndex++;
      }

      BratEntity triggerEntity = find(trigger);
      if (triggerEntity == null) {
        triggerEntity = convert(trigger, annIndex, "Basic-Trigger");
        bratdoc.addAnnotation(triggerEntity);
        annIndex++;
      }

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex);
        relation.setType("Interaction-" + subtype);
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        relation.putArgument("Trigger", triggerEntity.getId());
        bratdoc.addAnnotation(relation);
        relIndex++;
      }
    }
  }

  private boolean hasComplex(DependencyGraphVertex s) {
    //    if (s.getLemma().contains("-")) {
    //      return false;
    //    }
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(s)) {
      if (e.getDependency().getName().equals(DependencyName.PART_WHOLE)
          && e.getDependency().getSpecific().orElse("").contains("complex")) {
        return true;
      }
    }
    return false;
  }

  boolean has(DependencyGraph graph, DependencyGraphVertex v,
      DependencyName dependencyName) {
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(v)) {
      if (e.getDependency().getName().equals(dependencyName)) {
        return true;
      }
    }
    return false;
  }

  boolean between(DependencyGraphVertex e1, DependencyGraphVertex e2, DependencyGraphVertex v) {
    RangeSet<Integer> set = TreeRangeSet.create();
    set.add(Range.closedOpen(e1.getBeginPosition(), e1.getEndPosition()));
    set.add(Range.closedOpen(e2.getBeginPosition(), e2.getEndPosition()));
    Range<Integer> r = set.span();
    return r.contains(v.getBeginPosition()) && r.contains(v.getEndPosition());
  }

  Optional<DependencyGraphVertex> getNeg(DependencyGraphVertex trigger) {
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(trigger)) {
      if (e.getDependency().getName().equals(DependencyName.NEG)) {
        return Optional.of(graph.getEdgeTarget(e));
      }
    }
    for (DependencyGraphVertex s : graph.incomingEdgeSourcesOf(trigger)) {
      if (s.getLemma().equals("fail")) {
        return Optional.of(s);
      }
      if (s.getLemma().equals("cannot")) {
        return Optional.of(s);
      }
    }
    return Optional.empty();
  }

  BratRelation find(String e1Id, String e2Id) {
    return bratdoc.getRelations().stream()
        .filter(r -> r.containsRole("Arg1") && r.containsRole("Arg2"))
        .filter(r -> (r.getArgId("Arg1").equals(e1Id) && r.getArgId("Arg2").equals(e2Id))
            || (r.getArgId("Arg2").equals(e1Id) && r.getArgId("Arg1").equals(e2Id)))
        .findFirst()
        .orElse(null);
  }

  BratEntity find(DependencyGraphVertex v) {
    return bratdoc.getEntities().stream()
        .filter(e -> e.beginPosition() == v.getBeginPosition()
            && e.endPosition() == v.getEndPosition())
        .findFirst()
        .orElse(null);
  }

  BratEntity convert(DependencyGraphVertex v, int index, String type) {
    BratEntity e = new BratEntity();
    e.setId("T" + index);
    e.setType(type);
    e.setText(v.getWord());
    e.addSpan(v.getBeginPosition(), v.getEndPosition());
    return e;
  }
}
