package com.pengyifan.biocreativev;

import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.math.RangeUtils;
import com.pengyifan.nlp.ssdg.PartWholeLinkAdder;
import java.util.Optional;

public class PPIExtractorComplex {

  final DependencyGraph graph;
  final BratDocument bratdoc;

  public PPIExtractorComplex(DependencyGraph graph, BratDocument bratdoc) {
    this.graph = graph;
    this.bratdoc = bratdoc;
  }

  public void extract() {
    extractDirect();
  }

  private void extractDirect() {
    /**
     * complex of e1 and e2
     * <p>
     * e1 and e2 interaction
     */
    final DependencyGraphPattern p1 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPrepPredicate("of"))
        .addEdge(1, 2, Dependency.ofPrepPredicate("of"))
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ));
    extractPredicateArgument(p1, null, "Complex");

    /**
     * complex between e1 and e2
     */

    final DependencyGraphPattern p2 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1, "between"))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0, "between"));
    extractPredicateArgument(p2, null, "Complex");

    /**
     * e1-e2 complex
     */
    final DependencyGraphPattern p3 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NN)
            .or(Dependency.ofPredicate(DependencyName.NUM)))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.NN)
            .or(Dependency.ofPredicate(DependencyName.NUM)))
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ, "dash"));
    extractPredicateArgument(p3, null, "Complex");

    /**
     * e1 complexed with e2
     */
    final DependencyGraphPattern p4 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1, "vmod"))
        .addEdge(1, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p4, null, "Complex");

    /**
     * e1 and e2 form a complex
     */
    final DependencyGraphPattern p5 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("form"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ))
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.DOBJ));
    extractPredicateArgument(p5, null, "Complex");

    /**
     * e1 form complex with e2
     */
    final DependencyGraphPattern p6 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("form"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(3, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p6, null, "Complex");

    /**
     * e1 found in complex with e2
     */
    final DependencyGraphPattern p7 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("find"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 3, Dependency.ofPrepPredicate("in"))
        .addEdge(3, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p7, null, "Complex");

    /**
     * e1 in complex with e2
     */
    final DependencyGraphPattern p8 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, PartWholeLinkAdder.complexPredicate)
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(0, 1, Dependency.ofPrepPredicate("in"))
        .addEdge(1, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p8, null, "Complex");

    /**
     * e1 exist in complex with e2
     */
    final DependencyGraphPattern p9 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("exist"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 3, Dependency.ofPrepPredicate("in"))
        .addEdge(3, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p9, null, "Complex");

    /**
     * e1 assembled as complex with e2
     */
    final DependencyGraphPattern p10 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("assemble"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 3, Dependency.ofPrepPredicate("as"))
        .addEdge(3, 2, Dependency.ofPrepPredicate("with"));
    extractPredicateArgument(p10, null, "Complex");

    /**
     * complex is composed of e1 and e2
     */
    final DependencyGraphPattern p11 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // trigger
        .addVertex(1, v -> v.getLemma().equals("compose"))
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addVertex(3, PartWholeLinkAdder.complexPredicate)
        .addEdge(1, 3, Dependency.ofPredicate(DependencyName.ARG1))
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG0));
    extractPredicateArgument(p11, null, "Complex");

    /**
     * e1 - e2
     */
    final DependencyGraphPattern p12 = DependencyGraphPattern
        .of()
        // arg0
        .addVertex(0, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        // arg1
        .addVertex(2, v -> v.getNer().isProtein() || v.getNer().isProteinIt())
        .addEdge(0, 2, Dependency.ofPredicate(DependencyName.CONJ, "dash"));
    extractPredicateArgument_noComplex(p12, "Dash");
  }

  private void extractPredicateArgument_noComplex(DependencyGraphPattern pattern, String subtype) {
    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();
    DependencyGraphMatcher matcher = pattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(0);
      DependencyGraphVertex t = matcher.get(2);

      if (hasIncoming(s, DependencyName.NN) || hasIncoming(t, DependencyName.NN)
          || hasIncoming(s, DependencyName.AMOD) || hasIncoming(t, DependencyName.AMOD)) {
        continue;
      }
      if (!isDash(s, t)) {
        continue;
      }

      BratEntity e1 = find(s);
      if (e1 == null) {
        e1 = convert(s, annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(t);
      if (e2 == null) {
        e2 = convert(t, annIndex, "Protein");
        bratdoc.addAnnotation(e2);
        annIndex++;
      }

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex);
        relation.setType("Interaction-" + subtype);
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        bratdoc.addAnnotation(relation);
        relIndex++;
      }
    }
  }

  private boolean isDash(DependencyGraphVertex s, DependencyGraphVertex t) {
    if (s.getBeginPosition() > t.getBeginPosition()) {
      DependencyGraphVertex tmp = s;
      s = t;
      t = tmp;
    }
    String subtext = bratdoc.getText().substring(s.getEndPosition(), t.getBeginPosition());
    return subtext.equals("-");
  }

  /**
   * 0 and 2 are two arguments, 1 is trigger
   *
   * @param pattern input toUnderstandPattern
   */
  private void extractPredicateArgument(DependencyGraphPattern pattern,
      DependencyName dependencyName, String subtype) {

    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();

    DependencyGraphMatcher matcher = pattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(0);
      DependencyGraphVertex t = matcher.get(2);
      DependencyGraphVertex trigger = matcher.get(1);

      // three complex
      //      int num = 0;
      //      for(DependencyGraphVertex v: graph.outgoingEdgeTargetsOf(trigger)) {
      //        if (v.getNer().isProtein()) {
      //          num ++;
      //        }
      //      }
      //      if (num > 2) {
      //        continue;
      //      }

      Optional<DependencyGraphVertex> neg = getNeg(trigger);
      if (neg.isPresent() && between(s, t, neg.get())) {
        continue;
      }

      if (dependencyName != null && hasOutgoing(trigger, dependencyName)) {
        continue;
      }

      if (RangeUtils.isOverlapped(Range.closedOpen(s.getBeginPosition(), s.getEndPosition()),
          Range.closedOpen(t.getBeginPosition(), t.getEndPosition()))) {
        continue;
      }

      if (s.getLemma().equals(t.getLemma())) {
        continue;
      }

      //      if ((hasComplex(s) && !hasComplex(t)) || (!hasComplex(s) && hasComplex(t))) {
      //        continue;
      //      }

      BratEntity e1 = find(s);
      if (e1 == null) {
        e1 = convert(s, annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(t);
      if (e2 == null) {
        e2 = convert(t, annIndex, "Protein");
        bratdoc.addAnnotation(e2);
        annIndex++;
      }

      BratEntity triggerEntity = find(trigger);
      if (triggerEntity == null) {
        triggerEntity = convert(trigger, annIndex, "Complex-Trigger");
        bratdoc.addAnnotation(triggerEntity);
        annIndex++;
      }

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex);
        relation.setType("Interaction-" + subtype);
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        relation.putArgument("Trigger", triggerEntity.getId());
        bratdoc.addAnnotation(relation);
        relIndex++;
      }
    }
  }

  boolean hasIncoming(DependencyGraphVertex target, DependencyName dependencyName) {
    return graph.hasIncomingEdgesOf(target, Dependency.ofPredicate(dependencyName));
  }

  boolean hasIncoming(DependencyGraphVertex target, String sourceLemma) {
    for (DependencyGraphVertex v : graph.incomingEdgeSourcesOf(target)) {
      if (v.getLemma().equals(sourceLemma)) {
        return true;
      }
    }
    return false;
  }

  boolean hasOutgoing(DependencyGraphVertex v, DependencyName dependencyName) {
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(v)) {
      if (e.getDependency().getName().equals(dependencyName)) {
        return true;
      }
    }
    return false;
  }

  boolean between(DependencyGraphVertex e1, DependencyGraphVertex e2, DependencyGraphVertex v) {
    RangeSet<Integer> set = TreeRangeSet.create();
    set.add(Range.closedOpen(e1.getBeginPosition(), e1.getEndPosition()));
    set.add(Range.closedOpen(e2.getBeginPosition(), e2.getEndPosition()));
    Range<Integer> r = set.span();
    return r.contains(v.getBeginPosition()) && r.contains(v.getEndPosition());
  }

  Optional<DependencyGraphVertex> getNeg(DependencyGraphVertex trigger) {
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(trigger)) {
      if (e.getDependency().getName().equals(DependencyName.NEG)) {
        return Optional.of(graph.getEdgeTarget(e));
      }
    }
    for (DependencyGraphVertex s : graph.incomingEdgeSourcesOf(trigger)) {
      if (s.getLemma().equals("fail")) {
        return Optional.of(s);
      }
      if (s.getLemma().equals("cannot")) {
        return Optional.of(s);
      }
    }
    return Optional.empty();
  }

  BratRelation find(String e1Id, String e2Id) {
    return bratdoc.getRelations().stream()
        .filter(r -> r.containsRole("Arg1") && r.containsRole("Arg2"))
        .filter(r -> (r.getArgId("Arg1").equals(e1Id) && r.getArgId("Arg2").equals(e2Id))
            || (r.getArgId("Arg2").equals(e1Id) && r.getArgId("Arg1").equals(e2Id)))
        .findFirst()
        .orElse(null);
  }

  BratEntity find(DependencyGraphVertex v) {
    return bratdoc.getEntities().stream()
        .filter(e -> e.beginPosition() == v.getBeginPosition()
            && e.endPosition() == v.getEndPosition())
        .findFirst()
        .orElse(null);
  }

  BratEntity convert(DependencyGraphVertex v, int index, String type) {
    BratEntity e = new BratEntity();
    e.setId("T" + index);
    e.setType(type);
    e.setText(v.getWord());
    e.addSpan(v.getBeginPosition(), v.getEndPosition());
    return e;
  }
}
