package com.pengyifan.biocreativev;

import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.commons.io.BasenameUtils;
import com.pengyifan.kernel.MyProperties;
import com.pengyifan.kernel.allpathgraph.module.ApgConfiguration;
import com.pengyifan.kernel.allpathgraph.module.DefaultApgConfiguration;
import com.pengyifan.nlp.trees.TreebankReader;
import edu.stanford.nlp.trees.MemoryTreebank;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.Treebank;
import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.logging.Logger;
import org.apache.commons.io.FilenameUtils;

public class PPIExtractorComplexBatch extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("aimed");
    String dataset = "dataset";
    Path root = prop.getTempDir().resolve(dataset);

    PPIExtractorComplexBatch s = new PPIExtractorComplexBatch(
        new DefaultApgConfiguration(),
        BasenameUtils.getTextBasenames(root.resolve("brat")));

    s.bratInputDir = root.resolve("brat");
    s.parsingInputDir = root.resolve("ptb2");
    s.outputDir = root.resolve("ppi-complex");
    s.process();
  }


  protected static final Logger logger = Logger.getAnonymousLogger();

  final ApgConfiguration config;

  protected Path bratInputDir;
  protected Path parsingInputDir;
  protected Path outputDir;

  String text;
  BratDocument a1Doc;
  BratDocument corefDoc;
  BratDocument abbrDoc;
  Treebank treebank;

  BratDocument ppiDoc;

  protected PPIExtractorComplexBatch(ApgConfiguration config, Collection<String> basenames) {
    super(basenames);
    this.config = config;
  }

  private boolean isFileEmtpy(Path file) throws IOException {
    BufferedReader r = Files.newBufferedReader(file);
    boolean b = r.readLine() == null;
    r.close();
    return b;
  }

  private boolean isFiltered(Path file) throws IOException {
    String basename = FilenameUtils.getBaseName(file.toString());
    return basename.equals("23904265_12301");
  }

  @Override
  protected void preprocess()
      throws IOException {
    if (!Files.exists(outputDir)) {
      Files.createDirectory(outputDir);
    }
  }

  @Override
  protected void preprocessFile(String basename)
      throws IOException {

    Path textFile = bratInputDir.resolve(basename + ".txt");
    text = new String(Files.readAllBytes(textFile));

    Path a1File = bratInputDir.resolve(basename + ".a1");
    if (isFileEmtpy(a1File) || isFiltered(a1File)) {
      treebank = new MemoryTreebank();
      return;
    }

    try {
      a1Doc = BratIOUtils.read(Files.newBufferedReader(a1File), basename);
    } catch (Exception e) {
      a1Doc = new BratDocument();
    }

    corefDoc = new BratDocument();
    abbrDoc = new BratDocument();

    Path ptbFile = parsingInputDir.resolve(basename + ".ptb");
    if (Files.exists(ptbFile)) {
      TreebankReader treebankReader = new TreebankReader(ptbFile.toFile(),
          config.getTreeBuilder());
      treebank = treebankReader.read();
      treebankReader.close();
      //      checkState(text.isEmpty() || !treebank.isEmpty(), "Treebank is empty");
    } else {
      treebank = new MemoryTreebank();
    }
  }

  @Override
  protected void processFile(String basename)
      throws Exception {

    logger.info(String.format("Processing file: %s", basename));

    ppiDoc = new BratDocument();
    ppiDoc.setDocId(basename);
    ppiDoc.setText(text);

    try {
      for (Tree tree : treebank) {
        if (tree.getLeaves().size() >= 100) {
          continue;
        }
        // get sdg
        DependencyGraph sdg = new SDGCreater(config.getGrammaticalStructureBuilder())
            .setDependencyType(config.getDependencyType())
            .setText(text)
            .setTree(tree)
            .create();
        // get ssdg
        DependencyGraph ssdg = new EDGCreater()
            .setText(text)
            .setA1Doc(a1Doc)
            .setCorefDoc(corefDoc)
            .setAbbrDoc(abbrDoc)
            .setDependencyGraph(sdg)
            .create();
        // extract
        PPIExtractorComplex ppi = new PPIExtractorComplex(ssdg, ppiDoc);
        ppi.extract();
      }
    } catch (Exception e) {

    }
  }

  @Override
  protected void postprocessFile(String basename) throws Exception {
    Path ppiFile = outputDir.resolve(basename + ".ann");
    BratIOUtils.write(Files.newBufferedWriter(ppiFile), ppiDoc);
  }
}
