package com.pengyifan.biocreativev;

import com.google.common.collect.Range;
import com.pengyifan.bionlpst.trigger.Trigger;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.dependencygraph.*;
import com.pengyifan.commons.math.RangeUtils;

import java.util.Collection;
import java.util.Optional;

public class PPIExtractorExtended extends PPIExtractorBasic {

  public PPIExtractorExtended(DependencyGraph graph, BratDocument bratdoc,
      Collection<Trigger> triggers) {
    super(graph, bratdoc, triggers);
  }

  public void extract() {
    /**
     * e1 < arg0 < trigger > arg1 > e2
     */
    final DependencyGraphPattern basicExtendedPattern = DependencyGraphPattern
        .of()
            // arg0
        .addVertex(0, v -> v.getNer().isProtein())
            // trigger
        .addVertex(1, this::isExtendedTrigger)
            // arg1
        .addVertex(2, v -> v.getNer().isProtein())
        .addEdge(1, 0, Dependency.ofPredicate(DependencyName.ARG0))
        .addEdge(1, 2, Dependency.ofPredicate(DependencyName.ARG1));
    extractExtendedPredicateArgument(basicExtendedPattern, null);
  }

  private boolean isExtendedTrigger(DependencyGraphVertex v) {
    for (Trigger trigger : triggers) {
      if (trigger.getType().equals("Extended")) {
        if (v.getLemma().equalsIgnoreCase(trigger.getLemma().get())) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * 0 and 2 are two arguments, 1 is trigger
   *
   * @param pattern input toUnderstandPattern
   */
  private void extractExtendedPredicateArgument(DependencyGraphPattern pattern,
      DependencyName dependencyName) {

    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();

    DependencyGraphMatcher matcher = pattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(0);
      DependencyGraphVertex t = matcher.get(2);
      DependencyGraphVertex trigger = matcher.get(1);

      Optional<DependencyGraphVertex> neg = getNeg(trigger);
      if (neg.isPresent() && between(s, t, neg.get())) {
        continue;
      }

      if (dependencyName != null && has(matcher.getGraph(), trigger, dependencyName)) {
        continue;
      }

      if (RangeUtils.isOverlapped(Range.closedOpen(s.getBeginPosition(), s.getEndPosition()),
          Range.closedOpen(t.getBeginPosition(), t.getEndPosition()))) {
        continue;
      }

      if (s.getLemma().equals(t.getLemma())) {
        continue;
      }

      BratEntity e1 = find(s);
      if (e1 == null) {
        e1 = convert(s, annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(t);
      if (e2 == null) {
        e2 = convert(t, annIndex, "Protein");
        bratdoc.addAnnotation(e2);
        annIndex++;
      }

      BratEntity triggerEntity = find(trigger);
      if (triggerEntity == null) {
        triggerEntity = convert(trigger, annIndex, "Extended-Trigger");
        bratdoc.addAnnotation(triggerEntity);
        annIndex++;
      }

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex);
        relation.setType("Interaction-Extended");
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        relation.putArgument("Trigger", triggerEntity.getId());
        bratdoc.addAnnotation(relation);
        relIndex++;
      }
    }
  }
}
