package com.pengyifan.biocreativev;

import com.google.common.collect.Range;
import com.google.common.collect.Sets;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class PPIExtractorNeg {

  static final Logger logger = Logger.getAnonymousLogger();

  final static Pattern toUnderstandPattern = Pattern.compile("^(to) (assess|explore|investigate|understand)",
      Pattern.CASE_INSENSITIVE);
  final static Pattern weTestPattern = Pattern.compile("\\bwe " +
          "(test|devise|explorer|investigat|hypothesiz)",
      Pattern.CASE_INSENSITIVE);

  final String text;
  final int offset;
  final DependencyGraph graph;
  final List<DependencyGraphVertex> list;
  final BratDocument ppiDoc;
  final BratDocument outputDoc;

  public PPIExtractorNeg(int offset, String text, DependencyGraph graph, BratDocument ppiDoc,
      BratDocument
          outputDoc) {
    this.offset = offset;
    this.text = text;
    this.graph = graph;
    this.ppiDoc = ppiDoc;
    this.outputDoc = outputDoc;
    this.list = graph.getSortedVertexList();
  }

  public void extract() {
    int relId = outputDoc.getRelations().size();
    int entId = outputDoc.getEntities().size();
    for (BratRelation bratRelation : ppiDoc.getRelations()) {
      BratEntity e1 = ppiDoc.getEntity(bratRelation.getArgId("Arg1"));
      int e1Index = indexOfVertex(e1);
      if (e1Index == -1) {
        continue;
      }
      BratEntity e2 = ppiDoc.getEntity(bratRelation.getArgId("Arg2"));
      int e2Index = indexOfVertex(e2);
      if (e2Index == -1) {
        continue;
      }
      BratEntity trigger = ppiDoc.getEntity(bratRelation.getArgId("Trigger"));
      int triggerIndex = indexOfVertex(trigger);
      if (triggerIndex == -1) {
        continue;
      }
      Optional<DependencyGraphVertex> negOpt = hasNeg(e1Index, triggerIndex);
      if (!negOpt.isPresent()) {
        negOpt = hasNeg(e2Index, triggerIndex);
      }
      if (negOpt.isPresent()) {
        BratEntity newE1 = new BratEntity(e1);
        newE1.setId("T" + entId++);
        outputDoc.addAnnotation(newE1);

        BratEntity newE2 = new BratEntity(e2);
        newE2.setId("T" + entId++);
        outputDoc.addAnnotation(newE2);

        BratEntity newT = new BratEntity(trigger);
        newT.setId("T" + entId++);
        outputDoc.addAnnotation(newT);

        //neg
        BratEntity neg = convert(negOpt.get(), entId++, "Neg");
        outputDoc.addAnnotation(neg);

        // rel
        BratRelation rel = new BratRelation();
        rel.setId("R" + relId++);
        rel.setType("Interaction-NegNot");
        rel.putArgument("Arg1", newE1.getId());
        rel.putArgument("Arg2", newE2.getId());
        rel.putArgument("Trigger", newT.getId());
        rel.putArgument("Neg", neg.getId());
        outputDoc.addAnnotation(rel);
      }
      if (isProteinPlural(e1Index)) {
        BratEntity newE1 = new BratEntity(e1);
        newE1.setId("T" + entId++);
        outputDoc.addAnnotation(newE1);

        BratEntity newE2 = new BratEntity(e2);
        newE2.setId("T" + entId++);
        outputDoc.addAnnotation(newE2);

        BratEntity newT = new BratEntity(trigger);
        newT.setId("T" + entId++);
        outputDoc.addAnnotation(newT);

        // rel
        BratRelation rel = new BratRelation();
        rel.setId("R" + relId++);
        rel.setType("Interaction-NegPlural");
        rel.putArgument("Arg1", newE1.getId());
        rel.putArgument("Arg2", newE2.getId());
        rel.putArgument("Trigger", newT.getId());
        rel.putArgument("Neg", newE1.getId());
        outputDoc.addAnnotation(rel);
      } else if (isProteinPlural(e2Index)) {
        BratEntity newE1 = new BratEntity(e1);
        newE1.setId("T" + entId++);
        outputDoc.addAnnotation(newE1);

        BratEntity newE2 = new BratEntity(e2);
        newE2.setId("T" + entId++);
        outputDoc.addAnnotation(newE2);

        BratEntity newT = new BratEntity(trigger);
        newT.setId("T" + entId++);
        outputDoc.addAnnotation(newT);

        // rel
        BratRelation rel = new BratRelation();
        rel.setId("R" + relId++);
        rel.setType("Interaction-NegPlural");
        rel.putArgument("Arg1", newE1.getId());
        rel.putArgument("Arg2", newE2.getId());
        rel.putArgument("Trigger", newT.getId());
        rel.putArgument("Neg", newE2.getId());
        outputDoc.addAnnotation(rel);
      }
      // to understand etc
      Matcher matcher = toUnderstandPattern.matcher(text);
      if (matcher.find()) {

        Optional<DependencyGraphVertex> testOpt = getVertex(matcher.start(1), matcher.end(1));
        if (!testOpt.isPresent()) {
          logger.warning(
              String.format("cannot find the sentence: %d %d", matcher.start(1), matcher.end(1)));
          continue;
        }

        BratEntity newE1 = new BratEntity(e1);
        newE1.setId("T" + entId++);
        outputDoc.addAnnotation(newE1);

        BratEntity newE2 = new BratEntity(e2);
        newE2.setId("T" + entId++);
        outputDoc.addAnnotation(newE2);

        BratEntity newT = new BratEntity(trigger);
        newT.setId("T" + entId++);
        outputDoc.addAnnotation(newT);

        //neg
        BratEntity neg = convert(testOpt.get(), entId++, "Neg");
        outputDoc.addAnnotation(neg);

        // rel
        BratRelation rel = new BratRelation();
        rel.setId("R" + relId++);
        rel.setType("Interaction-NegTo");
        rel.putArgument("Arg1", newE1.getId());
        rel.putArgument("Arg2", newE2.getId());
        rel.putArgument("Trigger", newT.getId());
        rel.putArgument("Neg", neg.getId());
        outputDoc.addAnnotation(rel);
      }
      matcher = weTestPattern.matcher(text);
      if (matcher.find()) {

        Optional<DependencyGraphVertex> testOpt = getVertex(matcher.start(1), matcher.end(1));
        if (!testOpt.isPresent()) {
          logger.warning(
              String.format("cannot find the sentence: %d %d", matcher.start(1), matcher.end(1)));
          continue;
        }

        BratEntity newE1 = new BratEntity(e1);
        newE1.setId("T" + entId++);
        outputDoc.addAnnotation(newE1);

        BratEntity newE2 = new BratEntity(e2);
        newE2.setId("T" + entId++);
        outputDoc.addAnnotation(newE2);

        BratEntity newT = new BratEntity(trigger);
        newT.setId("T" + entId++);
        outputDoc.addAnnotation(newT);

        //neg
        BratEntity neg = convert(testOpt.get(), entId++, "Neg");
        outputDoc.addAnnotation(neg);

        // rel
        BratRelation rel = new BratRelation();
        rel.setId("R" + relId++);
        rel.setType("Interaction-NegTest");
        rel.putArgument("Arg1", newE1.getId());
        rel.putArgument("Arg2", newE2.getId());
        rel.putArgument("Trigger", newT.getId());
        rel.putArgument("Neg", neg.getId());
        outputDoc.addAnnotation(rel);
      }
    }
  }

  private Optional<DependencyGraphVertex> getVertex(int start, int end) {
    return list.stream().filter(v ->
        v.getBeginPosition() <= start + offset
            && end + offset <= v.getEndPosition())
        .findFirst();
  }

  private boolean isProteinPlural(int argIndex) {
    DependencyGraphVertex arg = list.get(argIndex);
    return arg.getTag() == PartOfSpeech.NNPS || arg.getTag() == PartOfSpeech.NNS;
  }

  Optional<DependencyGraphVertex> getNeg(DependencyGraphVertex trigger) {
    for (DependencyGraphEdge e : graph.outgoingEdgesOf(trigger)) {
      if (e.getDependency().getName().equals(DependencyName.NEG)) {
        return Optional.of(graph.getEdgeTarget(e));
      }
    }
    for (DependencyGraphVertex v : graph.incomingEdgeSourcesOf(trigger)) {
      if (negatives.contains(v.getLemma())) {
        return Optional.of(v);
      }
    }
    if (trigger.getLemma().startsWith("activ")) {
      for (DependencyGraphVertex v : graph.outgoingEdgeTargetsOf(trigger)) {
        if (v.getLemma().startsWith("transcription")) {
          return Optional.of(v);
        }
      }
    }
    return Optional.empty();
  }

  static final Set<String> negatives = Sets.newHashSet(
      "abolish", "cannot", "fail", "not"
  );

  private Optional<DependencyGraphVertex> hasNeg(int argIndex, int triggerIndex) {
    Optional<DependencyGraphVertex> negOpt = getNeg(list.get(triggerIndex));
    if (negOpt.isPresent()) {
      int start = Math.min(argIndex, triggerIndex);
      int end = Math.max(argIndex, triggerIndex);
      int negIndex = list.indexOf(negOpt.get());
      if (start <= negIndex && negIndex <= end) {
        return negOpt;
      }
    }
    int start = Math.min(argIndex, triggerIndex);
    int end = Math.max(argIndex, triggerIndex);
    for (int i = start; i <= end; i++) {
      DependencyGraphVertex v = list.get(i);
      if (negatives.contains(v.getLemma())) {
        return Optional.of(v);
      }
    }
    return Optional.empty();
  }

  BratEntity convert(DependencyGraphVertex v, int index, String type) {
    BratEntity e = new BratEntity();
    e.setId("T" + index);
    e.setType(type);
    e.setText(v.getWord());
    e.addSpan(v.getBeginPosition(), v.getEndPosition());
    return e;
  }

  private int indexOfVertex(BratEntity e) {
    for (int i = 0; i < list.size(); i++) {
      DependencyGraphVertex v = list.get(i);
      if (e.beginPosition() == v.getBeginPosition()
          && e.endPosition() == v.getEndPosition()) {
        return i;
      }
    }
    return -1;
  }

  private Range<Integer> getRange() {
    return Range.closedOpen(
        list.get(0).getBeginPosition(),
        list.get(list.size() - 1).getEndPosition());
  }
}
