package com.pengyifan.biocreativev;

import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.commons.io.BasenameUtils;
import com.pengyifan.kernel.MyProperties;
import com.pengyifan.kernel.allpathgraph.module.ApgConfiguration;
import com.pengyifan.kernel.allpathgraph.module.DefaultApgConfiguration;
import com.pengyifan.nlp.trees.TreebankReader;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.trees.MemoryTreebank;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.Treebank;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.StringJoiner;
import java.util.logging.Logger;
import org.apache.commons.io.FilenameUtils;

public class PPIExtractorNegBatch extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";
    Path root = prop.getTempDir().resolve(dataset);

    PPIExtractorNegBatch s = new PPIExtractorNegBatch(
        new DefaultApgConfiguration(),
        BasenameUtils.getTextBasenames(root.resolve("brat")));

    s.bratInputDir = root.resolve("brat");
    s.parsingInputDir = root.resolve("ptb2");
    s.basicBratDir = root.resolve("ppi-basic");
    s.techBratDir = root.resolve("ppi-tech");
    s.extendedBratDir = root.resolve("ppi-extended");
    s.annotationBratDir = root.resolve("ppi-annotation");
    s.propagateBratDir = root.resolve("ppi-propagate");
    s.outputDir = root.resolve("ppi-neg");
    s.process();
  }


  protected static final Logger logger = Logger.getAnonymousLogger();

  final ApgConfiguration config;

  protected Path bratInputDir;
  protected Path parsingInputDir;
  private Path basicBratDir;
  private Path techBratDir;
  private Path extendedBratDir;
  private Path annotationBratDir;
  private Path propagateBratDir;
  protected Path outputDir;

  String text;
  BratDocument a1Doc;
  BratDocument corefDoc = new BratDocument();
  BratDocument abbrDoc = new BratDocument();
  Treebank treebank;

  BratDocument basicDoc;
  BratDocument extendedDoc;
  BratDocument techDoc;
  BratDocument annotateDoc;
  BratDocument propagateDoc;
  BratDocument ppiDoc;

  protected PPIExtractorNegBatch(ApgConfiguration config, Collection<String> basenames) {
    super(basenames);
    this.config = config;
  }

  private boolean isFileEmtpy(Path file) throws IOException {
    return (Files.newBufferedReader(file).readLine() == null);
  }

  private boolean isFiltered(Path file) throws IOException {
    String basename = FilenameUtils.getBaseName(file.toString());
    return basename.equals("23904265_12301");
  }

  @Override
  protected void preprocess()
      throws IOException {
    if (!Files.exists(outputDir)) {
      Files.createDirectory(outputDir);
    }
  }

  @Override
  protected void preprocessFile(String basename)
      throws IOException {

    Path textFile = bratInputDir.resolve(basename + ".txt");
    text = new String(Files.readAllBytes(textFile));

    Path a1File = bratInputDir.resolve(basename + ".a1");
    if (isFileEmtpy(a1File) || isFiltered(a1File)) {
      treebank = new MemoryTreebank();
      return;
    }

    a1Doc = BratIOUtils.read(Files.newBufferedReader(a1File), basename);

    Path ptbFile = parsingInputDir.resolve(basename + ".ptb");
    if (Files.exists(ptbFile)) {
      TreebankReader treebankReader = new TreebankReader(ptbFile.toFile(),
          config.getTreeBuilder());
      treebank = treebankReader.read();
      treebankReader.close();
    } else {
      treebank = new MemoryTreebank();
    }

    Path basicFile = basicBratDir.resolve(basename + ".ann");
    basicDoc = BratIOUtils.read(Files.newBufferedReader(basicFile), basename);

    Path extendedFile = extendedBratDir.resolve(basename + ".ann");
    extendedDoc = BratIOUtils.read(Files.newBufferedReader(extendedFile), basename);

    Path techFile = techBratDir.resolve(basename + ".ann");
    techDoc = BratIOUtils.read(Files.newBufferedReader(techFile), basename);

    Path annFile = annotationBratDir.resolve(basename + ".ann");
    annotateDoc = BratIOUtils.read(Files.newBufferedReader(annFile), basename);

    Path propFile = propagateBratDir.resolve(basename + ".ann");
    propagateDoc = BratIOUtils.read(Files.newBufferedReader(propFile), basename);
  }

  @Override
  protected void processFile(String basename)
      throws Exception {

    logger.info(String.format("Processing file: %s", basename));

    ppiDoc = new BratDocument();
    ppiDoc.setText(text);

    for (Tree tree : treebank) {
      if (tree.getLeaves().size() >= 100) {
        continue;
      }
      // get sdg
      DependencyGraph sdg = new SDGCreater(config.getGrammaticalStructureBuilder())
          .setDependencyType(config.getDependencyType())
          .setText(text)
          .setTree(tree)
          .create();
      // get ssdg
      DependencyGraph ssdg = new EDGCreater()
          .setText(text)
          .setA1Doc(a1Doc)
          .setCorefDoc(corefDoc)
          .setAbbrDoc(abbrDoc)
          .setDependencyGraph(sdg)
          .create();
      // extract
      String sentext = getText(tree);
      int offset = getOffset(tree);
      PPIExtractorNeg ppi = new PPIExtractorNeg(offset, sentext, ssdg, basicDoc, ppiDoc);
      ppi.extract();
      ppi = new PPIExtractorNeg(offset, sentext, ssdg, extendedDoc, ppiDoc);
      ppi.extract();
      ppi = new PPIExtractorNeg(offset, sentext, ssdg, techDoc, ppiDoc);
      ppi.extract();
      ppi = new PPIExtractorNeg(offset, sentext, ssdg, annotateDoc, ppiDoc);
      ppi.extract();
      ppi = new PPIExtractorNeg(offset, sentext, ssdg, propagateDoc, ppiDoc);
      ppi.extract();
    }
  }

  private int getOffset(Tree t) {
    return ((CoreLabel) t.getLeaves().get(0).label()).beginPosition();
  }

  private String getText(Tree t) {
    StringJoiner sj = new StringJoiner(" ");
    for (Tree leaf : t.getLeaves()) {
      sj.add(((CoreLabel) leaf.label()).word());
    }
    return sj.toString();
  }

  @Override
  protected void postprocessFile(String basename) throws Exception {
    Path ppiFile = outputDir.resolve(basename + ".ann");
    BratIOUtils.write(Files.newBufferedWriter(ppiFile), ppiDoc);
  }
}
