package com.pengyifan.biocreativev;

import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.bionlpst.trigger.SimpleTriggerReader;
import com.pengyifan.bionlpst.trigger.Trigger;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.brat.util.BratMerge;
import com.pengyifan.commons.collections.Pairs;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.commons.io.BasenameUtils;
import com.pengyifan.kernel.MyProperties;
import com.pengyifan.kernel.allpathgraph.module.ApgConfiguration;
import com.pengyifan.kernel.allpathgraph.module.DefaultApgConfiguration;
import com.pengyifan.nlp.process.ExtractLemma;
import com.pengyifan.nlp.trees.TreeUtils;
import com.pengyifan.nlp.trees.TreebankReader;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.trees.MemoryTreebank;
import edu.stanford.nlp.trees.Tree;
import org.apache.commons.io.FilenameUtils;
import org.javatuples.Pair;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.logging.Logger;

public class PPIExtractorPropagateBatch extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";
    Path root = prop.getTempDir().resolve(dataset);

    PPIExtractorPropagateBatch s = new PPIExtractorPropagateBatch(
        new DefaultApgConfiguration(),
        BasenameUtils.getTextBasenames(root.resolve("brat")));

    s.bratInputDir = root.resolve("brat");
    s.parsingInputDir = root.resolve("ptb2");
    s.basicBratDir = root.resolve("ppi-basic");
    s.techBratDir = root.resolve("ppi-tech");
    s.extendedBratDir = root.resolve("ppi-extended");
    s.annotationBratDir = root.resolve("ppi-annotation");

    s.outputDir = root.resolve("ppi-propagate");
    s.triggerFile = prop.getDropboxResourcesDir().resolve("triggers/biocreative-all-triggers.txt");
    s.process();
  }


  protected static final Logger logger = Logger.getAnonymousLogger();
  private static final ExtractLemma lemma = new ExtractLemma();

  final ApgConfiguration config;

  private Path bratInputDir;
  private Path parsingInputDir;
  private Path basicBratDir;
  private Path techBratDir;
  private Path extendedBratDir;
  private Path annotationBratDir;

  protected Path triggerFile;
  protected Path outputDir;

  Collection<Trigger> triggers;
  String text;
  BratDocument a1Doc;
  MemoryTreebank treebank;

  BratDocument basicDoc;
  BratDocument extendedDoc;
  BratDocument techDoc;
  BratDocument annotateDoc;

  String prefix;
  BratDocument allPPIInDocument;

  BratDocument ppiDoc;

  protected PPIExtractorPropagateBatch(ApgConfiguration config, Collection<String> basenames) {
    super(basenames);
    this.config = config;
  }

  private boolean isFileEmtpy(Path file) throws IOException {
    return (Files.newBufferedReader(file).readLine() == null);
  }

  private boolean isFiltered(Path file) throws IOException {
    String basename = FilenameUtils.getBaseName(file.toString());
    return basename.equals("23904265_12301");
  }

  @Override
  protected void preprocess()
      throws IOException {
    triggers = SimpleTriggerReader.readAll(triggerFile);
    if (!Files.exists(outputDir)) {
      Files.createDirectory(outputDir);
    }
  }

  @Override
  protected void preprocessFile(String basename)
      throws IOException {

    Path textFile = bratInputDir.resolve(basename + ".txt");
    text = new String(Files.readAllBytes(textFile));

    Path a1File = bratInputDir.resolve(basename + ".a1");
    if (isFileEmtpy(a1File) || isFiltered(a1File)) {
      treebank = new MemoryTreebank();
      return;
    }

    a1Doc = BratIOUtils.read(Files.newBufferedReader(a1File), basename);

    Path ptbFile = parsingInputDir.resolve(basename + ".ptb");
    if (Files.exists(ptbFile)) {
      TreebankReader treebankReader = new TreebankReader(ptbFile.toFile(),
          config.getTreeBuilder());
      treebank = treebankReader.read();
      treebankReader.close();
      //      checkState(text.isEmpty() || !treebank.isEmpty(), "Treebank is empty");
    } else {
      treebank = new MemoryTreebank();
    }


    Path basicFile = basicBratDir.resolve(basename + ".ann");
    basicDoc = BratIOUtils.read(Files.newBufferedReader(basicFile), basename);

    Path extendedFile = extendedBratDir.resolve(basename + ".ann");
    extendedDoc = BratIOUtils.read(Files.newBufferedReader(extendedFile), basename);

    Path techFile = techBratDir.resolve(basename + ".ann");
    techDoc = BratIOUtils.read(Files.newBufferedReader(techFile), basename);

    Path annFile = annotationBratDir.resolve(basename + ".ann");
    annotateDoc = BratIOUtils.read(Files.newBufferedReader(annFile), basename);

    String thisPrefix = basename.substring(0, basename.indexOf('_'));
    if (thisPrefix.equals(prefix)) {
      return;
    } else {
      prefix = thisPrefix;
      BratFileVisitor visitor = new BratFileVisitor(prefix);
      Files.walkFileTree(basicBratDir, visitor);
      Files.walkFileTree(techBratDir, visitor);
      Files.walkFileTree(extendedBratDir, visitor);
      //    Files.walkFileTree(annotationBratDir, visitor);
      allPPIInDocument = visitor.merge.getDoc();
    }
  }

  @Override
  protected void processFile(String basename)
      throws Exception {

    logger.info(String.format("Processing file: %s", basename));

    ppiDoc = new BratDocument();
    ppiDoc.setText(text);

    // mark PPI
    boolean[] hasPPI = new boolean[treebank.size()];
    for (int i = 0; i < treebank.size(); i++) {
      Tree tree = treebank.get(i);
      Range<Integer> range = getTreeRange(tree);
      if (hasPPI(range.lowerEndpoint(), range.upperEndpoint())) {
        hasPPI[i] = true;
      }
    }

    // extending passage
    int entId = a1Doc.getEntities().size();
    int relId = 0;
    for (int i = 0; i < hasPPI.length; i++) {
      if (!hasPPI[i]) {
        continue;
      }
      if (i < hasPPI.length - 1 && !hasPPI[i + 1]) {
        Pair<Integer, Integer> pair = addIfHasBasicAndProteins(i+1, entId, relId, hasPPI);
        entId = pair.getValue0();
        relId = pair.getValue1();
      }
      if (i < hasPPI.length - 2 && !hasPPI[i + 2]) {
        Pair<Integer, Integer> pair = addIfHasBasicAndProteins(i+2, entId, relId, hasPPI);
        entId = pair.getValue0();
        relId = pair.getValue1();
      }
    }

    // lack of patterns
    for (int i = 0; i < hasPPI.length; i++) {
      if (hasPPI[i]) {
        continue;
      }
      Pair<Integer, Integer> pair = addIfHasBasicAndProteins(i, entId, relId, hasPPI);
      entId = pair.getValue0();
      relId = pair.getValue1();
    }
  }

  private Pair<Integer, Integer> addIfHasBasicAndProteins(int treeIndex, int entId, int relId, boolean[]
      hasPPI) {
    Tree tree = treebank.get(treeIndex);
    Range<Integer> range = getTreeRange(tree);

    // proteins
    List<BratEntity> proteins = Lists.newArrayList();
    for (BratEntity entity : a1Doc.getEntities()) {
      if (range.encloses(entity.totalSpan())) {
        boolean hasSame = proteins.stream()
            .filter(e -> e.getText().equals(entity.getText()))
            .findAny()
            .isPresent();
        if (!hasSame) {
          proteins.add(entity);
        }
      }
    }
    if (proteins.size() < 2) {
      return Pair.with(entId, relId);
    }

    Optional<Tree> triggerLeaf = hasBasicTrigger(tree);
    if (triggerLeaf.isPresent()) {
      int start = ((CoreLabel)triggerLeaf.get().label()).beginPosition();
      int end = ((CoreLabel)triggerLeaf.get().label()).endPosition();
      BratEntity trigger = new BratEntity();
      trigger.setId("T" + entId++);
      trigger.setType("Prop-Trigger");
      trigger.addSpan(start, end);
      trigger.setText(text.substring(start, end));

      for(Pair<BratEntity, BratEntity> pair: Pairs.getPairs(proteins)) {
        BratEntity e1 = pair.getValue0();
        BratEntity e2 = pair.getValue1();
        // PPI there;
        if (hasPPI(e1, e2)) {
          BratRelation rel = new BratRelation();
          rel.setId("R" + relId++);
          rel.setType("Interaction-Prop2");
          rel.putArgument("Arg1", e1.getId());
          rel.putArgument("Arg2", e2.getId());
          rel.putArgument("Trigger", trigger.getId());
          ppiDoc.addAnnotation(rel);
          if (!ppiDoc.containsId(trigger.getId())) {
            ppiDoc.addAnnotation(trigger);
          }
          if (!ppiDoc.containsId(e1.getId())) {
            ppiDoc.addAnnotation(e1);
          }
          if (!ppiDoc.containsId(e2.getId())) {
            ppiDoc.addAnnotation(e2);
          }
          hasPPI[treeIndex] = true;
        }
      }
    }
    return Pair.with(entId, relId);
  }

  private boolean hasPPI(BratEntity e1, BratEntity e2) {
    for(BratRelation relation: allPPIInDocument.getRelations()) {
      BratEntity te1 = allPPIInDocument.getEntity(relation.getArgId("Arg1"));
      BratEntity te2 = allPPIInDocument.getEntity(relation.getArgId("Arg2"));
      if ((te1.getText().equalsIgnoreCase(e1.getText())
          && te2.getText().equalsIgnoreCase(e2.getText()))
        || (te2.getText().equalsIgnoreCase(e1.getText())
          && te1.getText().equalsIgnoreCase(e2.getText()))) {
        return true;
      }
    }
    return false;
  }

  private Optional<Tree> hasBasicTrigger(Tree tree) {
    for (Tree leaf : tree.getLeaves()) {
      String word = ((CoreLabel) leaf.label()).word();
      String tag = ((CoreLabel) leaf.parent(tree).label()).tag();
      String l = lemma.lemmatize(word, tag);
      if (isBasicTrigger(l)) {
        return Optional.of(leaf);
      }
    }
    return Optional.empty();
  }

  private int addIfHasBasic(int treeIndex, int entId, boolean[] hasPPI) {
    Tree tree = treebank.get(treeIndex);
    if (hasBasicTrigger(tree).isPresent()) {
      Range<Integer> range = getTreeRange(tree);
      BratEntity ann = new BratEntity();
      ann.setId("T" + entId++);
      ann.setType("Interaction-Prop1");
      ann.addSpan(range);
      ann.setText(text.substring(range.lowerEndpoint(), range.upperEndpoint()));
      ppiDoc.addAnnotation(ann);
      hasPPI[treeIndex] = true;
    }
    return entId;
  }

  private boolean isBasicTrigger(String lemma) {
    for (Trigger trigger : triggers) {
      if (trigger.getType().equals("Basic") || trigger.getType().equals("PTM")) {
        if (lemma.equalsIgnoreCase(trigger.getLemma().get())) {
          return true;
        }
      }
    }
    return false;
  }

  private Range<Integer> getTreeRange(Tree tree) {
    Tree first = TreeUtils.leftMostLeaf(tree);
    Tree last = TreeUtils.rightMostLeaf(tree);
    int start = ((CoreLabel) first.label()).beginPosition();
    int end = ((CoreLabel) last.label()).endPosition();
    return Range.closedOpen(start, end);
  }

  private boolean hasPPI(int start, int end) {
    boolean b = hasPPI(start, end, basicDoc);
    if (b) {
      return true;
    }
    b = hasPPI(start, end, extendedDoc);
    if (b) {
      return true;
    }
    b = hasPPI(start, end, techDoc);
    if (b) {
      return true;
    }
    b = hasPPI(start, end, annotateDoc);
    if (b) {
      return true;
    }
    return false;
  }

  private boolean hasPPI(int start, int end, BratDocument doc) {
    for (BratRelation relation : doc.getRelations()) {
      BratEntity e1 = doc.getEntity(relation.getArgId("Arg1"));
      BratEntity e2 = doc.getEntity(relation.getArgId("Arg2"));
      BratEntity trigger = doc.getEntity(relation.getArgId("Trigger"));
      RangeSet<Integer> rs = TreeRangeSet.create();
      rs.addAll(e1.getSpans());
      rs.addAll(e2.getSpans());
      rs.addAll(trigger.getSpans());
      Range<Integer> totalSpan = rs.span();
      if (start <= totalSpan.lowerEndpoint() && totalSpan.upperEndpoint() <= end) {
        return true;
      }
    }
    for (BratEntity entity : doc.getEntities()) {
      if (entity.getType().equals("Interaction-Sentence")) {
        Range<Integer> totalSpan = entity.totalSpan();
        if (start <= totalSpan.lowerEndpoint() && totalSpan.upperEndpoint() <= end) {
          return true;
        }
      }
    }
    return false;
  }

  @Override
  protected void postprocessFile(String basename) throws Exception {
    Path ppiFile = outputDir.resolve(basename + ".ann");
    BratIOUtils.write(Files.newBufferedWriter(ppiFile), ppiDoc);
  }

  private static class BratFileVisitor extends SimpleFileVisitor<Path> {
    String prefix;
    BratMerge merge;

    public BratFileVisitor(String prefix) {
      this.prefix = prefix;
      merge = new BratMerge();
    }

    @Override
    public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
      String basename = FilenameUtils.getBaseName(file.toString());
      if (basename.startsWith(prefix)) {
        try {
        BratDocument bratDocument = BratIOUtils.read(Files.newBufferedReader(file), basename);
          merge.addDocument(bratDocument);
        } catch(Exception e) {
          logger.warning(String.format("Cannot parse: %s", file));
          e.printStackTrace();
          System.exit(1);
        }
      }
      return FileVisitResult.CONTINUE;
    }
  }
}
