package com.pengyifan.biocreativev;

import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.Pairs;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.math.RangeUtils;
import edu.emory.mathcs.backport.java.util.Collections;
import org.javatuples.Pair;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

public class PPIExtractorTechnique extends PPIExtractorBasic {

  private final Collection<List<String>> techniques;

  public PPIExtractorTechnique(DependencyGraph graph, BratDocument bratdoc,
      Collection<List<String>> techniques) {
    super(graph, bratdoc, Collections.emptyList());
    this.techniques = techniques;
  }

  public void extract() {
    List<DependencyGraphVertex> proteins = Lists.newArrayList();
    List<DependencyGraphVertex> vertexList = graph.getSortedVertexList();
    for(DependencyGraphVertex v: vertexList) {
      if (v.getNer().isProtein()) {
        proteins.add(v);
      }
    }
    if (proteins.size() < 2) {
      return;
    }

    int relIndex = bratdoc.getRelations().size();
    int annIndex = bratdoc.getEntities().size();

    // has technique word
    LinkedList<DependencyGraphVertex> techVertex = Lists.newLinkedList();
    for(int i=0; i< vertexList.size(); i++) {
      techVertex.clear();
      for(List<String> technique: techniques) {
        if (i + technique.size() > vertexList.size()) {
          continue;
        }
        for(int j=0; j<technique.size(); j++) {
          String senLemma = vertexList.get(i + j).getLemma();
          if (!senLemma.equals(technique.get(j))) {
            break;
          }
          techVertex.add(vertexList.get(i + j));
        }
        if (techVertex.size() == technique.size()) {
          break;
        } else {
          techVertex.clear();
        }
      }
      if (!techVertex.isEmpty()) {
        break;
      }
    }

    if (techVertex.isEmpty()) {
      return;
    }

    for(Pair<DependencyGraphVertex, DependencyGraphVertex> p: Pairs.getPairs(proteins)) {
      if (p.getValue0().getLemma().equals(p.getValue1().getLemma())) {
        continue;
      }

      if (RangeUtils.isOverlapped(
          Range.closedOpen(p.getValue0().getBeginPosition(), p.getValue0().getEndPosition()),
          Range.closedOpen(p.getValue1().getBeginPosition(), p.getValue1().getEndPosition()))) {
        continue;
      }

      BratEntity e1 = find(p.getValue0());
      if (e1 == null) {
        e1 = convert(p.getValue0(), annIndex, "Protein");
        bratdoc.addAnnotation(e1);
        annIndex++;
      }

      BratEntity e2 = find(p.getValue1());
      if (e2 == null) {
        e2 = convert(p.getValue1(), annIndex++, "Protein");
        bratdoc.addAnnotation(e2);
      }

      BratEntity trigger = convert(techVertex, annIndex++, "Technique-Trigger");
      bratdoc.addAnnotation(trigger);

      BratRelation relation = find(e1.getId(), e2.getId());
      if (relation == null) {
        relation = new BratRelation();
        relation.setId("R" + relIndex++);
        relation.setType("Interaction-Technique");
        relation.putArgument("Arg1", e1.getId());
        relation.putArgument("Arg2", e2.getId());
        relation.putArgument("Trigger", trigger.getId());
        bratdoc.addAnnotation(relation);
      }
    }
  }

  private BratEntity convert(LinkedList<DependencyGraphVertex> vs, int index, String type) {
    BratEntity e = new BratEntity();
    e.setId("T" + index);
    e.setType(type);
    e.setText(Joiner.on(" ").join(vs.stream().map(v -> v.getWord()).toArray()));
    e.addSpan(vs.getFirst().getBeginPosition(), vs.getLast().getEndPosition());
    return e;
  }
}
