package com.pengyifan.biocreativev.cmd;

import com.google.common.collect.Iterables;
import com.pengyifan.bioc.BioCAnnotation;
import com.pengyifan.bioc.BioCCollection;
import com.pengyifan.bioc.BioCLocation;
import com.pengyifan.bioc.BioCPassage;
import com.pengyifan.bioc.BioCSentence;
import com.pengyifan.bioc.io.BioCCollectionReader;
import com.pengyifan.bioc.util.BioCPassageIterator;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.commons.convert.BiocToBratConverter;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.kernel.MyProperties;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.io.FileUtils;

import static com.google.common.base.Preconditions.checkArgument;

public class BioC2BratNer extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {
    MyProperties prop = new MyProperties("biocreative5");

    String dataset = "dataset";
    Path root = prop.getDropboxDir().resolve(dataset);

    BioC2BratNer b = new BioC2BratNer(BiocreativeUtils.getTextBasenames(dataset));
    b.textDir = root.resolve("origin/brat");
    b.biocDir = root.resolve("team312/bioc");
    b.bioc2Dir = root.resolve("team312/bioc2");
    b.annDir = root.resolve("team312/brat");
    b.process();
  }

  private static final Logger logger = Logger.getAnonymousLogger();
  private Path textDir;
  private Path biocDir;
  private Path bioc2Dir;
  private Path annDir;

  private int numOfFailed;

  private BioCCollection originCollection;
  private BioCCollection collection;

  public BioC2BratNer(Collection<String> basenames) {
    super(basenames);
  }

  @Override
  protected void preprocess() throws Exception {
    super.preprocess();
    numOfFailed = 0;
    if (!Files.exists(bioc2Dir)) {
      Files.createDirectory(bioc2Dir);
    }
    if (!Files.exists(annDir)) {
      Files.createDirectory(annDir);
    }
  }

  @Override
  protected void postprocess() throws Exception {
    logger.info(String.format("Number of failed conversion: %d", numOfFailed));
  }

  @Override
  protected void preprocessFile(String basename) throws Exception {
    super.preprocessFile(basename);

    logger.info("process " + basename);

    Path biocFile = biocDir.resolve(basename + ".xml");
    // fixed
    if (!Files.exists(biocFile)) {
      return;
    }
    String xmlStr = FileUtils.readFileToString(biocFile.toFile());
    xmlStr = fix(xmlStr);

    Path bioc2File = bioc2Dir.resolve(basename + ".xml");
    FileUtils.write(bioc2File.toFile(), xmlStr);

    BioCCollectionReader biocReader = new BioCCollectionReader(bioc2File.toFile());
    collection = biocReader.readCollection();
    biocReader.close();

    Path originBioCFile = textDir.getParent().resolve("bioc").resolve(basename + ".xml");
    biocReader = new BioCCollectionReader(originBioCFile.toFile());
    originCollection = biocReader.readCollection();
    biocReader.close();
  }

  private static final Pattern p = Pattern.compile("<annotation id=\"(.*?)\">");
  //private static final Pattern p2 = Pattern.compile("(\\d+)(<length>)");

  private String fix(String s) {
    StringBuffer buffer = new StringBuffer();
    int i = 0;
    Matcher m = p.matcher(s);
    while (m.find()) {
      m.appendReplacement(buffer, "<annotation id=\"O" + i++ + "\">");
    }
    m.appendTail(buffer);

    //    buffer = new StringBuffer();
    //    m = p2.matcher(s);
    //    while (m.find()) {
    //      m.appendReplacement(buffer, m.group(1) + "</length>");
    //    }
    //    m.appendTail(buffer);

    return buffer.toString();
  }

  @Override
  protected void postprocessFile(String basename) throws Exception {
    super.postprocessFile(basename);

    BioCPassageIterator passageIterator = new BioCPassageIterator(originCollection);
    while (passageIterator.hasNext()) {
      BioCPassage bioCPassage = passageIterator.next();
      int passageOffset = bioCPassage.getOffset();
      Path annFile = BiocreativeUtils.getFile(annDir, basename, passageOffset, "ann");
      if (!Files.exists(annFile)) {
        annFile.toFile().createNewFile();
      }
    }
  }

  @Override
  protected void processFile(String basename) throws Exception {

    BiocToBratConverter converter = new BiocToBratConverter();

    try {
      int index = 0;
      BioCPassageIterator passageIterator = new BioCPassageIterator(collection);
      while (passageIterator.hasNext()) {
        BioCPassage bioCPassage = passageIterator.next();
        int passageOffset = bioCPassage.getOffset();

        Path textFile = BiocreativeUtils.getFile(textDir, basename, passageOffset, "txt");
        String text = FileUtils.readFileToString(textFile.toFile());

        BratDocument bratDoc = new BratDocument();
        for (BioCAnnotation bioCAnnotation : bioCPassage.getAnnotations()) {
          bioCAnnotation.setID("T" + index++);
          // check
          String expected = bioCAnnotation.getText().get();
          BioCLocation loc = Iterables.getOnlyElement(bioCAnnotation.getLocations());

          int beginPosition = loc.getOffset() - passageOffset;
          int endPosition = loc.getOffset() - passageOffset + loc.getLength();

          try {
            String actual = text.substring(beginPosition, endPosition);
            checkArgument(expected.equals(actual), "expect [%s] is not equal to actual [%s]\n%s",
                expected, actual, bioCAnnotation);
          } catch (Exception e) {
            System.err.println("Annotation: " + bioCAnnotation);
            e.printStackTrace();
            numOfFailed ++;
            continue;
          }
          BratEntity bratEntity = converter.convert(bioCAnnotation);
          BratEntity shiftedBratEntity = BratEntity.shift(bratEntity, -passageOffset);
          bratDoc.addAnnotation(shiftedBratEntity);
        }
        for (BioCSentence bioCSentence : bioCPassage.getSentences()) {
          for (BioCAnnotation bioCAnnotation : bioCSentence.getAnnotations()) {
            bioCAnnotation.setID("T" + index++);
            // check
            String expected = bioCAnnotation.getText().get();
            BioCLocation loc = Iterables.getOnlyElement(bioCAnnotation.getLocations());

            int beginPosition = loc.getOffset() - passageOffset;
            int endPosition = loc.getOffset() - passageOffset + loc.getLength();

            String actual = text.substring(beginPosition, endPosition);
            try {
              checkArgument(expected.equals(actual), "expect [%s] is not equal to actual [%s]\n%s",
                  expected, actual, bioCAnnotation);
            } catch (Exception e) {
              e.printStackTrace();
              numOfFailed ++;
              continue;
            }
            BratEntity bratEntity = converter.convert(bioCAnnotation);
            BratEntity shiftedBratEntity = BratEntity.shift(bratEntity, -passageOffset);
            bratDoc.addAnnotation(shiftedBratEntity);
          }
        }

        Path annFile = BiocreativeUtils.getFile(annDir, basename, passageOffset, "ann");
        BratIOUtils.write(Files.newBufferedWriter(annFile), bratDoc);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}
