package com.pengyifan.biocreativev.cmd;

import com.google.common.collect.Maps;
import com.pengyifan.bioc.BioCAnnotation;
import com.pengyifan.bioc.BioCCollection;
import com.pengyifan.bioc.BioCLocation;
import com.pengyifan.bioc.BioCNode;
import com.pengyifan.bioc.BioCPassage;
import com.pengyifan.bioc.BioCRelation;
import com.pengyifan.bioc.io.BioCCollectionReader;
import com.pengyifan.bioc.io.BioCCollectionWriter;
import com.pengyifan.bioc.util.BioCPassageIterator;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.kernel.MyProperties;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;

public class BioCGenerator extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";

    Path root = prop.getDropboxDir().resolve(dataset);

    BioCGenerator b = new BioCGenerator(BiocreativeUtils.getBioCBasenames(dataset));
    b.biocDir = root.resolve("origin/bioc");
    b.ppiBratDir = prop.getTempDir().resolve(dataset + "/ppi");
    b.outputDir = prop.getTempDir().resolve(dataset + "/v5");
    b.process();
  }

  private static final Logger logger = Logger.getAnonymousLogger();

  private Path biocDir;
  private Path ppiBratDir;
  private Path outputDir;

  private BioCCollection bioCCollection;

  public BioCGenerator(Collection<String> basenames) {
    super(basenames);
  }

  @Override
  protected void preprocess() throws Exception {
    super.preprocess();
    if (!Files.exists(outputDir)) {
      Files.createDirectory(outputDir);
    }
//    if (!Files.exists(outputDir.resolve("no-ppi"))) {
//      Files.createDirectory(outputDir.resolve("no-ppi"));
//    }
  }

  @Override
  protected void preprocessFile(String basename) throws Exception {
    logger.info("process " + basename);
    super.preprocessFile(basename);

    Path biocFile = biocDir.resolve(basename + ".xml");
    BioCCollectionReader bioCCollectionReader = new BioCCollectionReader(
        Files.newBufferedReader(biocFile));
    bioCCollection = bioCCollectionReader.readCollection();
    bioCCollectionReader.close();
  }

  /**
   * <pre>
   * <annotationid="E2">
   * <infonkey="type">PPImention</infon>
   * <locationoffset="208"length="98"/>
   * <text>
   * Here, we report that Aip1p also interacts with the ubiquitous actin depolymerizing factor
   * cofilin.
   * </text>
   * </annotation>
   * </pre>
   */
  @Override
  protected void processFile(String basename) throws Exception {

    int annId = 0;
    int relId = 0;
    BioCPassageIterator passageIterator = new BioCPassageIterator(bioCCollection);
    while (passageIterator.hasNext()) {
      BioCPassage bioCPassage = passageIterator.next();

      bioCPassage.clearAnnotations();
      bioCPassage.clearRelations();

      int offset = bioCPassage.getOffset();

      Path ppiBratFile = BiocreativeUtils.getFile(ppiBratDir, basename, offset, "ann");
      if (!Files.exists(ppiBratFile)) {
        continue;
      }
//      if (bioCPassage.getText().isPresent()
//          && bioCPassage.getText().get().equalsIgnoreCase("references")) {
//        break;
//      }

      Map<String, String> idMap = Maps.newHashMap();

      BratDocument ppiDoc = BratIOUtils.read(Files.newBufferedReader(ppiBratFile), basename);
      for (BratEntity bratEntity : ppiDoc.getEntities()) {
        String biocId = "T" + annId++;
        idMap.put(bratEntity.getId(), biocId);

        BioCAnnotation bioCAnnotation = new BioCAnnotation();
        bioCAnnotation.setID(biocId);
        bioCAnnotation.addLocation(new BioCLocation(
            bratEntity.beginPosition() + offset,
            bratEntity.endPosition() - bratEntity.beginPosition()));
        if (bratEntity.getType().endsWith("Protein")) {
          bioCAnnotation.putInfon("type", bratEntity.getType());
        } else if (bratEntity.getType().endsWith("Trigger")) {
          int dash = bratEntity.getType().indexOf('-');
          bioCAnnotation.putInfon("type", bratEntity.getType().substring(dash + 1));
          bioCAnnotation.putInfon("run", bratEntity.getType().substring(0, dash));
        } else if (bratEntity.getType().startsWith("Interaction")) {
          int dash = bratEntity.getType().indexOf('-');
          bioCAnnotation.putInfon("run", bratEntity.getType().substring(dash + 1));
          bioCAnnotation.putInfon("type", bratEntity.getType().substring(0, dash));
          // score
          String run = bioCAnnotation.getInfon("run").get();
          switch(run) {
          case "Sentence":
            bioCAnnotation.putInfon("score", "0.8");
            break;
          default:
            System.err.println("Cannot recognize run: " + run);
            System.exit(1);
          }
        } else if (bratEntity.getType().startsWith("Neg")) {
          bioCAnnotation.putInfon("type", bratEntity.getType());
        } else {
          logger.warning(String.format("Cannot recognize type :%s", bratEntity.getType()));
          System.exit(1);
        }
        try {
          bioCAnnotation.setText(bioCPassage.getText().get().substring(
              bratEntity.beginPosition(),
              bratEntity.endPosition()
          ));
          bioCPassage.addAnnotation(bioCAnnotation);
        } catch (Exception e) {
          logger.warning(String.format("Cannot match passage text :%s",
              bioCPassage.getText().get()));
        }
      }

      for (BratRelation bratRelation : ppiDoc.getRelations()) {
        BioCRelation bioCRelation = new BioCRelation();
        bioCRelation.setID("R" + relId++);
        if (bratRelation.getType().startsWith("Interaction")) {
          int dash = bratRelation.getType().indexOf('-');
          bioCRelation.putInfon("run", bratRelation.getType().substring(dash + 1));
          bioCRelation.putInfon("type", bratRelation.getType().substring(0, dash));
        } else {
          logger.warning(String.format("Cannot recognize type :%s", bratRelation.getType()));
          System.exit(1);
        }
        for (Entry<String, String> pair : bratRelation.getArguments().entrySet()) {
          bioCRelation.addNode(new BioCNode(idMap.get(pair.getValue()), pair.getKey()));
        }
        // score
        String run = bioCRelation.getInfon("run").get();
        switch(run) {
        case "Basic":
        case "Technique":
          bioCRelation.putInfon("score", "1.0");
          break;
        case "Prop2":
          bioCRelation.putInfon("score", "0.6");
          break;
        case "NegPlural":
        case "NegTo":
        case "NegTest":
        case "NegNot":
          bioCRelation.putInfon("score", "-1");
          break;
        case "Extended":
          bioCRelation.putInfon("score", "0.9");
          break;
        default:
          System.err.println("Cannot recognize type: " + run);
          System.exit(1);
        }
        bioCPassage.addRelation(bioCRelation);
      }
    }

    Path ppiBioCFile = null;
    if (annId > 0) {
      ppiBioCFile = outputDir.resolve(basename + ".xml");
    } else {
      ppiBioCFile = outputDir.resolve("no-ppi").resolve(basename + ".xml");
    }
    BioCCollectionWriter writer = new BioCCollectionWriter(Files.newBufferedWriter(ppiBioCFile));
    writer.writeCollection(bioCCollection);
    writer.close();
  }

//  private boolean addIfNotExsitAnnotation(BioCAnnotation target, BioCPassage bioCPassage) {
//    for (BioCAnnotation ann : bioCPassage.getAnnotations()) {
//      if (ann.getTotalLocation().equals(target.getTotalLocation())) {
//        // ann.putInfon("run", target.getInfon("run").get());
//        return ann;
//      }
//    }
//    bioCPassage.addAnnotation(target);
//    return target;
//  }
}
