package com.pengyifan.biocreativev.cmd;

import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.bioc.BioCAnnotation;
import com.pengyifan.bioc.BioCCollection;
import com.pengyifan.bioc.BioCDocument;
import com.pengyifan.bioc.BioCLocation;
import com.pengyifan.bioc.BioCNode;
import com.pengyifan.bioc.BioCPassage;
import com.pengyifan.bioc.BioCRelation;
import com.pengyifan.bioc.BioCSentence;
import com.pengyifan.bioc.io.BioCCollectionReader;
import com.pengyifan.bioc.io.BioCCollectionWriter;
import com.pengyifan.bioc.util.BioCPassageIterator;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.kernel.MyProperties;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.logging.Logger;
import org.apache.commons.math3.stat.StatUtils;

import static com.google.common.base.Preconditions.checkArgument;

public class BioCGeneratorFinal extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";

    Path root = prop.getDropboxDir().resolve(dataset);

    BioCGeneratorFinal b = new BioCGeneratorFinal(BiocreativeUtils.getBioCBasenames(dataset));
    b.biocDir = root.resolve("origin/bioc");
    b.jnlpDir = root.resolve("jnlp/bioc2");
    b.ppiBioCDir = prop.getTempDir().resolve(dataset + "/v5");
    b.outputDir = prop.getTempDir().resolve(dataset + "/run_score");
    b.process();
  }

  private static final Logger logger = Logger.getAnonymousLogger();

  private Path biocDir;
  private Path jnlpDir;
  private Path ppiBioCDir;
  private Path outputDir;

  private BioCCollection originCollection;
  private BioCCollection ppiCollection;
  private BioCCollection jnlpCollection;

  public BioCGeneratorFinal(Collection<String> basenames) {
    super(basenames);
  }

  @Override
  protected void preprocess() throws Exception {
    super.preprocess();
    if (!Files.exists(outputDir)) {
      Files.createDirectory(outputDir);
    }
    if (!Files.exists(outputDir.resolve("no-ppi"))) {
      Files.createDirectory(outputDir.resolve("no-ppi"));
    }
  }

  @Override
  protected void preprocessFile(String basename) throws Exception {
    logger.info("process " + basename);
    super.preprocessFile(basename);

    Path biocFile = biocDir.resolve(basename + ".xml");
    BioCCollectionReader bioCCollectionReader = new BioCCollectionReader(
        Files.newBufferedReader(biocFile));
    originCollection = bioCCollectionReader.readCollection();
    bioCCollectionReader.close();

    Path ppiBioCFile = ppiBioCDir.resolve(basename + ".xml");
    bioCCollectionReader = new BioCCollectionReader(Files.newBufferedReader(ppiBioCFile));
    ppiCollection = bioCCollectionReader.readCollection();
    bioCCollectionReader.close();

    Path jnlpFile = jnlpDir.resolve(basename + ".xml");
    bioCCollectionReader = new BioCCollectionReader(
        Files.newBufferedReader(jnlpFile));
    jnlpCollection = bioCCollectionReader.readCollection();
    bioCCollectionReader.close();
  }

  /**
   * <pre>
   * <annotationid="E2">
   * <infonkey="type">PPImention</infon>
   * <locationoffset="208"length="98"/>
   * <text>
   * Here, we report that Aip1p also interacts with the ubiquitous actin depolymerizing factor
   * cofilin.
   * </text>
   * </annotation>
   * </pre>
   */
  @Override
  protected void processFile(String basename) throws Exception {

    int annId = 0;

    checkArgument(ppiCollection.getDocmentCount() == 1,
        "ppi collection contains %s document", ppiCollection.getDocmentCount());
    checkArgument(originCollection.getDocmentCount() == 1,
        "origin collection contains %s document", originCollection.getDocmentCount());

    BioCDocument ppiDocument = ppiCollection.getDocument(0);
    BioCDocument originDocument = originCollection.getDocument(0);
    checkArgument(ppiDocument.getPassageCount() == originDocument.getPassageCount(),
        "different passage counts: ppi[%s] != origin[%s]",
        ppiDocument.getPassageCount(),
        originDocument.getPassageCount());

    for (int i = 0; i < ppiDocument.getPassageCount(); i++) {
      BioCPassage ppiPassage = ppiDocument.getPassage(i);
      BioCPassage originPassage = originDocument.getPassage(i);
      BioCPassage sentencePassage = getSentencePassage(originPassage);

      originPassage.clearAnnotations();
      originPassage.clearRelations();

      if (sentencePassage == null) {
        logger.severe(String.format("Cannot find sentence passage: %d", originPassage.getOffset()));
      }

      List<BioCAnnotation> ppiMentions = Lists.newArrayList();
      for (BioCAnnotation annotation : ppiPassage.getAnnotations()) {
        if (annotation.getInfon("type").get().equals("Interaction")) {
          BioCAnnotation ann = new BioCAnnotation(annotation);
          BioCLocation loc = ann.getTotalLocation();
          String expected = ann.getText().get();
          String actual = originPassage.getText().get().substring(
              loc.getOffset() - originPassage.getOffset(),
              loc.getOffset() - originPassage.getOffset() + loc.getLength()
          );
          checkArgument(expected.equals(actual), "expect [%s] is not equal to actual [%s]\n%s",
              expected, actual, ann);
          ann.putInfon("type", "PPImention");
          ppiMentions.add(ann);
        }
      }

      // remove neg
      List<BioCRelation> relations = Lists.newArrayList(ppiPassage.getRelations());
      boolean[] isNeg = new boolean[relations.size()];
      for (int j = 0; j < relations.size(); j++) {
        BioCRelation relation = relations.get(j);
        if (relation.getInfon("type").get().equals("Interaction")
            && relation.getInfon("run").get().startsWith("Neg")
            && !relation.getInfon("run").get().endsWith("Plural")) {
          isNeg[j] = true;
          for (int k = 0; k < relations.size(); k++) {
            BioCRelation r = relations.get(k);
            if (equal(r, relation)) {
              isNeg[k] = true;
            }
          }
        }
      }

      for (int j = 0; j < relations.size(); j++) {
        if (isNeg[j]) {
          continue;
        }
        BioCRelation relation = relations.get(j);
        if (relation.getInfon("type").get().equals("Interaction")) {
          RangeSet<Integer> rs = TreeRangeSet.create();
          for (BioCNode node : relation.getNodes()) {
            Optional<BioCAnnotation> opt = ppiPassage.getAnnotation(node.getRefid());
            if (!opt.isPresent()) {
              logger.warning("Cannot find annotation: " + node.getRefid());
              System.exit(1);
            }
            BioCLocation loc = opt.get().getTotalLocation();
            rs.add(Range.closedOpen(loc.getOffset(), loc.getOffset() + loc.getLength()));
          }

          Range<Integer> totalSpan = rs.span();
          int start = totalSpan.lowerEndpoint();
          int end = totalSpan.upperEndpoint();

          // find sentence
          int sentenceIndex = indexOfSentence(start, end, sentencePassage);
          if (sentenceIndex != -1) {
            BioCSentence s = sentencePassage.getSentence(sentenceIndex);
            start = s.getOffset();
            end = s.getOffset() + s.getText().get().length();
          } else {
            logger.severe(String.format("Cannot find sentence: %d %d", start, end));
          }

          BioCAnnotation ann = new BioCAnnotation();
          ann.addLocation(new BioCLocation(start, end - start));
          ann.putInfon("type", "PPImention");
          ann.putInfon("score", relation.getInfon("score").get());
          try {
            ann.setText(originPassage.getText().get().substring(
                start - originPassage.getOffset(),
                end - originPassage.getOffset()));
          } catch (Exception e) {
            logger.warning(String.format("Cannot match passage text: %s",
                originPassage.getText().get()));
          }

          ppiMentions.add(ann);
        }
      }

      if (ppiMentions.isEmpty()) {
        continue;
      }

      boolean[] hasPPIs = new boolean[sentencePassage.getSentenceCount()];
      double[] scores = new double[sentencePassage.getSentenceCount()];
      for (BioCAnnotation ppiMention : ppiMentions) {
        BioCLocation loc = ppiMention.getTotalLocation();
        int sentenceIndex = indexOfSentence(
            loc.getOffset(), loc.getOffset() + loc.getLength(),
            sentencePassage);
        hasPPIs[sentenceIndex] = true;
        scores[sentenceIndex] = Double.parseDouble(ppiMention.getInfon("score").get());
      }

      //      // remove neg
      //      for (BioCAnnotation ppiMention : negMentions) {
      //        BioCLocation loc = ppiMention.getTotalLocation();
      //        int sentenceIndex = indexOfSentence(
      //            loc.getOffset(), loc.getOffset() + loc.getLength(),
      //            sentencePassage);
      //        hasPPIs[sentenceIndex] = false;
      //      }

      // merge
      for (int j = 0; j < hasPPIs.length; j++) {
        if (!hasPPIs[j]) {
          continue;
        }
        BioCSentence start = sentencePassage.getSentence(j);
        int k = j + 1;
        while (k < hasPPIs.length) {
          if (!hasPPIs[k]) {
            break;
          }
          k++;
        }
        BioCSentence end = sentencePassage.getSentence(k - 1);
        double score = StatUtils.max(scores, j, k - j);
        j = k;

        BioCAnnotation ann = new BioCAnnotation();
        ann.setID("TEAM_336_PPI" + annId++);
        ann.addLocation(new BioCLocation(
            start.getOffset(),
            end.getOffset() + end.getText().get().length() - start.getOffset()));
        ann.putInfon("type", "PPImention");
        ann.putInfon("score", Double.toString(score));
        try {
          BioCLocation loc = ann.getTotalLocation();
          ann.setText(originPassage.getText().get().substring(
              loc.getOffset() - originPassage.getOffset(),
              loc.getOffset() + loc.getLength() - originPassage.getOffset()));
        } catch (Exception e) {
          logger.warning(String.format("Cannot match passage text: %s",
              originPassage.getText().get()));
        }
        originPassage.addAnnotation(ann);
      }
    }

    Path ppiBioCFile = null;
    if (annId > 0) {
      ppiBioCFile = outputDir.resolve(basename + ".xml");
    } else {
      ppiBioCFile = outputDir.resolve("no-ppi").resolve(basename + ".xml");
    }
    BioCCollectionWriter writer = new BioCCollectionWriter(Files.newBufferedWriter(ppiBioCFile));
    writer.writeCollection(originCollection);
    writer.close();
  }

  private boolean equal(BioCRelation r1, BioCRelation r2) {
    BioCNode r1Arg1 = null;
    BioCNode r1Arg2 = null;
    for (BioCNode n : r1.getNodes()) {
      if (n.getRole().equals("Arg1")) {
        r1Arg1 = n;
      }
      if (n.getRole().equals("Arg2")) {
        r1Arg2 = n;
      }
    }
    BioCNode r2Arg1 = null;
    BioCNode r2Arg2 = null;
    for (BioCNode n : r2.getNodes()) {
      if (n.getRole().equals("Arg1")) {
        r2Arg1 = n;
      }
      if (n.getRole().equals("Arg2")) {
        r2Arg2 = n;
      }
    }
    return r1Arg1.getRefid().equals(r2Arg1.getRefid())
        && r1Arg2.getRefid().equals(r2Arg2.getRefid());
  }

  private BioCPassage getSentencePassage(BioCPassage originPassage) {
    BioCPassageIterator passageIterator = new BioCPassageIterator(jnlpCollection);
    while (passageIterator.hasNext()) {
      BioCPassage p = passageIterator.next();
      if (originPassage.getOffset() == p.getOffset()) {
        return p;
      }
    }
    return null;
  }

  private int indexOfSentence(int start, int end, BioCPassage sentencePassage) {
    for (int i = 0; i < sentencePassage.getSentenceCount(); i++) {
      BioCSentence bioCSentence = sentencePassage.getSentence(i);
      int sStart = bioCSentence.getOffset();
      int sEnd = bioCSentence.getOffset() + bioCSentence.getText().get().length();
      if (sStart <= start && end <= sEnd) {
        return i;
      }
    }
    return -1;
  }
}
