package com.pengyifan.biocreativev.cmd;

import com.google.common.collect.Iterables;
import com.pengyifan.bioc.BioCAnnotation;
import com.pengyifan.bioc.BioCCollection;
import com.pengyifan.bioc.BioCLocation;
import com.pengyifan.bioc.BioCPassage;
import com.pengyifan.bioc.BioCSentence;
import com.pengyifan.bioc.io.BioCCollectionReader;
import com.pengyifan.bioc.util.BioCPassageIterator;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.kernel.MyProperties;
import com.pengyifan.nlp.trees.PtbWriter;
import com.pengyifan.nlp.trees.TreeUtils;
import edu.stanford.nlp.ling.CoreLabel;
import edu.stanford.nlp.trees.Tree;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import javax.xml.stream.XMLStreamException;
import org.apache.commons.io.FileUtils;

import static com.google.common.base.Preconditions.checkArgument;

public class Jnlp2Ptb extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";

    Path root = prop.getDropboxDir().resolve(dataset);

    Jnlp2Ptb b = new Jnlp2Ptb(BiocreativeUtils.getTextBasenames(dataset));
    b.jnlpDir = root.resolve("jnlp/bioc");
    b.jnlp2Dir = root.resolve("jnlp/bioc2");
    b.textDir = root.resolve("origin/brat");
    b.sentDir = root.resolve("jnlp/brat");
    //    b.ptbDir = root.resolve("jnlp/ptb");
    b.process();
  }

  private static final Logger logger = Logger.getAnonymousLogger();
  private Path textDir;
  private Path jnlpDir;
  private Path jnlp2Dir;
  private Path sentDir;

  public Jnlp2Ptb(Collection<String> basenames) {
    super(basenames);
  }

  @Override
  protected void preprocess() throws Exception {
    super.preprocess();
    if (!Files.exists(jnlp2Dir)) {
      Files.createDirectory(jnlp2Dir);
    }
    if (!Files.exists(sentDir)) {
      Files.createDirectory(sentDir);
    }
  }

  @Override
  protected void processFile(String basename) throws Exception {
    logger.info("process " + basename);

    Path jnlpFile = jnlpDir.resolve(basename + "_jnlp.xml");
    Path jnlp2File = jnlp2Dir.resolve(basename + ".xml");

    String xmlStr = FileUtils.readFileToString(jnlpFile.toFile());
    xmlStr = fix(xmlStr);
    FileUtils.write(jnlp2File.toFile(), xmlStr);

    convert(basename, jnlp2File);
  }

  private static final Pattern p = Pattern.compile("<relation>");

  private String fix(String s) {
    // info
    s = s.replace("<source></source>", "<source>PMC</source>");
    s = s.replace("<date></date>", "<date>20140719</date>");
    s = s.replace("<key></key>", "<key>ascii.key</key>");
    s = s.replace("<source/>", "<source>PMC</source>");
    s = s.replace("<date/>", "<date>20140719</date>");
    s = s.replace("<key/>", "<key>ascii.key</key>");

    StringBuffer buffer = new StringBuffer();
    int relationId = 0;
    Matcher m = p.matcher(s);
    while (m.find()) {
      m.appendReplacement(buffer, "<relation id=\"" + relationId++ + "\">");
    }
    m.appendTail(buffer);
    return buffer.toString();
  }

  public void convert(String basename, Path biocFile) throws IOException,
      XMLStreamException {

    BioCCollectionReader biocReader = new BioCCollectionReader(biocFile.toFile());
    BioCCollection collection = biocReader.readCollection();
    biocReader.close();

    BioCPassageIterator passageIterator = new BioCPassageIterator(collection);
    while (passageIterator.hasNext()) {
      BioCPassage bioCPassage = passageIterator.next();

      int passageOffset = bioCPassage.getOffset();
      Path textFile = BiocreativeUtils.getFile(textDir, basename, passageOffset, "txt");
      String text = FileUtils.readFileToString(textFile.toFile());

      StringBuilder ptbSb = new StringBuilder();
      StringBuilder senSb = new StringBuilder();
      for (BioCSentence bioCSentence : bioCPassage.getSentences()) {
        // ptb
        Tree tree = getTree(bioCSentence, text, passageOffset);
        ptbSb.append(PtbWriter.print(tree)).append('\n');
        senSb.append(bioCSentence.getText().get()).append('\n');
      }

      //      Path ptbFile = BiocreativeUtils.getFile(ptbDir, basename, passageOffset, "ptb");
      //      FileUtils.write(ptbFile.toFile(), ptbSb);

      Path sentFile = BiocreativeUtils.getFile(sentDir, basename, passageOffset, "txt");
      FileUtils.write(sentFile.toFile(), senSb);
    }
  }

  private Tree getTree(BioCSentence bioCSentence, String text, int passageOffset) {
    String ptbStr = bioCSentence.getInfon("PennTree").get();
    Tree tree = Tree.valueOf(ptbStr);
    // offset
    List<Tree> leaves = tree.getLeaves();
    List<BioCAnnotation> annotatioins = bioCSentence.getAnnotations().stream()
        .filter(ann -> ann.getInfon("type").get().equals("token"))
        .sorted((ann1, ann2) -> {
          BioCLocation loc1 = Iterables.getOnlyElement(ann1.getLocations());
          BioCLocation loc2 = Iterables.getOnlyElement(ann2.getLocations());
          return Integer.compare(loc1.getOffset(), loc2.getOffset());
        })
        .collect(Collectors.toList());
    checkArgument(leaves.size() == annotatioins.size(), "different size");

    for (int i = 0; i < leaves.size(); i++) {
      CoreLabel label = (CoreLabel) leaves.get(i).label();
      BioCLocation loc = Iterables.getOnlyElement(annotatioins.get(i).getLocations());

      // check offset
      String actual = TreeUtils.adaptValue(label.word());
      int beginPosition = loc.getOffset() - passageOffset;
      int endPosition = loc.getOffset() - passageOffset + loc.getLength();
      String expected = text.substring(beginPosition, endPosition);
      if ((expected.equals("[") || expected.equals("<")) && actual.equals("(")) {
        ;
      } else if ((expected.equals("]") || expected.equals(">")) && actual.equals(")")) {
        ;
      } else if (actual.equals(expected + ".")) {
        label.setBeginPosition(beginPosition);
        label.setEndPosition(endPosition + 1);
        continue;
      } else {
        try {
          checkArgument(actual.equalsIgnoreCase(expected), "Not equal: expected[%s] actual[%s]\n%s",
              expected, actual, bioCSentence);
        } catch (Exception e) {
          e.printStackTrace();
          return tree;
        }
      }

      label.setBeginPosition(beginPosition);
      label.setEndPosition(endPosition);
    }
    return tree;
  }
}
