package com.pengyifan.biocreativev.cmd;

import com.google.common.base.Joiner;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.pengyifan.bioc.*;
import com.pengyifan.bioc.io.BioCCollectionReader;
import com.pengyifan.bioc.util.BioCPassageIterator;
import com.pengyifan.commons.io.AbstractBatchProcessor;
import com.pengyifan.experimental.PPIEval.PPI;
import com.pengyifan.kernel.MyProperties;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.WordUtils;

import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.logging.Logger;

import static com.google.common.base.Preconditions.checkNotNull;

public class PPICounter extends AbstractBatchProcessor {

  public static void main(String[] args)
      throws Exception {

    MyProperties prop = new MyProperties("biocreative5");
    String dataset = "dataset";

    Path root = prop.getDropboxDir().resolve(dataset);

    PPICounter b = new PPICounter(BiocreativeUtils.getBioCBasenames(dataset));
    b.biocDir = prop.getTempDir().resolve(dataset + "/v5");
    b.process();
  }

  private static final Logger logger = Logger.getAnonymousLogger();

  private Path biocDir;
  private BioCCollection bioCCollection;
  private Set<PPI> ppis;

  public PPICounter(Collection<String> basenames) {
    super(basenames);
  }

  @Override
  protected void postprocess() throws Exception {
    List<PPI> list = Lists.newArrayList(ppis);
    Collections.sort(list, (p1, p2) -> {
      int c = p1.docId.compareTo(p2.docId);
      if (c == 0) {
        c = p1.arg1.compareTo(p2.arg1);
      }
      if (c == 0) {
        c = p1.arg2.compareTo(p2.arg2);
      }
      return c;
    });

    StringBuilder sb = new StringBuilder();
    for (PPI ppi: list) {
      sb.append(Joiner.on('\t').join(ppi.docId, ppi.arg1, ppi.arg2)).append('\n');
    }
    FileUtils.write(Paths.get("tmp/biocreative5/dataset/pred.csv").toFile(), sb.toString());
  }

  @Override
  protected void preprocess() throws Exception {
    super.preprocess();
    ppis = Sets.newHashSet();
  }

  @Override
  protected void preprocessFile(String basename) throws Exception {
    logger.info("process " + basename);
    super.preprocessFile(basename);

    Path biocFile = biocDir.resolve(basename + ".xml");
    BioCCollectionReader bioCCollectionReader = new BioCCollectionReader(
        Files.newBufferedReader(biocFile));
    bioCCollection = bioCCollectionReader.readCollection();
    bioCCollectionReader.close();
  }

  @Override
  protected void processFile(String basename) throws Exception {
    /**
     <node refid="T1" role="Arg2"/>
     <node refid="T2" role="Trigger"/>
     <node refid="T4" role="Arg1"/>
     */
    BioCPassageIterator passageIterator = new BioCPassageIterator(bioCCollection);
    while (passageIterator.hasNext()) {
      BioCPassage bioCPassage = passageIterator.next();
      for (BioCRelation bioCRelation : bioCPassage.getRelations()) {
        if (bioCRelation.getInfon("type").get().equals("Interaction")) {
          String arg1Id = null;
          String arg2Id = null;
          for(BioCNode node: bioCRelation.getNodes()) {
            if (node.getRole().equals("Arg1")) {
              arg1Id = node.getRefid();
            } else if (node.getRole().equals("Arg2")) {
              arg2Id = node.getRefid();
            }
          }
          checkNotNull(arg1Id);
          checkNotNull(arg2Id);
          BioCAnnotation arg1 = getAnnotation(bioCPassage, arg1Id);
          BioCAnnotation arg2 = getAnnotation(bioCPassage, arg2Id);
          checkNotNull(arg1);
          checkNotNull(arg2);
          PPI ppi = new PPI(basename, arg1.getText().get(), arg2.getText().get());
          ppis.add(ppi);
        }
      }
    }
  }

  private BioCAnnotation getAnnotation(BioCPassage passage, String annId) {
    for(BioCAnnotation annotation: passage.getAnnotations()) {
      if (annotation.getID().equals(annId)) {
        return annotation;
      }
    }
    return null;
  }
}
