package com.pengyifan.bionlpst.trigger;

import java.io.Closeable;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVParser;
import org.apache.commons.csv.CSVRecord;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;

public class TriggerReader implements Closeable {

  private static final String TYPE_HEADER = "type";
  private static final String WORD_HEADER = "word";
  private static final String LEMMA_HEADER = "lemma";
  private static final String TAG_HEADER = "tag";
  private static final String FREQ_HEADER = "frequency";
  private static final String MARK_HEADER = "mark";
  private static final String PATTERNS_HEADER = "patterns";

  private static final CSVFormat FORMAT = CSVFormat.DEFAULT
      .withDelimiter('\t')
      .withCommentMarker('#')
      .withIgnoreEmptyLines(true)
      .withAllowMissingColumnNames(true)
      .withHeader(
          TYPE_HEADER,
          WORD_HEADER,
          LEMMA_HEADER,
          TAG_HEADER,
          FREQ_HEADER,
          MARK_HEADER,
          PATTERNS_HEADER);

  private final CSVParser parser;

  public TriggerReader(Path file)
      throws IOException {
    parser = FORMAT.parse(Files.newBufferedReader(file));
  }

  @Override
  public void close()
      throws IOException {
    parser.close();
  }

  public List<Trigger> readAll()
      throws IOException {
    List<Trigger> triggers = Lists.newArrayList();
    for (CSVRecord csvRecord : parser) {
      triggers.add(createTrigger(csvRecord));
    }
    return ImmutableList.copyOf(triggers);
  }

  private Trigger createTrigger(CSVRecord csvRecord) {
    Trigger trigger = new Trigger();
    readBasic(trigger, csvRecord);
    readPatterns(trigger, csvRecord);
    return trigger;
  }

  private void readBasic(Trigger trigger, CSVRecord csvRecord) {
    trigger
        .setType(csvRecord.get(TYPE_HEADER))
        .setWord(csvRecord.get(WORD_HEADER))
        .setLemma(csvRecord.get(LEMMA_HEADER))
        .setTag(csvRecord.get(TAG_HEADER))
        .setFrequency(Integer.parseInt(csvRecord.get(FREQ_HEADER)))
        .setMark(csvRecord.get(MARK_HEADER));
  }

  private void readPatterns(Trigger trigger, CSVRecord csvRecord) {
    if (csvRecord.isConsistent()) {
      for (String pattern : csvRecord.get(PATTERNS_HEADER).split(",")) {
        if (!pattern.isEmpty()) {
          trigger.addPattern(pattern);
        }
      }
    }
  }

}
