package com.pengyifan.bionlpst.utils;

import com.google.common.base.Joiner;
import com.google.common.collect.Range;
import org.apache.commons.io.FilenameUtils;

import javax.swing.text.BadLocationException;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.HTMLWriter;
import javax.swing.text.html.StyleSheet;
import java.io.*;
import java.util.List;

public class AnalysisHtmlPrinter implements Closeable {

  private final StringBuilder totalHtmlText;
  private final File outputFile;

  public AnalysisHtmlPrinter(File outputFile) {
    this.totalHtmlText = new StringBuilder();
    this.outputFile = outputFile;
  }

  public void writeInstance(
      String title,
      String text,
      List<EntityRange> annotations) {
    // title
    totalHtmlText.append(String.format("<h2 id=\"%s\">%s</h2>", title, title));
    // add text
    writeInstance(text, annotations);
  }

  public void writeInstance(String text, List<EntityRange> annotations) {
    // add text
    totalHtmlText.append("<p>");
    for (int i = 0; i < text.length(); i++) {
      appendStartTags(totalHtmlText, i, annotations);
      appendChar(totalHtmlText, text.charAt(i));
      appendEndTags(totalHtmlText, i, annotations);
    }
    totalHtmlText.append("</p>");
  }

  @Override
  public void close()
      throws IOException {
    insertTitle(FilenameUtils.getBaseName(outputFile.getAbsolutePath()));
    HTMLDocument htmlDoc = createHTMLDocument();
    modifyStyleSheet(htmlDoc);
    writeToFile(outputFile, htmlDoc);
  }

  private void appendEndTags(StringBuilder sb, int offset, List<EntityRange> annotations) {
    for (int i = annotations.size() - 1; i >= 0; i--) {
      EntityRange p = annotations.get(i);
      if (p.range.upperEndpoint() == offset + 1) {
        sb.append(String.format("<sub>%s</sub></span>",
            Joiner.on(':').skipNulls().join(p.tag, null)));//, p.range)));
      }
    }
  }

  private void appendStartTags(StringBuilder sb, int offset, List<EntityRange> annotations) {
    for (int i = 0; i < annotations.size(); i++) {
      EntityRange p = annotations.get(i);
      if (p.range.lowerEndpoint() == offset) {
        sb.append(String.format("<span class=\"%s\">", p.selector==null ? "T" : p.selector));
      }
    }
  }

  private void appendChar(StringBuilder sb, char c) {
    if (c == '\n') {
      sb.append("</p><p>");
    } else {
      sb.append(c);
    }
  }

  private void insertTitle(String title) {
    totalHtmlText.insert(
        0,
        String.format("<head><title>%s</title></head>", title));
  }

  private HTMLDocument createHTMLDocument()
      throws IOException {
    HTMLEditorKit kit = new HTMLEditorKit();
    HTMLDocument htmlDoc = (HTMLDocument) kit.createDefaultDocument();
    try {
      kit.read(new StringReader(totalHtmlText.toString()), htmlDoc, 0);
    } catch (BadLocationException e) {
      throw new IOException(e);
    }
    return htmlDoc;
  }

  private void writeToFile(File file, HTMLDocument htmlDoc)
      throws IOException {
    FileWriter out = new FileWriter(file);
    try {
      new HTMLWriter(out, htmlDoc).write();
    } catch (BadLocationException e) {
      throw new IOException(e);
    } finally {
      out.flush();
      out.close();
    }
  }

  private void modifyStyleSheet(HTMLDocument htmlDoc) {
    StyleSheet styleSheet = htmlDoc.getStyleSheet();
    styleSheet.addRule("p {margin-top:5px;margin-bottom:5px;text-indent:25px}");
    styleSheet.addRule("h1 {font-size:150%;font-weight:bold;}");
    styleSheet.addRule("h2 {font-size:100%;font-weight:bold;}");
    styleSheet.addRule(".T {color:blue; background-color:yellow; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".A {color:green; background-color:yellow; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".C {background-color:yellow; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".D {color:brown; background-color:yellow; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".E {color:#800080; background-color:yellow; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".F {background-color:#00FFFF; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".G {background-color:#E0FFFF; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".Z {background-color:red; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".A1A2 {color:#DF01D7; font-weight:bold;font-size:100%}");
    styleSheet.addRule(".eventid {font-weight:bold;}");
    styleSheet.addRule(".title {margin-top:10px;font-weight:bold; font-size:150%}");
    styleSheet.addRule("sub {font-size:70%}");
  }

  public static class EntityRange {
    public final Range<Integer> range;
    public final String selector;
    public final String tag;

    public EntityRange(Range<Integer> range, String tag, String selector) {
      this.range = range;
      this.selector = selector;
      this.tag = tag;
    }
  }
}
