package com.pengyifan.commons.collections.dependencygraph;

import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.util.StringJoiner;
import java.util.function.Predicate;

import javax.annotation.Nonnull;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.pengyifan.commons.collections.ImmutableCollectors;
import com.pengyifan.commons.collections.indexgraph.IndexGraph;
import org.apache.commons.lang3.builder.ToStringBuilder;

public class DependencyGraph
    extends IndexGraph<DependencyGraphVertex, DependencyGraphEdge> {

  public DependencyGraph() {
    super(DependencyGraphEdge.class);
  }

  public DependencyGraph(DependencyGraph graph) {
    this();
    checkNotNull(graph, "Graph should not be null.");
    // nodes
    graph.vertexSet().stream()
        .forEach(v -> addVertex(new DependencyGraphVertex(
            v.getIndex(),
            v.getWord(),
            v.getTag(),
            v.getLemma(),
            v.getNer(),
            v.getPosition(),
            v.getBeginPosition(),
            v.getEndPosition())));

    // edges
    graph.edgeSet().stream()
        .forEach(e -> addEdge(
            getVertex(graph.getEdgeSource(e).getIndex()),
            getVertex(graph.getEdgeTarget(e).getIndex()),
            new DependencyGraphEdge(e.getIndex(), e)));

    validateIndex();
  }

  public boolean containsEdge(
      @Nonnull DependencyGraphVertex source,
      @Nonnull DependencyGraphVertex target, 
      @Nonnull Predicate<Dependency> pred) {
    return getAllEdges(source, target).stream()
        .filter(e -> pred.test(e.getDependency()))
        .findFirst()
        .isPresent();
  }
  
  public ImmutableList<DependencyGraphVertex> getSortedVertexList() {
    ImmutableList<DependencyGraphVertex> list = vertexSet().stream()
        .sorted((v1, v2) -> Integer.compare(
            v1.getBeginPosition(),
            v2.getBeginPosition()))
        .collect(ImmutableCollectors.toList());
    // validate the offset
//    DependencyGraphVertex before = list.get(0);
//    for (int i = 1; i < list.size(); i++) {
//      DependencyGraphVertex next = list.get(i);
//      checkState(
//          before.getEndPosition() <= next.getBeginPosition(),
//          "The graph is not sequence: [before=%s, next=%s]", before, next);
//      before = next;
//    }
    return list;
  }

  public boolean hasIncomingEdgesOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return incomingEdgesOf(v).stream()
        .filter(e -> pred.test(e.getDependency()))
        .findFirst()
        .isPresent();
  }

  public boolean hasOutgoingEdgesOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return outgoingEdgesOf(v).stream()
        .filter(e -> pred.test(e.getDependency()))
        .findFirst()
        .isPresent();
  }
  
  public ImmutableSet<DependencyGraphEdge> incomingEdgesOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return incomingEdgesOf(v)
        .stream()
        .filter(e -> pred.test(e.getDependency()))
        .collect(ImmutableCollectors.toSet());
  }

  public ImmutableSet<DependencyGraphVertex> incomingEdgeSourcesOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return incomingEdgesOf(v).stream()
        .filter(e -> pred.test(e.getDependency()))
        .map(this::getEdgeSource)
        .collect(ImmutableCollectors.toSet());
  }

  public ImmutableSet<DependencyGraphEdge> outgoingEdgesOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return outgoingEdgesOf(v)
        .stream()
        .filter(e -> pred.test(e.getDependency()))
        .collect(ImmutableCollectors.toSet());
  }
  
  public ImmutableSet<DependencyGraphVertex> outgoingEdgeTargetsOf(
      @Nonnull DependencyGraphVertex v, 
      @Nonnull Predicate<Dependency> pred) {
    return outgoingEdgesOf(v)
        .stream()
        .filter(e -> pred.test(e.getDependency()))
        .map(this::getEdgeTarget)
        .collect(ImmutableCollectors.toSet());
  }
}