package com.pengyifan.experimental;

import static com.google.common.base.Preconditions.checkArgument;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.builder.ToStringBuilder;

import com.google.common.base.Joiner;
import com.google.common.base.Splitter;
import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.google.common.collect.RangeSet;
import com.google.common.collect.TreeRangeSet;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.brat.io.BratIOUtils;
import com.pengyifan.commons.collections.dependencygraph.MyStyle;
import com.pengyifan.kernel.MyProperties;

public class ParseChristinCorpus {
  public static void main(String[] args) throws IOException {
    ParseChristinCorpus p = new ParseChristinCorpus();
    p.doMain(args);
  }
  
  private boolean debugGenesInvolvedMoreThan2 = true;
  private boolean debugGenesInvolvedLessThan2 = true;
  private boolean debugGenesInvolved = false;
  private boolean debugGenesCited = false;
  private boolean debugDemonstrated = false;
  private boolean debugVocabulary = false;

  private void doMain(String[] args) throws IOException {
    MyProperties properties = new MyProperties("biocreative2");

    Path dropboxDir = properties.getDropboxDir();

    Path corpusFile = dropboxDir.resolve("origin/ISS_TRAIN/ISS_BASIC/Christine_Brun_corpus.txt");
    Corpus c = parse(corpusFile);

    for (Block b : c.blocks) {
      Line pmid = get(b, "PMID", "UI");
      Line sentence = get(b, "SENTENCE");
      // PositiveNegativeStatus status = getStatus(lines);


      BratDocument a1Doc = new BratDocument();
      a1Doc.setText(sentence.text);
      a1Doc.setDocId("BC.d" + pmid.text + ".s" + b.lineOffset);

      List<BratEntity> bratEntities = getGenesCited(b, sentence, a1Doc.getDocId());
      for (BratEntity ent : bratEntities) {
        a1Doc.addAnnotation(ent);
      }

      List<List<BratRelation>> bratRelations = getGenesInvolved(b, a1Doc);
      bratRelations = getDemonstrated(b, bratRelations);
      bratRelations = getVocabulary(b, sentence, bratRelations, a1Doc);
      
      BratDocument a2Doc = new BratDocument();
      int relIndex = 0;
      for(List<BratRelation> list: bratRelations) {
        for (BratRelation rel : list) {
          rel.setId("R" + a1Doc.getDocId() + "p." + relIndex++);
          a2Doc.addAnnotation(rel);
        }
      }

      Path textFile = properties.getBratPositive().resolve(a1Doc.getDocId() + ".txt");
      FileUtils.write(textFile.toFile(), a1Doc.getText());
      Path a1DocFile = properties.getBratPositive().resolve(a1Doc.getDocId() + ".a1");
      BratIOUtils.write(Files.newBufferedWriter(a1DocFile), a1Doc);
      Path a2DocFile = properties.getBratPositive().resolve(a1Doc.getDocId() + ".a2");
      BratIOUtils.write(Files.newBufferedWriter(a2DocFile), a2Doc);
    }
  }
  
  private List<List<BratRelation>> getVocabulary(Block b, Line sentence, 
      List<List<BratRelation>> relations, BratDocument doc) {
    List<Line> sublines = Lists.newArrayList();
    boolean isVoc = false;
    for (int i = 0; i < b.lines.size(); i++) {
      if (b.lines.get(i).text.equals("INTERACTION VOCABULARY")) {
        isVoc = true;
        continue;
      }
      if (isVoc) {
        sublines.add(b.lines.get(i));
      }
    }
    if(sublines.size() != relations.size()) {
      printMsg(debugVocabulary, "vocabulary len[%s] != rel len[%s] at %s",
        sublines.size(), relations.size(), b.lineOffset);
      return relations;
    }
    for(int i=0; i<sublines.size(); i++) {
      Line line = sublines.get(i);
      line.text = trimLeadingNumber(line.text);
      RangeSet<Integer> rs = TreeRangeSet.create();
      for(String part: line.text.split("[.]{3}")) {
        // completely matched
        Pattern pattern = Pattern.compile(part, Pattern.LITERAL | Pattern.CASE_INSENSITIVE);
        Matcher matcher = pattern.matcher(sentence.text);
        while (matcher.find()) {
          rs.add(Range.closedOpen(matcher.start(), matcher.end()));
        }
      }
      filter(relations.get(i), doc, rs);
    }
    
    return relations;
  }
  
  private List<BratRelation> filter(List<BratRelation> relations, BratDocument doc,
      RangeSet<Integer> rs) {
    Iterator<BratRelation> itr = relations.iterator();
    while(itr.hasNext()) {
      BratRelation r = itr.next();
      BratEntity e1 = doc.getEntity(r.getArgId("Arg1"));
      BratEntity e2 = doc.getEntity(r.getArgId("Arg2"));
      RangeSet<Integer> ers = TreeRangeSet.create();
      ers.addAll(e1.getSpans());
      ers.addAll(e2.getSpans());
      if (!rs.enclosesAll(ers)) {
        itr.remove();
      }
    }
    return relations;
  }

  private List<List<BratRelation>> getDemonstrated(Block b, List<List<BratRelation>> relations) {
    List<Line> sublines = Lists.newArrayList();
    boolean isDemon = false;
    for (int i = 0; i < b.lines.size(); i++) {
      if (b.lines.get(i).text.equals("DEMONSTRATED IN SENTENCE")) {
        isDemon = true;
        continue;
      }
      if (isDemon && (b.lines.get(i).text.equals("METHODS")
          || b.lines.get(i).text.equals("INTERACTION TYPE"))) {
        break;
      }
      if (isDemon) {
        sublines.add(b.lines.get(i));
      }
    }
    if(sublines.size() != relations.size()) {
      printMsg(debugDemonstrated, "demonstrated len[%s] != rel len[%s] at %s",
        sublines.size(), relations.size(), b.lineOffset);
      return relations;
    }
    for(int i=0; i<sublines.size(); i++) {
      if (sublines.get(i).text.contains("Y")) {
        for(BratRelation r: relations.get(i)) {
          r.setType(r.getType() + "_Y");
        }
      } else {
        for(BratRelation r: relations.get(i)) {
          r.setType(r.getType() + "_N");
        }
      }
    }
    return relations;
  }

  private List<List<BratRelation>> getGenesInvolved(Block b, BratDocument a1Doc) {
    List<Line> sublines = Lists.newArrayList();
    boolean isInvolved = false;
    for (int i = 0; i < b.lines.size(); i++) {
      if (b.lines.get(i).text.equals("GENES INVOLVED")) {
        isInvolved = true;
        continue;
      }
      if (isInvolved && b.lines.get(i).text.equals("GENES CITED")) {
        break;
      }
      if (isInvolved) {
        sublines.add(b.lines.get(i));
      }
    }
    if (sublines.isEmpty()) {
      return Collections.emptyList();
    }

    List<List<BratRelation>> relations = Lists.newArrayList();
    for(Line line: sublines) {
      line.text = trimLeadingNumber(line.text);
      List<BratRelation> r = getGenesInvolved(line, a1Doc);
      relations.add(r);
    }
    checkArgument(relations.size() == sublines.size());
    return relations;
  }
  
  private String trimLeadingNumber(String s) {
    final Pattern pattern = Pattern.compile("^(\\d+[-])[ ]+(.*)");
    Matcher matcher = pattern.matcher(s);
    if (matcher.find()) {
      return matcher.group(2);
    }
    return s;
  }

  private List<BratRelation> getGenesInvolved(Line line, BratDocument a1Doc) {
    List<EntityMention> entities = Lists.newArrayList();
    for (String tok : Splitter.on(';').split(line.text)) {
      // BRCA1 [actor]
      int leftBracket = tok.lastIndexOf('[');
      String role = leftBracket == -1 ? "UNKNOWN" : getInvolvedType(tok);
      String word = leftBracket == -1 ? tok.trim() : tok.substring(0, leftBracket).trim();
      if (word.endsWith("*")) {
        word = word.substring(0, word.length() - 1);
      }
      EntityMention em = new EntityMention();
      em.word = word;
      em.role = role;
      
      for(BratEntity e: a1Doc.getEntities()) {
        if (e.getText().equalsIgnoreCase(word)) {
          em.entities.add(e);
        }
      }
      if (debugGenesInvolved && em.entities.isEmpty()) {
        printMsg(debugGenesInvolved, "Cannot parse GenesInvolved token %s in %s\n", tok, line);
      } else {
        entities.add(em);
      }
    }
    if (entities.size() < 2) {
      printMsg(debugGenesInvolvedLessThan2, "Less than 2 proteins: %s", line);
      return Collections.emptyList();
    }
    if (entities.size() == 2) {
      List<BratRelation> relations = Lists.newArrayList();
      getRelations(entities.get(0).entities, entities.get(1).entities, a1Doc, relations);
      return relations;
    }
    if (entities.size() > 2) {
      List<BratRelation> relations = Lists.newArrayList();
      for(int i=0; i<entities.size(); i++) {
        EntityMention em1 = entities.get(i);
        if (!em1.role.equalsIgnoreCase("actor")
            && !em1.role.equalsIgnoreCase("no actor")) {
          continue;
        }
        for (int j=0; j<entities.size(); j++) {
          EntityMention em2 = entities.get(j);
          if (!em2.role.equalsIgnoreCase("subject") && !em2.role.equalsIgnoreCase("no subject")) {
            continue;
          }
          getRelations(em1.entities, em2.entities, a1Doc, relations);
        }
      }
      if (relations.isEmpty()) {
        printMsg(debugGenesInvolvedMoreThan2, "More than 2 proteins: %s", line);
      }
    }
    return Collections.emptyList();
  }
  
  private void getRelations(List<BratEntity> e1s, List<BratEntity> e2s, BratDocument doc,
      List<BratRelation> relations) {
    for(BratEntity e1: e1s) {
      for(BratEntity e2: e2s) {
        BratRelation r = new BratRelation();
        r.setType("Interaction-True");
        r.putArgument("Arg1", e1.getId());
        r.putArgument("Arg2", e2.getId());
        relations.add(r);
      }
    }
  }

//  private PositiveNegativeStatus getStatus(Block b) {
//    Line line = get(b, "INTERACTION");
//    if (line.text.equals("N")) {
//      return PositiveNegativeStatus.NEGATIVE;
//    }
//    if (line.text.equals("Y")) {
//      return PositiveNegativeStatus.POSITIVE;
//    }
//    checkArgument(false, "Cannot parse INTERACTION [%s] in\n%s", line, b);
//    return null;
//  }

  private Line get(Block b, String... atts) {
    for (int i = 0; i < b.lines.size(); i++) {
      for (String att : atts) {
        if (b.lines.get(i).text.equals(att)) {
          return b.lines.get(i + 1);
        }
      }
    }
    checkArgument(false, "Cannot find %s in\n %s", Joiner.on(',').join(atts), b.lines);
    return null;
  }

  private List<BratEntity> getGenesCited(Block b, Line sentence, String docId) {
    Line line = get(b, "GENES CITED");

    List<BratEntity> entities = Lists.newArrayList();
    int index = 0;
    for (String tok : Splitter.on(';').split(line.text)) {
      // BRCA1 [H, P]
      int leftBracket = tok.lastIndexOf('[');
      String type = leftBracket == -1 ? "UNKNOWN" : getCitedType(tok);
      String word = leftBracket == -1 ? tok.trim() : tok.substring(0, leftBracket).trim();
      if (word.endsWith("*")) {
        word = word.substring(0, word.length() - 1);
      }

      Pattern pattern = Pattern.compile(word, Pattern.LITERAL | Pattern.CASE_INSENSITIVE);
      Matcher matcher = pattern.matcher(sentence.text);
      if (!matcher.find()) {
        printMsg(debugGenesCited, "Check %s in %s", word, line);
      }
      matcher = pattern.matcher(sentence.text);
      while (matcher.find()) {
        BratEntity ent = new BratEntity();
        ent.setId("T" + docId + ".e" + index++);
        ent.setText(matcher.group());
        ent.addSpan(matcher.start(), matcher.end());
        ent.setType(type);
        entities.add(ent);
      }
    }
    if (entities.size() < 2) {
      System.err.println(String.format("Check %s", line));
    }
    return entities;
  }

  // E1A [actor]
  private String getInvolvedType(String tok) {
    int rightBracket = tok.lastIndexOf(']');
    checkArgument(rightBracket != -1, "Cannot parse %s", tok);
    int leftBracket = tok.lastIndexOf('[');
    return tok.substring(leftBracket + 1, rightBracket).trim();
  }

  // BRCA1 [H, P]
  private String getCitedType(String tok) {
    int rightBracket = tok.lastIndexOf(']');
    checkArgument(rightBracket != -1, "Cannot parse %s", tok);
    int leftBracket = tok.lastIndexOf('[');
    checkArgument(leftBracket != -1, "Cannot parse %s", tok);
    String types = tok.substring(leftBracket + 1, rightBracket);
    if (types.contains("P") || types.contains("p")) {
      return "Protein";
    } else if (types.contains("D")) {
      return "DNA";
    } else if (types.contains("R")) {
      return "RNA";
    }
    checkArgument(false, "Cannot parse type [%s] in %s", types, tok);
    return null;
  }
  
  private void printMsg(boolean b, String format, Object ... args) {
    if (b) {
      System.err.println(String.format(format, args));
    }
  }

  private Corpus parse(Path file) throws IOException {
    List<String> lines = Files.readAllLines(file);

    Corpus c = new Corpus();
    Block b = new Block();
    b.lineOffset = 1;
    for (int i = 0; i < lines.size(); i++) {
      String line = lines.get(i);
      if (line.isEmpty()) {
        checkArgument(lines.get(i + 1).isEmpty(), "Cannot parse line: %s", i + 1);
        checkArgument(lines.get(i + 2).isEmpty(), "Cannot parse line: %s", i + 1);
        i += 2;
        c.blocks.add(b);
        b = new Block();
        b.lineOffset = i + 2;
      } else {
        b.lines.add(new Line(i + 1, lines.get(i)));
      }
    }
    if (!b.lines.isEmpty()) {
      c.blocks.add(b);
    }
    return c;
  }

  private class Corpus {
    List<Block> blocks;

    Corpus() {
      blocks = Lists.newArrayList();
    }
  }

  private class Block {
    int lineOffset;
    List<Line> lines;

    Block() {
      lines = Lists.newArrayList();
    }

    @Override
    public String toString() {
      return lineOffset + "\n" + Joiner.on('\n').join(lines);
    }
  }

  private class Line {
    int lineOffset;
    String text;

    Line(int lineOffset, String line) {
      this.lineOffset = lineOffset;
      this.text = line;
    }

    @Override
    public String toString() {
      return new ToStringBuilder(this, new MyStyle())
          .append("offset", lineOffset)
          .append("text", text)
          .toString();
    }
  }
  
  private class EntityMention {
    List<BratEntity> entities;
    String word;
    String role;
    
    EntityMention () {
      entities = Lists.newArrayList();
    }
  }
}
