package com.pengyifan.kernel.allpathgraph.singleprocess;

import com.google.common.collect.Lists;
import com.google.common.collect.Range;
import com.google.common.primitives.Doubles;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.NamedEntity;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.commons.collections.dependencygraph.Position;
import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang3.math.NumberUtils;

import static com.google.common.base.Preconditions.checkArgument;

public class DependencyGraphVertexSpliter {

  private DependencyGraph graph;
  private Collection<BratEntity> proteins;
  private static final Pattern PATTERN_SLASH_DASH = Pattern
      .compile("[-/.:]");
  private static final Pattern PATTERN_RELATED = Pattern
      .compile("(in)?dependent|like|specific|positive");
  private static final Pattern PATTERN_VBG = Pattern
      .compile("[a-z]+(ing)");
  private static final Pattern PATTERN_VBN = Pattern
      .compile("[a-z]+(ed)");

  public DependencyGraphVertexSpliter(DependencyGraph graph,
      Collection<BratEntity> proteins) {
    this.graph = graph;
    this.proteins = proteins;
  }

  public DependencyGraph split() {
    return split1();
  }

  private DependencyGraph split1() {
    DependencyGraphVertex newVertex;
    DependencyGraphEdge newEdge;

    int maxIndex = graph.getMaxIndex() + 1;
    for (DependencyGraphVertex v : graph.vertexSet()) {
      BratEntity token = getEntity(
          v.getWord(),
          Range.closedOpen(v.getBeginPosition(), v.getEndPosition()),
          v.getTag().toString());
      token.setType("Others");
      List<BratEntity> tokens = split(chopSymbol(token));
      if (tokens == null) {
        continue;
      }
      if (tokens.size() == 1) {
        continue;
      }
      String type = getType(tokens);
      switch (type) {
      case "Empty-Protein":
        v.setWord(tokens.get(1).getText());
        v.setBeginPosition(tokens.get(1).beginPosition());
        break;
      case "Protein-Empty":
        v.setWord(tokens.get(0).getText());
        v.setEndPosition(tokens.get(0).endPosition());
        break;
      case "Protein-Protein":
      case "ProteinEnd-Protein":
      case "ProteinEnd-ProteinStart":
      case "Protein-ProteinStart":
        v.setWord(tokens.get(0).getText());
        v.setBeginPosition(tokens.get(0).beginPosition());
        v.setEndPosition(tokens.get(0).endPosition());
        // add new vertex
        newVertex = createNewVertex(maxIndex, tokens.get(1), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        // add new edge
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.CONJ, "dash"));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Protein-Dependant":
        v.setWord(tokens.get(1).getText());
        v.setBeginPosition(tokens.get(1).beginPosition());
        v.setEndPosition(tokens.get(1).endPosition());
        // add new vertex
        newVertex = createNewVertex(maxIndex, tokens.get(0), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        // add new edge
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.NN, null));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Protein-VBG":
        v.setWord(tokens.get(1).getText());
        v.setBeginPosition(tokens.get(1).beginPosition());
        v.setEndPosition(tokens.get(1).endPosition());
        v.setTag(PartOfSpeech.VBG);
        // add new vertex
        newVertex = createNewVertex(maxIndex, tokens.get(0), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        // add new edge
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.ARG0, null));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Protein-VBN":
        v.setWord(tokens.get(1).getText());
        v.setBeginPosition(tokens.get(1).beginPosition());
        v.setEndPosition(tokens.get(1).endPosition());
        v.setTag(PartOfSpeech.VBN);
        // add new vertex
        newVertex = createNewVertex(maxIndex, tokens.get(0), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        // add new edge
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.ARG1, null));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Protein-Protein-Protein":
        v.setWord(tokens.get(0).getText());
        v.setBeginPosition(tokens.get(0).beginPosition());
        v.setEndPosition(tokens.get(0).endPosition());
        // second
        newVertex = createNewVertex(maxIndex, tokens.get(1), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.CONJ, "dash"));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        // third
        newVertex = createNewVertex(maxIndex, tokens.get(2), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.CONJ, "dash"));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Protein-Protein-Dependant":
        v.setWord(tokens.get(0).getText());
        v.setBeginPosition(tokens.get(0).beginPosition());
        v.setEndPosition(tokens.get(0).endPosition());
        // second
        newVertex = createNewVertex(maxIndex, tokens.get(2), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.CONJ, "dash"));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        // third
        DependencyGraphVertex newVertex2 = createNewVertex(
            maxIndex,
            tokens.get(1),
            PartOfSpeech.NN,
            NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex2);
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.NN, null));
        maxIndex++;
        graph.addEdge(newVertex, newVertex2, newEdge);
        break;
      case "Protein-Protein-Others":
        v.setWord(tokens.get(0).getText());
        v.setBeginPosition(tokens.get(0).beginPosition());
        v.setEndPosition(tokens.get(0).endPosition());
        // add new vertex
        newVertex = createNewVertex(maxIndex, tokens.get(1), PartOfSpeech.NN, NamedEntity.PROTEIN_X);
        maxIndex++;
        graph.addVertex(newVertex);
        // add new edge
        newEdge = DependencyGraphEdge.of(maxIndex, new Dependency(DependencyName.CONJ, "dash"));
        maxIndex++;
        graph.addEdge(v, newVertex, newEdge);
        break;
      case "Others-Protein":
      case "Protein-Others":
        break;
      }
    }

    return graph;
  }

  private DependencyGraphVertex createNewVertex(int index, BratEntity token,
      PartOfSpeech tag, NamedEntity ne) {
    return new DependencyGraphVertex(index,
        token.getText(),
        tag,
        null,
        ne,
        Position.UNKNOWN,
        token.beginPosition(),
        token.endPosition());
  }

  private BratEntity chopSymbol(BratEntity token) {
    Range<Integer> range = Range.closedOpen(
        token.beginPosition() + 1,
        token.endPosition());
    if (startsWithSymbol(token.getText())) {
      token = getEntity(
          token.getText().substring(1),
          range,
          "LBRAC");
      token.setType(token.getType());
    }
    range = Range.closedOpen(
        token.beginPosition(),
        token.endPosition() - 1);
    if (endsWithSymbol(token.getText())) {
      token = getEntity(
          token.getText().substring(0, token.getText().length() - 1),
          range,
          "RBRAC");
      token.setType(token.getType());
    }
    return token;
  }

  private List<BratEntity> split(BratEntity token) {
    boolean hasDash = false;
    Matcher matcher = PATTERN_SLASH_DASH.matcher(token.getText());
    while (matcher.find()) {
      hasDash = true;
      if (isProtein(token.totalSpan())) {
        token.setType("Protein");
        return Lists.newArrayList(token);
      }
      if (isSymbol(token.getText())) {
        token.setType("Symbol");
        return Lists.newArrayList(token);
      }
      Range<Integer> range = Range.closedOpen(
          token.beginPosition(),
          token.beginPosition() + matcher.start());
      if (range.isEmpty()) {
        continue;
      }
      // left
      BratEntity leftToken = getEntity(
          token.getText().substring(0, matcher.start()),
          range,
          token.getType());
      range = Range.closedOpen(
          token.beginPosition() + matcher.end(),
          token.endPosition());
      if (range.isEmpty()) {
        continue;
      }
      // right
      BratEntity rightToken = getEntity(
          token.getText().substring(matcher.end()),
          range,
          token.getType());

      if (isProtein(leftToken.totalSpan()) && isProtein(rightToken.totalSpan())) {
        leftToken.setType("Protein");
        rightToken.setType("Protein");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (endsWithProtein(leftToken.totalSpan())
          && isProtein(rightToken.totalSpan())) {
        leftToken.setType("ProteinEnd");
        rightToken.setType("Protein");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (endsWithProtein(leftToken.totalSpan())
          && startsWithProtein(rightToken.totalSpan())) {
        leftToken.setType("ProteinEnd");
        rightToken.setType("ProteinStart");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(leftToken.totalSpan())
          && startsWithProtein(rightToken.totalSpan())) {
        leftToken.setType("Protein");
        rightToken.setType("ProteinStart");

        // try once more
        List<BratEntity> tokens = split(rightToken);
        if (tokens == null) {
          return Lists.newArrayList(leftToken, rightToken);
        } else {
          tokens.add(0, leftToken);
          return tokens;
        }
      } else if (isProtein(leftToken.totalSpan())
          && isDependant(rightToken.getText())) {
        leftToken.setType("Protein");
        rightToken.setType("Dependant");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(leftToken.totalSpan())
          && isVBG(rightToken.getText())) {
        leftToken.setType("Protein");
        rightToken.setType("VBG");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(leftToken.totalSpan())
          && isVBN(rightToken.getText())) {
        leftToken.setType("Protein");
        rightToken.setType("VBN");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(leftToken.totalSpan())
          && isEmpty(rightToken.totalSpan())) {
        leftToken.setType("Protein");
        rightToken.setType("Empty");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(leftToken.totalSpan())) {
        leftToken.setType("Protein");
        rightToken.setType("Others");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isEmpty(leftToken.totalSpan())
          && isProtein(rightToken.totalSpan())) {
        leftToken.setType("Empty");
        rightToken.setType("Protein");
        return Lists.newArrayList(leftToken, rightToken);
      } else if (isProtein(rightToken.totalSpan())) {
        leftToken.setType("Others");
        rightToken.setType("Protein");
        return Lists.newArrayList(leftToken, rightToken);
      }
    }
    if (hasDash) {
      return Lists.newArrayList(token);
    } else {
      return null;
    }
  }

  private boolean isEmpty(Range<Integer> range) {
    return range.isEmpty();
  }

  private boolean startsWithSymbol(String token) {
    return (token.charAt(0) == '['
        || token.charAt(0) == ']'
        || token.charAt(0) == '('
        || token.charAt(0) == ')')
        && token.length() > 1;
  }

  private boolean endsWithSymbol(String token) {
    return (token.endsWith("[")
        || token.endsWith("]")
        || token.endsWith("(")
        || token.endsWith(")"))
        && token.length() > 1;
  }

  private boolean isSymbol(String token) {
    return token.equals("[")
        || token.equals("]")
        || token.equals("(")
        || token.equals(")")
        // || token.equals("-LRB-")
        // || token.equals("-RRB-")
        // || token.equals("-LSB-")
        // || token.equals("-RSB-")
        || token.equals(".")
        || token.equals(".")
        || NumberUtils.isNumber(token)
        || Doubles.tryParse(token) != null;
  }

  private boolean isProtein(Range<Integer> range) {
    for (BratEntity entity : proteins) {
      if (entity.totalSpan().equals(range)) {
        return true;
      }
    }
    return false;
  }

  private boolean isDependant(String word) {
    return PATTERN_RELATED.matcher(word).matches();
  }

  private boolean isVBG(String word) {
    return PATTERN_VBG.matcher(word).matches();
  }

  private boolean isVBN(String word) {
    return PATTERN_VBN.matcher(word).matches();
  }

  private BratEntity getEntity(String token, Range<Integer> range, String type) {
    checkArgument(!range.isEmpty(), "Begin >= end: %s%s", range, token);
    BratEntity entity = new BratEntity();
    entity.addSpan(range);
    entity.setText(token);
    entity.setType(type);
    return entity;
  }

  private boolean endsWithProtein(Range<Integer> range) {
    for (BratEntity entity : proteins) {
      if (range.upperEndpoint() == (entity.endPosition())) {
        return true;
      }
    }
    return false;
  }

  private boolean startsWithProtein(Range<Integer> range) {
    for (BratEntity entity : proteins) {
      if (range.lowerEndpoint() == (entity.beginPosition())) {
        return true;
      }
    }
    return false;
  }

  private String getType(List<BratEntity> tokens) {
    String s = tokens.get(0).getType();
    for (int i = 1; i < tokens.size(); i++) {
      s += "-" + tokens.get(i).getType();
    }
    return s;
  }
}
