package com.pengyifan.nlp.ssdg;

import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;

class AbilityLinkAdder extends EdgeAdder {

  public AbilityLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    numOfEdges += addAbility();
    numOfEdges += addItsAbility();
    numOfEdges += addIsCapable();
    numOfEdges += addCapable();
    return numOfEdges;
  }

  /**
   * <pre>
   *             vmod/dep
   *   +------------------------+
   *   |  prep_of      arg0     |
   *   |+-------++-------------+|
   *   ||       vv             |v
   * ability of N   to < aux < VB
   * </pre>
   */
  private static final DependencyGraphPattern abilityPattern = DependencyGraphPattern
      .of()
      // ability
      .addVertex(0, v -> "ability".equals(v.getLemma()))
      // n
      .addVertex(1, v -> true)
      // aux
      .addVertex(2, v -> true)
      // vb
      .addVertex(3, v -> true)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"))
      .addEdge(0, 3, Dependency.ofPredicate(DependencyName.VMOD, DependencyName.DEP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.AUX));

  private int addAbility() {
    return add(abilityPattern, 3, 1, DependencyName.ARG0, "ability");
  }

  /**
   * <pre>
   *             arg0
   *  +-------------------------------+
   *  |                    vmod       |
   *  |              +---------------+|
   *  v              |               v|
   * its < poss < ability to < aux < VB
   * </pre>
   */
  private static final DependencyGraphPattern itsAbilityPattern = DependencyGraphPattern
      .of()
      // its
      .addVertex(0, v -> true)
      // ability
      .addVertex(1, v -> "ability".equals(v.getLemma()))
      // aux
      .addVertex(2, v -> true)
      // vb
      .addVertex(3, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.POSS))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.VMOD, DependencyName.DEP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.AUX));

  private int addItsAbility() {
    return add(itsAbilityPattern, 3, 0, DependencyName.ARG0, "ability");
  }

  /**
   * <pre>
   *           arg0
   * +----------------------------+
   * |    nsubj        prepc_of  |
   * | +-------------+ +--------+ |
   * v v             + +        v +
   *  N  is < cop < capable of  VBG
   * </pre>
   */
  private static final DependencyGraphPattern isCapablePattern = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> true)
      // is
      .addVertex(1, v -> true)
      // capable
      .addVertex(2, v -> "capable".equals(v.getLemma()))
      // vbg
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VBG)
      .addEdge(2, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.COP))
      .addEdge(2, 3, Dependency.ofPrepPredicate("of"));

  private int addIsCapable() {
    return add(isCapablePattern, 3, 0, DependencyName.ARG0, "capable");
  }

  /**
   * <pre>
   *           arg0            
   * +------------------------+
   * |   amod      prep_of    |
   * | +-------+ +----------+ |
   * v +       v +          v +
   *  N      capable   of   VBG
   * </pre>
   */
  private static final DependencyGraphPattern capablePattern = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> true)
      // capable
      .addVertex(1, v -> "capable".equals(v.getLemma()))
      // vbg
      .addVertex(2, v -> v.getTag() == PartOfSpeech.VBG)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.AMOD))
      .addEdge(1, 2, Dependency.ofPrepPredicate("of"));

  private int addCapable() {
    return add(capablePattern, 2, 0, DependencyName.ARG0, "capable");
  }
}
