package com.pengyifan.nlp.ssdg;

import java.util.Set;
import java.util.function.Predicate;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.nlp.ling.string.GreekString;

class CTermFTermLinkAdder extends EdgeAdder {

  private static final Set<String> keywords = Sets.newHashSet(
      "aa",
      "all",
      "amount",
      "antigen",
      "cell",
      //"chain",
      "coactivator",
      "co-receptor",
      "deposit",
      "element",
      "factor",
      "form",
      "gene",
      "glycoform",
      "holoprotein",
      "homolog",
      "isoform",
      "kinase",
      "level",
      "mab",
      "molecule",
      "monomer",
      "mutant",
      "mutation",
      "oncoprotein",
      "partner",
      "peptide",
      "product",
//      "promoter",
      "protein",
      "proto-oncogene",
      "structure",
      "subtilis",
      "suppressor",
      "transporter",
      "variant",
      "variety"
      );

  private static boolean isKeyWord(String keyword, Set<String> set) {
    return set.contains(keyword);
  }

  private final static Predicate<DependencyGraphVertex> ftermP = v -> isKeyWord(
      v.getLemma(),
      keywords);
  private final static Predicate<DependencyGraphVertex> ctermP = v -> v
      .getNer()
      .isProtein();

  CTermFTermLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  int add() {
    int numOfEdges = 0;
    numOfEdges += addCTermFterm();
    numOfEdges += addPrepOf();
    numOfEdges += addFtermCTerm();
    numOfEdges += addSubunitCTermFterm();
    return numOfEdges;
  }

  /**
   * cterm < nn < fterm
   */
  private static final DependencyGraphPattern subunitCtermFtermPattern = DependencyGraphPattern
      .of()
          // cterm
      .addVertex(0, v -> GreekString.isEndWithGreek(v.getLemma()))
          // fterm
      .addVertex(1, v -> v.getLemma().equals("subunit"))
      .addEdge(1, 0, Dependency.NnPredicate);
  private int addSubunitCTermFterm() {
    return add(subunitCtermFtermPattern, 1, 0, DependencyName.DESCRIPTIVE, "nn");
  }

  /**
   * cterm < nn < fterm
   */
  private static final DependencyGraphPattern ctermFtermPattern = DependencyGraphPattern
      .of()
      // cterm
      .addVertex(0, ctermP)
          // fterm
      .addVertex(1, ftermP)
      .addEdge(1, 0, Dependency.NnPredicate);

  private int addCTermFterm() {
    return add(ctermFtermPattern, 1, 0, DependencyName.DESCRIPTIVE, "nn");
  }

  /**
   * fterm < nn/amod < cterm
   * 
   * or
   * 
   * fterm > dep > cterm
   */
  private static final DependencyGraphPattern ftermCtermPattern1 = DependencyGraphPattern
      .of()
      .addVertex(0, ctermP)
      .addVertex(1, ftermP)
      .addEdge(0, 1, Dependency.NnPredicate);

  private static final DependencyGraphPattern ftermCtermPattern2 = DependencyGraphPattern
      .of()
      .addVertex(0, ftermP)
      .addVertex(1, ctermP)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.DEP));

  private int addFtermCTerm() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    /*
     * level < nn/amod < protein
     */
    DependencyGraphMatcher matcher = ftermCtermPattern1.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex cterm = matcher.get(0);
      DependencyGraphVertex fterm = matcher.get(1);
      if (cterm.getBeginPosition() > fterm.getBeginPosition()) {
        edgeIndex = addIfNotExist(
            fterm,
            cterm,
            DependencyName.DESCRIPTIVE,
            fterm.getLemma(),
            edgeIndex);
      }
    }
    /*
     * fterm > dep > protein
     */
    matcher = ftermCtermPattern2.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex fterm = matcher.get(0);
      DependencyGraphVertex cterm = matcher.get(1);
      if (cterm.getBeginPosition() > fterm.getBeginPosition()) {
        edgeIndex = addIfNotExist(
            fterm,
            cterm,
            DependencyName.DESCRIPTIVE,
            fterm.getLemma(),
            edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * fterm > of > cterm
   */
  private static final DependencyGraphPattern prepOfPattern = DependencyGraphPattern
      .of()
      .addVertex(0, ftermP)
      .addVertex(1, ctermP)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"));
  /**
   * fterm > of > x > descriptive > cterm
   */
  private static final DependencyGraphPattern prepOfPattern2 = DependencyGraphPattern
      .of()
      .addVertex(0, ftermP)
      .addVertex(1, v -> true)
      .addVertex(2, ctermP)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DESCRIPTIVE));
  private int addPrepOf() {
    int numOfEdges = 0;
    numOfEdges += add(prepOfPattern, 0, 1, DependencyName.DESCRIPTIVE, "of");
    numOfEdges += add(prepOfPattern2, 0, 1, DependencyName.DESCRIPTIVE, "of");
    return numOfEdges;
  }
}
