package com.pengyifan.nlp.ssdg;

import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.pengyifan.brat.BratDocument;
import com.pengyifan.brat.BratEntity;
import com.pengyifan.brat.BratRelation;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.kernel.allpathgraph.singleprocess.DependencyGraphVertexFinder;

class CoreferenceAdder {

  private final DependencyGraph graph;
  private final BratDocument corefDoc;

  public CoreferenceAdder(DependencyGraph graph, BratDocument corefDoc) {
    this.graph = graph;
    this.corefDoc = corefDoc;
  }

  /**
   * If referee is protein, add edge from referer to referee;
   */
  public int add() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (BratRelation r : corefDoc.getRelations()) {
      // its
      BratEntity referer = corefDoc.getEntity(r.getArgId("Referer"));
      // if ("their".equalsIgnoreCase(referer.getText())) {
      // continue;
      // }
      if ("they".equalsIgnoreCase(referer.getText())) {
        continue;
      }
      // protein
      BratEntity referee = corefDoc.getEntity(r.getArgId("Referee"));

      // skip if the reference is forward.
      if (referer.beginPosition() < referee.beginPosition()) {
        continue;
      }

      List<DependencyGraphVertex> refererVertexList =
          new DependencyGraphVertexFinder(graph,
              referer.totalSpan()).find();
      DependencyGraphVertex refererVertex = refererVertexList.get(0);

      List<DependencyGraphVertex> refereeVertexList =
          new DependencyGraphVertexFinder(graph,
              referee.totalSpan()).find();
      if (refereeVertexList.isEmpty()) {
        continue;
      }

      if ("its".equalsIgnoreCase(referer.getText())
          || "it".equalsIgnoreCase(referer.getText())
          || "itself".equalsIgnoreCase(referer.getText())) {
        DependencyGraphVertex refereeHeadVertex = changeHead(getHead(refereeVertexList));
        edgeIndex = addEdge(refererVertex, refereeHeadVertex, edgeIndex);
      } else if ("their".equalsIgnoreCase(referer.getText())) {
        // filter proteins
        List<DependencyGraphVertex> proteins = refereeVertexList.stream()
            .filter(v -> v.getNer().isProtein()).collect(Collectors.toList());
        if (proteins.size() > 1) {
          for (DependencyGraphVertex protein : proteins) {
            edgeIndex = addEdge(refererVertex, protein, edgeIndex);
          }
        }
      } else {
        DependencyGraphVertex refereeHeadVertex = getHead(refereeVertexList);
        if (refereeHeadVertex.getTag() == PartOfSpeech.NNS) {
          edgeIndex = addEdge(refererVertex, refereeHeadVertex, edgeIndex);
        } else {
          for (DependencyGraphVertex refereeVertex : refereeVertexList) {
            edgeIndex = addEdge(refererVertex, refereeVertex, edgeIndex);
          }
        }
      }
    }
    return edgeIndex - startIndex;
  }

  private DependencyGraphVertex getHead(List<DependencyGraphVertex> vertexList) {
    Set<Integer> indices = vertexList.stream().map(v -> v.getIndex())
        .collect(Collectors.toSet());
    // the vertex whose incoming is outside
    List<DependencyGraphVertex> heads = Lists.newArrayList();
    for (DependencyGraphVertex v : vertexList) {
      for (DependencyGraphVertex inV : graph.incomingEdgeSourcesOf(v)) {
        if (!indices.contains(inV.getIndex())) {
          heads.add(v);
          break;
        }
      }
    }
    if (heads.isEmpty()) {
      return Iterables.getLast(vertexList);
    }
    if (heads.size() == 1) {
      return heads.get(0);
    }
    return Iterables.getLast(heads);
  }

  private DependencyGraphVertex changeHead(DependencyGraphVertex v) {
    if (v.getNer().isProtein()) {
      return v;
    }
    int numberOfProtein = 0;
    DependencyGraphVertex protein = null;
    for (DependencyGraphVertex n : graph.outgoingEdgeTargetsOf(
        v,
        Dependency.ofPredicate(DependencyName.PREP, "of"))) {
      if (n.getNer().isProtein()) {
        protein = n;
        numberOfProtein++;
      }
    }
    if (numberOfProtein == 1) {
      return protein;
    }
    return v;
  }

  private int addEdge(DependencyGraphVertex source,
      DependencyGraphVertex target, int edgeIndex) {
    if (!graph.containsEdge(
        source,
        target,
        Dependency.ofPredicate(DependencyName.COREF))) {
      graph.addEdge(source, target,
          DependencyGraphEdge.of(edgeIndex, new Dependency(DependencyName.COREF, null)));
      edgeIndex++;
    }
    return edgeIndex;
  }
}
