package com.pengyifan.nlp.ssdg;

import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.NamedEntity;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;

class IsALinkAdder extends EdgeAdder {

  public IsALinkAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    numOfEdges += addAppositive();
    numOfEdges += addNamed();
    numOfEdges += addDesignated();
    numOfEdges += addIdentifyAs();
    numOfEdges += addKnownAs();
    numOfEdges += addFunctionAs();
    numOfEdges += addIdentifiedAs();
    numOfEdges += addServeAs();
    numOfEdges += addIsA();
    numOfEdges += addAre();
    numOfEdges += addEncode();
    numOfEdges += addEncoding();
    return numOfEdges;
  }

  /**
   * n < nsubj < encode > dobj > n
   */
  private static final DependencyGraphPattern encodePattern = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> v.getTag().isNoun())
          // encodes
      .addVertex(1, v -> v.getLemma().equals("encode"))
        // n
      .addVertex(2, v -> v.getTag().isNoun())
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addEncode() {
    return add(encodePattern, 0, 2, DependencyName.IS_A, "encode");
  }

  /**
   * <pre>
   *               is-a
   * +-----------------------------+
   * |                             v
   * N > vmod > encoding > dobj > N
   * </pre>
   */
  private static final DependencyGraphPattern encodingPattern = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> v.getTag().isNoun())
          // encoding
      .addVertex(1, v -> v.getWord().equals("encoding"))
          // n
      .addVertex(2, v -> v.getTag().isNoun())
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addEncoding() {
    return add(encodingPattern, 0, 1, DependencyName.IS_A, "encode");
  }

  private static final DependencyGraphPattern apposPattern = DependencyGraphPattern
      .of()
      // source
      .addVertex(0, v -> true)
      // target
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.APPOS));

  private int addAppositive() {
    return add(apposPattern, 0, 1, DependencyName.IS_A, "appos");
  }

  /**
   * <pre>
   *                 is-a
   * +-------------------------------------+
   * |      nsubjpass                      |
   * + +----------------------+            |
   * v v                      |            |
   *  N      is < auxpass < named > dobj > N
   * </pre>
   */
  private static final DependencyGraphPattern namedPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // named
      .addVertex(1, v -> "named".equals(v.getWord()))
      // n2
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addNamed() {
    return add(namedPattern, 2, 0, DependencyName.IS_A, "named");
  }

  /**
   * <pre>
   *             is-a
   * +-----------------------------+
   * v                             |
   * N > vmod > designated > dep > N
   *            termed
   * </pre>
   */
  private static final DependencyGraphPattern designatedPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // designated
      .addVertex(1, v -> "designated".equals(v.getWord())
          || "termed".equals(v.getWord())
          || "named".equals(v.getWord()))
      // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DEP));

  private int addDesignated() {
    return add(designatedPattern, 2, 0, DependencyName.IS_A, "named");
  }

  /**
   * <pre>
   *            is-a
   * +-------------------------+
   * |               prep_as   |
   * |            +-----------+|
   * |            |           vv
   * N > vmod > known  as     N
   * </pre>
   */
  private static final DependencyGraphPattern knownPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // known
      .addVertex(1, v -> "known".equals(v.getWord()))
      // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.PREP, "as"));

  private int addKnownAs() {
    return add(knownPattern, 0, 2, DependencyName.IS_A, "known-as");
  }

  /**
   * <pre>
   *            is-a
   * +-------------------------+
   * |               prep_as   |
   * |            +-----------+|
   * |            |           vv
   * N > dep > functions as    N
   * </pre>
   */
  private static final DependencyGraphPattern functionPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // function
      .addVertex(1, v -> v.getLemma().equals("function"))
      // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.DEP))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.PREP, "as"));

  private int addFunctionAs() {
    return add(functionPattern, 0, 2, DependencyName.IS_A, "function-as");
  }

  /**
   * <pre>
   *           prep_as
   *     +----------------------+
   *     |                is-a  |
   *     |             +-------+|
   *     |             v       |v
   * identify > dobj > N1  as  N2
   */
  private static final DependencyGraphPattern identifyAsPattern = DependencyGraphPattern
      .of()
      // identify
      .addVertex(0, v -> v.getLemma().startsWith("identif"))
      // n1
      .addVertex(1, v -> true)
      // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.DOBJ))
      .addEdge(0, 2, Dependency.ofPredicate(DependencyName.PREP, "as"));

  private int addIdentifyAs() {
    return add(identifyAsPattern, 2, 1, DependencyName.IS_A, "identified-as");
  }

  /**
   * <pre>
   *                 is-a
   * +-------------------------------------+
   * |      nsubjpass              prep_as |
   * ++----------------------+    +-------+|
   * vv                      |    |       v|
   * N      is < auxpass < identified as  N
   * </pre>
   */
  private static final DependencyGraphPattern identifiedAsPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // identified
      .addVertex(1, v -> "identified".equals(v.getWord())
          || "known".equals(v.getWord())
          || "characterized".equals(v.getWord())
          || "considered".equals(v.getWord()))
      // n2
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.PREP, "as"));

  private int addIdentifiedAs() {
    return add(identifiedAsPattern, 0, 2, DependencyName.IS_A, "identified-as");
  }

  /**
   * <pre>
   *              is
   * +--------------------------+
   * |                 prep_as  |
   * |              +---------+ |
   * |              |         v v
   * N < nsubj < serve     as  N
   *             functions
   * </pre>
   */
  private static final DependencyGraphPattern servePattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // function
      .addVertex(1, v -> v.getLemma().equals("act")
          || v.getLemma().equals("function")
          || v.getLemma().equals("serve"))
      // n2
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.PREP, "as"));

  private int addServeAs() {
    return add(servePattern, 0, 2, DependencyName.IS_A, "serve-as");
  }

  /**
   * <pre>
   *          nsubj
   * +---------------------+
   * |            cop      |
   * |       +------------+|
   * v       V            ||
   * N      is  a < det < N
   * |                    ^
   * +--------------------+
   *         is-a
   * </pre>
   */
  private static final DependencyGraphPattern isAPattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // is
      .addVertex(1, v -> true)
      // a
      .addVertex(2, v -> true)
      // n2
      .addVertex(3, v -> true)
      .addEdge(3, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(3, 1, Dependency.ofPredicate(DependencyName.COP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.DET));

  private int addIsA() {
    return add(isAPattern, 0, 3, DependencyName.IS_A, "is-a");
  }

  /**
   * <pre>
   *          nsubj
   * +---------------------+
   * |            cop      |
   * |       +------------+|
   * v       V            ||
   * N     are  CD < num < N
   * |                    ^
   * +--------------------+
   * </pre>
   */
  private static final DependencyGraphPattern arePattern = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> true)
      // are
      .addVertex(1, v -> true)
      // n2
      .addVertex(3, v -> v.getTag() == PartOfSpeech.NNS)
      .addEdge(3, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(3, 1, Dependency.ofPredicate(DependencyName.COP));
  
  private int addAre() {
    return add(arePattern, 0, 3, DependencyName.IS_A, "are-num");
  }

  private boolean hasDet(DependencyGraphVertex v) {
    return graph.hasOutgoingEdgesOf(v, Dependency.ofPredicate(DependencyName.DET));
  }

  private boolean hasNNS(DependencyGraphVertex v) {
    return v.getTag() == PartOfSpeech.NNS;
  }

  private boolean hasNum(DependencyGraphVertex v) {
    return graph.hasOutgoingEdgesOf(v, Dependency.ofPredicate(DependencyName.NUM));
  }

  @Override
  protected int addIfNotExist(DependencyGraphVertex source,
      DependencyGraphVertex target,
      DependencyName name,
      String specific,
      int index) {
    if (source == target) {
      return index;
    }
    if (!hasNum(source) && hasNum(target)) {
      ;
    } else if (hasNum(source) && !hasNum(target)) {
      DependencyGraphVertex tmp = source;
      source = target;
      target = tmp;
    } else if (!hasNNS(source) && hasNNS(target)) {
      ;
    } else if (hasNNS(source) && !hasNNS(target)) {
      DependencyGraphVertex tmp = source;
      source = target;
      target = tmp;
    } else if (!hasDet(source) && hasDet(target)) {
      ;
    } else if (hasDet(source) && !hasDet(target)) {
      DependencyGraphVertex tmp = source;
      source = target;
      target = tmp;
    }
    index = addIfNotExist(source, target, specific, index);
    if (source.getNer() != NamedEntity.UNKNOWN
        && target.getNer() != NamedEntity.UNKNOWN) {
      index = addIfNotExist(target, source, specific, index);
    }
    return index;
  }

  private int addIfNotExist(DependencyGraphVertex source,
      DependencyGraphVertex target, String specific, int index) {
    if (graph.containsEdge(source, target, Dependency.ofPredicate(DependencyName.IS_A))
        || graph.containsEdge(target, source, Dependency.ofPredicate(DependencyName.IS_A))) {
      return index;
    }
    graph.addEdge(source, target,
        DependencyGraphEdge.of(index, new Dependency(DependencyName.IS_A, specific)));
    return index + 1;
  }
}
