package com.pengyifan.nlp.ssdg;

import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;

class MemberCollectionLinkAdder extends EdgeAdder {

  private static final Set<String> keywords = Sets.newHashSet(
      "all",
      "cluster",
      "family"
      );

  private static boolean isKeyWord(String keyword) {
    return keywords.contains(keyword);
  }

  public MemberCollectionLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    numOfEdges += addPrepIncluding();
    numOfEdges += addVmodIncluding();
    numOfEdges += addVerbInclude();
    numOfEdges += addContaining();
    numOfEdges += addContains();
    numOfEdges += addNnsConj();

    numOfEdges += addPrepOf();
    numOfEdges += addNN();

    return numOfEdges;
  }

  /**
   * collection > of > member
   */
  private static final DependencyGraphPattern prepOfPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> isKeyWord(v.getLemma()))
      // member
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"));

  private int addPrepOf() {
    return add(prepOfPattern, 1, 0, DependencyName.MEMBER_COLLECTION, "prep");
  }

  /**
   * !> isa > member < nn < collection
   */
  private final static DependencyGraphPattern nnPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> isKeyWord(v.getLemma()))
      // member
      .addVertex(1, v -> v.getNer().isProtein())
      .addEdge(0, 1, Dependency.NnPredicate);

  private int addNN() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = nnPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex collection = matcher.get(0);
      DependencyGraphVertex member = matcher.get(1);
      if (!graph.hasIncomingEdgesOf(member, Dependency.ofPredicate(DependencyName.IS_A))) {
        edgeIndex = addIfNotExist(
            member,
            collection,
            DependencyName.MEMBER_COLLECTION,
            "nn",
            edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * <pre>
   *               m-c
   * +-----------------------------+
   * v                             |
   * N > vmod > containing > dobj > N
   * </pre>
   */
  private static final DependencyGraphPattern containingPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NNS)
      // containing
      .addVertex(1, v -> v.getWord().equals("containing"))
      // member
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addContaining() {
    return add(containingPattern, 2, 0, DependencyName.MEMBER_COLLECTION, "contain");
  }

  /**
   * collection(not nns) < nsubj < contain > dobj > member
   */
  private static final DependencyGraphPattern containsPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NNS)
      // contains
      .addVertex(1, v -> "contain".equals(v.getLemma()) || "span".equals(v.getLemma()))
      // member
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addContains() {
    return add(containsPattern, 2, 0, DependencyName.MEMBER_COLLECTION, "contain");
  }

  private int addNnsConj() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;

    Predicate<DependencyGraphVertex> pred = v -> v.getTag() == PartOfSpeech.NNS;

    for (DependencyGraphVertex v : graph.vertexSet(pred)) {
      List<DependencyGraphVertex> conjunts = getDeps(graph.outgoingEdgesOf(v));
      if (conjunts.size() < 2) {
        continue;
      }
      DependencyGraphVertex n0 = conjunts.get(0);
      boolean isN0Conjunt = false;
      for (int i = 1; i < conjunts.size(); i++) {
        DependencyGraphVertex n1 = conjunts.get(i);
        if (containsConj(n0, n1)) {
          edgeIndex = addIfNotExist(n1, v, DependencyName.MEMBER_COLLECTION, "nns", edgeIndex);
          isN0Conjunt = true;
        }
      }
      if (isN0Conjunt) {
        edgeIndex = addIfNotExist(n0, v, DependencyName.MEMBER_COLLECTION, "nns", edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  private boolean
      containsConj(DependencyGraphVertex s, DependencyGraphVertex t) {
    return graph
        .getAllEdges(s, t)
        .stream()
        .anyMatch(edge -> edge.getDependency().getName() == DependencyName.CONJ);
  }

  private List<DependencyGraphVertex> getDeps(
      Collection<DependencyGraphEdge> edges) {
    return edges
        .stream()
        .filter(
            e -> e.getDependency().getName() == DependencyName.DEP
                && graph.getEdgeTarget(e).getTag().isNoun())
        .map(e -> graph.getEdgeTarget(e))
        .sorted(
            (v1, v2) -> Integer.compare(
                v1.getBeginPosition(),
                v2.getBeginPosition()))
        .collect(Collectors.toList());
  }

  /**
   * <pre>
   *               m-c
   * +-----------------------------+
   * v                             |
   * N > vmod > including > dobj > N
   * </pre>
   */
  private static final DependencyGraphPattern vmodIncludingPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> true)
      // including
      .addVertex(1, v -> v.getWord().equals("including"))
      // member
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addVmodIncluding() {
    return add(vmodIncludingPattern, 2, 0, DependencyName.MEMBER_COLLECTION, "include");
  }

  /**
   * <pre>
   *               m-c
   * +-----------------------------+
   * v                             |
   * N < nsubj < includes > dobj > N
   * </pre>
   */
  private static final DependencyGraphPattern verbIncludePattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> true)
      // include
      .addVertex(1, v -> "include".equals(v.getLemma()))
      // member
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addVerbInclude() {
    return add(verbIncludePattern, 2, 0, DependencyName.MEMBER_COLLECTION, "include");
  }

  /**
   * <pre>
   *          m-c
   * +----------------------------+
   * v                            |
   * N > prep_such_as/including > N
   * </pre>
   */
  private static final DependencyGraphPattern prepIncludingPattern = DependencyGraphPattern
      .of()
      // collection
      .addVertex(0, v -> true)
      // member
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.PREP, "including", "such_as"));

  private int addPrepIncluding() {
    return add(prepIncludingPattern, 1, 0, DependencyName.MEMBER_COLLECTION, "including");
  }
}
