package com.pengyifan.nlp.ssdg;

import java.util.Set;
import java.util.function.Predicate;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.nlp.ling.verb.VerbUtils;

class NullArgumentLinkAdder extends EdgeAdder {

  private static final Set<String> criticalKeywords = Sets.newHashSet(
      "critical",
      "crucial",
      "essential",
      "important",
      "necessary"
  );

  private static final Predicate<DependencyGraphVertex> vionVbgVingPredicate =
      v -> (v.getTag().isNoun() && VerbUtils.isVion(v.getLemma()))
          || v.getTag() == PartOfSpeech.VBG
          || (v.getTag().isNoun() && v.getWord().endsWith("ing"));

  private static final Predicate<DependencyGraphVertex> criticalPredicate =
      v -> criticalKeywords.contains(v.getLemma());

  public NullArgumentLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    numOfEdges += addProViaSubj();
    numOfEdges += addProViaSubjpass();
    numOfEdges += addProBy();
    numOfEdges += addInvolved();
    numOfEdges += addInvolving();
    numOfEdges += addInvolving2();
    numOfEdges += addInvolve();
    numOfEdges += addParticipate();
    numOfEdges += addIsCritical();
    numOfEdges += addCritical();
    numOfEdges += addIsRequired();
    numOfEdges += addDespite();
    numOfEdges += addOccur();
    return numOfEdges;
  }

  /**
   * <pre>
   *            arg1
   * +---------------------------------+
   * |                  prep_via       |
   * |            +------------------+ |
   * v            |                  v |
   * N < nsubj < VB > dobj > N via N-ion
   * </pre>
   */
  private static final DependencyGraphPattern proViaSubjPattern = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // vb
      .addVertex(1, v -> v.getTag().isActiveVerb())
          // dobj
      .addVertex(2, v -> true)
          // ion
      .addVertex(3, v -> v.getWord().endsWith("ion"))
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.PREP, "via", "through"));

  private int addProViaSubj() {
    return add(proViaSubjPattern, 3, 0, DependencyName.ARG1, "null_via");
  }

  /**
   * <pre>
   *            arg1
   * +-------------------------------------+
   * |                      prep_via       |
   * |                +------------------+ |
   * v                |                  v |
   * N < nsubjpass < VBN > agent > N via N-ion
   * </pre>
   */
  private static final DependencyGraphPattern proViaSubjpassPattern = DependencyGraphPattern
      .of()
          // nsubjpass
      .addVertex(0, v -> true)
          // vbn
      .addVertex(1, v -> v.getTag().isPassiveVerb())
          // agent
      .addVertex(2, v -> true)
          // ion
      .addVertex(3, v -> v.getWord().endsWith("ion"))
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.AGENT))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.PREP, "via", "through"));

  private int addProViaSubjpass() {
    return add(proViaSubjpassPattern, 3, 0, DependencyName.ARG1, "null_via");
  }

  /**
   * <pre>
   *            arg0
   * +------------------------------+
   * |                  prep_by     |
   * |            +---------------+ |
   * v            |               v |
   * N < nsubj < VB > dobj > N by VBG
   * </pre>
   */
  private static final DependencyGraphPattern proByPattern = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // vb
      .addVertex(1, v -> v.getTag().isActiveVerb() || v.getLemma().equals("function"))
          // dobj
      .addVertex(2, v -> true)
          // vbg
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VBG)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ))
      .addEdge(1, 3, Dependency.ofPrepPredicate("by"));
  /**
   * <pre>
   *            arg0
   * +------------------------------+
   * |                  prep_by     |
   * |            +---------------+ |
   * v            |               v |
   * N < nsubj < VB > dobj > N by VBG
   * </pre>
   */
  private static final DependencyGraphPattern proByPattern2 = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // vb
      .addVertex(1, v -> v.getLemma().equals("function"))
          // vbg
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VBG)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 3, Dependency.ofPrepPredicate("by"));
  private int addProBy() {
    int numOfEdges = 0;
    numOfEdges += add(proByPattern, 3, 0, DependencyName.ARG0, "null_by");
    numOfEdges += add(proByPattern2, 3, 0, DependencyName.ARG0, "null_by");
    return numOfEdges;
  }

  /**
   * <pre>
   *    amod    prep_for
   * +-------+ +---------+
   * |       v |         v
   * N    critical for ving.n
   * ^                   |
   * +-------------------+
   * arg0
   */
  private static final DependencyGraphPattern criticalPattern = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // critical
      .addVertex(1, criticalPredicate)
          // ving
      .addVertex(2, vionVbgVingPredicate)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.AMOD))
      .addEdge(1, 2, Dependency.ofPrepPredicate("for"));

  private int addCritical() {
    return add(criticalPattern, 2, 0, DependencyName.ARG0, "null_critical");
  }

  /**
   * <pre>
   *   nsubj
   * +--------+
   * |    cop |  prep_for
   * |  +----+|+---------+
   * v  v    |||         v
   * N is critical for ving.n
   * ^                   |
   * +-------------------+
   * arg0
   */
  private static final DependencyGraphPattern isCriticalPattern = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // be
      .addVertex(1, v -> true)
          // critical
      .addVertex(2, criticalPredicate)
          // ving
      .addVertex(3, vionVbgVingPredicate)
      .addEdge(2, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.COP))
      .addEdge(2, 3, Dependency.ofPrepPredicate("for", "in"));

  private int addIsCritical() {
    return add(isCriticalPattern, 3, 0, DependencyName.ARG0, "null_critical");
  }

  /**
   * <pre>
   *   nsubjpass
   * +------------+
   * |    auxpass |  prep_for
   * |  +--------+|+---------+
   * v  v        |||         v
   * N is     required for ving.n
   * ^                       |
   * +-----------------------+
   * arg0
   */
  private static final DependencyGraphPattern isRequiredPattern = DependencyGraphPattern
      .of()
          // nsubjpass
      .addVertex(0, v -> true)
          // be
      .addVertex(1, v -> true)
          // required
      .addVertex(2, v -> v.getWord().equals("required"))
          // ving
      .addVertex(3, vionVbgVingPredicate)
      .addEdge(2, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.AUXPASS))
      .addEdge(2, 3, Dependency.ofPrepPredicate("for"));

  private int addIsRequired() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = isRequiredPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex s = matcher.get(3);
      DependencyGraphVertex t = matcher.get(0);
      t = changeHead(t);
      edgeIndex = addIfNotExist(s, t, DependencyName.ARG0, "required", edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  private DependencyGraphVertex changeHead(DependencyGraphVertex v) {
    if (v.getNer().isProtein()) {
      return v;
    }
    int numberOfProtein = 0;
    DependencyGraphVertex protein = null;
    for (DependencyGraphVertex n : graph.outgoingEdgeTargetsOf(
        v,
        Dependency.ofPredicate(DependencyName.PREP, "of"))) {
      if (n.getNer().isProtein()) {
        protein = n;
        numberOfProtein++;
      }
    }
    if (numberOfProtein == 1) {
      return protein;
    }
    return v;
  }

  /**
   * <pre>
   *  nsubjpass   prep/prepc_in
   * +---------+ +------------+
   * v         | |            v
   * N  is  involved in VBG/Ving.n/Vion.n
   * ^                        |
   * +------------------------+
   *          arg0
   * </pre>
   */
  private static final DependencyGraphPattern involvedPattern1 = DependencyGraphPattern
      .of()
          // nsubjpass
      .addVertex(0, v -> true)
          // involve
      .addVertex(1, v -> v.getLemma().equals("involve"))
          // ving
      .addVertex(2, vionVbgVingPredicate)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 2, Dependency.ofPrepPredicate("in"));
  /**
   * <pre>
   *    vmod    prep_for
   * +-------+ +---------+
   * |       v |         v
   * N    involved for ving.n
   * ^                   |
   * +-------------------+
   *         arg0
   * </pre>
   */
  private static final DependencyGraphPattern involvedPattern2 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // involve
      .addVertex(1, v -> v.getLemma().equals("involve"))
          // ving
      .addVertex(2, vionVbgVingPredicate)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPrepPredicate("in"));

  private int addInvolved() {
    int n = 0;
    n += add(involvedPattern1, 2, 0, DependencyName.ARG0, "null_involve");
    n += add(involvedPattern2, 2, 0, DependencyName.ARG0, "null_involve");
    return n;
  }

  /**
   * <pre>
   *    nsubj     dobj
   * +---------+ +------------+
   * v         | |            v
   * N    involve/show VBG/Ving.n/Vion.n
   * ^                        |
   * +------------------------+
   *          arg0
   * </pre>
   */
  private static final DependencyGraphPattern involvePattern = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // required
      .addVertex(1, v -> v.getLemma().equals("involve")
          || v.getLemma().equals("show")
          || v.getLemma().equals("mediate")
          || v.getLemma().equals("require"))
          // ving
      .addVertex(2, vionVbgVingPredicate)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addInvolve() {
    return add(involvePattern, 2, 0, DependencyName.ARG0, "null_involve");
  }

  /**
   * <pre>
   *  nsubj     prep_in
   * +---------+ +------------+
   * v         | |            v
   * N    participate in VBG/Ving.n/Vion.n
   * ^                        |
   * +------------------------+
   *          arg0
   * </pre>
   */
  private static final DependencyGraphPattern participatePattern = DependencyGraphPattern
      .of()
          // nsubj
      .addVertex(0, v -> true)
          // participate
      .addVertex(1, v -> v.getLemma().equals("participate"))
          // ving
      .addVertex(2, vionVbgVingPredicate)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPrepPredicate("in"));

  private int addParticipate() {
    return add(participatePattern, 2, 0, DependencyName.ARG0, "null_participate");
  }

  /**
   * <pre>
   *     vmod         dobj
   * +---------+ +--------+
   * |         v |        v
   * Vion involving       N
   * |                    ^
   * +--------------------+
   *          arg1
   * </pre>
   */
  private static final DependencyGraphPattern involvingPattern = DependencyGraphPattern
      .of()
          // ving
      .addVertex(0, vionVbgVingPredicate)
          // involving
      .addVertex(1, v -> v.getWord().equals("involving"))
          // ving
      .addVertex(2, v -> v.getTag().isNoun())
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addInvolving() {
    return add(involvingPattern, 0, 2, DependencyName.ARG1, "null_involve");
  }

  /**
   * VBN by N1 involving N2
   */
  private static final DependencyGraphPattern involving2Pattern = DependencyGraphPattern
      .of()
          // vbn
      .addVertex(0, v -> v.getTag() == PartOfSpeech.VBN)
          // agent
      .addVertex(1, v -> true)
          // involving
      .addVertex(2, v -> v.getWord().equals("involving"))
          // dobj
      .addVertex(3, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.AGENT))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(2, 3, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addInvolving2() {
    return add(involving2Pattern, 0, 3, DependencyName.ARG1, "null_involve");
  }

  /**
   * Despite vbg, n
   */
  private static final DependencyGraphPattern despitePattern = DependencyGraphPattern
      .of()
          // vbg
      .addVertex(0, vionVbgVingPredicate)
          // nsubj
      .addVertex(1, v -> true)
          // v
      .addVertex(2, v -> true)
      .addEdge(2, 0, Dependency.ofPrepPredicate("despite"))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.NSUBJ));

  private int addDespite() {
    return add(despitePattern, 0, 1, DependencyName.ARG0, "null_despite");
  }

  /**
   * vbg occur n
   */
  private static final DependencyGraphPattern occurPattern = DependencyGraphPattern
      .of()
          // vbg
      .addVertex(0, vionVbgVingPredicate)
          // occur
      .addVertex(1, v -> v.getLemma().equals("occur"))
          // v
      .addVertex(2, v -> v.getTag().isNoun())
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPrepPredicate("in", "between"));

  private int addOccur() {
    return add(occurPattern, 0, 2, DependencyName.ARG0, "null_occur");
  }
}
