package com.pengyifan.nlp.ssdg;

import java.util.Set;
import java.util.function.Predicate;

import com.google.common.collect.Iterables;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;

class ParseError {

  private final DependencyGraph graph;
  private final String text;

  ParseError(String text, DependencyGraph graph) {
    this.graph = graph;
    this.text = text;
  }

  int add() {
    changeAssociated();

    int numOfEdge = 0;
    numOfEdge += addBinding();
    numOfEdge += propagateBetween();
    numOfEdge += addShowBind();
    numOfEdge += addBecomeAssociated();
    numOfEdge += addLigand();
    numOfEdge += addNsubjVBN();
    changeDep();
    return numOfEdge;
  }

  private void changeDep() {

    final Predicate<DependencyGraphEdge> predicate =
        e -> e.getDependency().getName() == DependencyName.DEP;

    for (DependencyGraphEdge edge : graph.edgeSet(predicate)) {
      DependencyGraphVertex source = graph.getEdgeSource(edge);
      DependencyGraphVertex target = graph.getEdgeTarget(edge);
      if (source.getEndPosition() >= target.getBeginPosition()) {
        continue;
      }
      String subtext = text.substring(source.getEndPosition(), target.getBeginPosition());
      if (subtext.trim().equals("(")) {
        edge.getDependency().setName(DependencyName.APPOS);
      }
    }
  }

  /**
   * N become associated
   */
  private int addBecomeAssociated() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;

    final Predicate<DependencyGraphEdge> predicate =
        e -> e.getDependency().getName() == DependencyName.NSUBJ;

    for (DependencyGraphEdge edge : graph.edgeSet(predicate)) {
      DependencyGraphVertex become = graph.getEdgeSource(edge);
      if (become.getLemma().equals("become")) {
        DependencyGraphVertex n = graph.getEdgeTarget(edge);
        for (DependencyGraphVertex vbn : graph.outgoingEdgeTargetsOf(
            become,
            Dependency.ofPredicate(DependencyName.ACOMP))) {
          if (vbn.getTag() == PartOfSpeech.VBN) {
            edgeIndex = addIfNotExsist(vbn, n, DependencyName.NSUBJPASS, null, edgeIndex);
          }
        }
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * N associated.vbn
   */
  private void changeAssociated() {
    final Predicate<DependencyGraphEdge> predicate =
        e -> e.getDependency().getName() == DependencyName.NSUBJ;

    for (DependencyGraphEdge edge : graph.edgeSet(predicate)) {
      DependencyGraphVertex v = graph.getEdgeSource(edge);
      if (v.getTag() == PartOfSpeech.VBN && v.getLemma().equals("associate")) {
        v.setTag(PartOfSpeech.VBD);
      }
    }
  }

  private int addNsubjVBN () {
    final Predicate<DependencyGraphEdge> predicate =
        e -> e.getDependency().getName() == DependencyName.NSUBJ;

    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge edge : graph.edgeSet(predicate)) {
      DependencyGraphVertex v = graph.getEdgeSource(edge);
      if (v.getTag() == PartOfSpeech.VBN
          && !graph.hasOutgoingEdgesOf(v, Dependency.ofPredicate(DependencyName.AUX))) {
        addIfNotExsist(graph.getEdgeTarget(edge), v, DependencyName.VMOD, null, edgeIndex++);
        graph.removeEdge(edge);
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * N is a ligand
   */
  private int addLigand() {
    for (DependencyGraphVertex v : graph.vertexSet()) {
      if (!v.getLemma().equals("ligand")) {
        continue;
      }
      Set<DependencyGraphEdge> auxpasses = graph.outgoingEdgesOf(
          v,
          Dependency.ofPredicate(DependencyName.AUXPASS));
      if (auxpasses.size() != 1) {
        continue;
      }
      boolean hasLigand = false;
      for (DependencyGraphEdge nsubjpass : graph.outgoingEdgesOf(
          v,
          Dependency.ofPredicate(DependencyName.NSUBJPASS))) {
        nsubjpass.getDependency().setName(DependencyName.NSUBJ);
        hasLigand = true;
      }
      if (hasLigand) {
        DependencyGraphEdge auxpass = Iterables.getOnlyElement(auxpasses);
        auxpass.getDependency().setName(DependencyName.COP);
      }
    }
    return 0;
  }

  private int propagateBetween() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge edge : graph.edgeSet()) {
      if (edge.getDependency().equals(DependencyName.PREP, "between")) {
        DependencyGraphVertex source = graph.getEdgeSource(edge);
        DependencyGraphVertex n1 = graph.getEdgeTarget(edge);
        for (DependencyGraphVertex n2 : graph.outgoingEdgeTargetsOf(
            n1,
            Dependency.ofPredicate(DependencyName.CONJ))) {
          edgeIndex = addIfNotExsist(source, n2, DependencyName.PREP, "between", edgeIndex);
        }
      }
    }
    for (DependencyGraphVertex v : graph.vertexSet()) {
      if (!v.getWord().equals("binding")) {
        continue;
      }
      // nn
      DependencyGraphVertex nn = null;
      for (DependencyGraphEdge edge : graph.outgoingEdgesOf(v)) {
        DependencyGraphVertex target = graph.getEdgeTarget(edge);
        if (target.getNer().isProtein()) {
          nn = target;
          break;
        }
      }
      if (nn == null) {
        continue;
      }

      // conj
      for (DependencyGraphEdge edge : graph.outgoingEdgesOf(v)) {
        DependencyGraphVertex target = graph.getEdgeTarget(edge);
        if (edge.getDependency().getName() == DependencyName.CONJ) {
          edgeIndex = addIfNotExsistSemantic(v, target, DependencyName.ARG1, "DOBJ", edgeIndex);
          edgeIndex = addIfNotExsistSemantic(v, nn, DependencyName.ARG0, "nn", edgeIndex);
        }
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * show N binds to
   */
  private int addShowBind() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphVertex v : graph.vertexSet()) {
      if (!v.getLemma().equals("show")) {
        continue;
      }
      // dobj
      Set<DependencyGraphVertex> dobjs = graph.outgoingEdgeTargetsOf(
          v,
          Dependency.ofPredicate(DependencyName.DOBJ));
      if (dobjs.isEmpty()) {
        continue;
      }
      // vb
      Set<DependencyGraphVertex> deps = graph.outgoingEdgeTargetsOf(
          v,
          Dependency.ofPredicate(DependencyName.DEP));
      if (deps.isEmpty()) {
        continue;
      }
      for (DependencyGraphVertex dep : deps) {
        if (dep.getTag().isActiveVerb()) {
          for (DependencyGraphVertex dobj : dobjs) {
            if (dep.getBeginPosition() > dobj.getBeginPosition()) {
              edgeIndex = addIfNotExsistSemantic(
                  dep,
                  dobj,
                  DependencyName.ARG0,
                  "parsing-error",
                  edgeIndex);
            }
          }
        }
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * N binding: N
   */
  private int addBinding() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphVertex v : graph.vertexSet()) {
      if (!v.getWord().equals("binding")) {
        continue;
      }
      // nn
      DependencyGraphVertex nn = null;
      for (DependencyGraphEdge edge : graph.outgoingEdgesOf(v)) {
        DependencyGraphVertex target = graph.getEdgeTarget(edge);
        if (target.getNer().isProtein()) {
          nn = target;
          break;
        }
      }
      if (nn == null) {
        continue;
      }

      // conj
      for (DependencyGraphEdge edge : graph.outgoingEdgesOf(v)) {
        DependencyGraphVertex target = graph.getEdgeTarget(edge);
        if (edge.getDependency().getName() == DependencyName.CONJ) {
          edgeIndex = addIfNotExsistSemantic(v, target, DependencyName.ARG1, "DOBJ", edgeIndex);
          edgeIndex = addIfNotExsistSemantic(v, nn, DependencyName.ARG0, "nn", edgeIndex);
        }
      }
    }
    return edgeIndex - startIndex;
  }

  private int addIfNotExsist(DependencyGraphVertex source,
      DependencyGraphVertex target,
      DependencyName name,
      String specific,
      int edgeIndex) {
    if (specific == null
        && !graph.containsEdge(source, target, Dependency.ofPredicate(name))) {
      graph.addEdge(source, target, createEdge(edgeIndex, name, null));
      edgeIndex++;
    } else if (specific != null
        && !graph.containsEdge(source, target, Dependency.ofPredicate(name, specific))) {
      graph.addEdge(source, target, createEdge(edgeIndex, name, specific));
      edgeIndex++;
    }
    return edgeIndex;
  }

  private int addIfNotExsistSemantic(DependencyGraphVertex source,
      DependencyGraphVertex target,
      DependencyName name,
      String specific,
      int edgeIndex) {
    if (!graph.containsEdge(source, target, Dependency.ofPredicate(name, specific))) {
      graph.addEdge(source, target, createEdge(edgeIndex, name, specific));
      edgeIndex++;
    }
    return edgeIndex;
  }

  private DependencyGraphEdge createEdge(int index, DependencyName name,
      String specific) {
    return DependencyGraphEdge.of(index, new Dependency(name, specific));
  }
}
