package com.pengyifan.nlp.ssdg;

import java.util.Set;
import java.util.function.Predicate;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;

public class PartWholeLinkAdder extends EdgeAdder {

  public static final Predicate<DependencyGraphVertex> complexPredicate =

      new Predicate<DependencyGraphVertex>() {

        private final Set<String> complexKeywords = Sets.newHashSet(
            "complex",
            "dimer",
            "heterodimer",
            "heteromer",
            "homodimer",
            "homomer",
            "homotrimer",
            "oligomer",
            "trimer"
        );

        @Override
        public boolean test(DependencyGraphVertex t) {
          return complexKeywords.contains(t.getLemma());
        }
      };

  public PartWholeLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    for (int i = 0; i < 2; i++) {
      numOfEdges += addPrepOf();
      numOfEdges += addNN();
      numOfEdges += addHas();
      numOfEdges += addContaining();
      numOfEdges += addContains();
      numOfEdges += addComposedOf();

//      numOfEdges += addComplexOf();
//      numOfEdges += addNNComplex();
//      numOfEdges += addFormComplex();
//      numOfEdges += addFormedComplex();
//      numOfEdges += addFormationComplex();
    }
    return numOfEdges;
  }

  private final static Set<String> keywords = Sets.newHashSet(
      //"aa",
      //"acid",
      "chain",
      //"cleft",
      // "construct",
      "domain",
      "ectodomain",
      //"epitope",
      "fraction",
      "fragment",
      "homeodomain",
      //      "homo-oligomer",
      "locus",
      "loop",
      "motif",
      "part",
      //      "peptide",
      "phosphotyrosine",
      //"portion",
      "region",
      //"repeat",
      "residue",
      "segment",
      "sequence",
      "site",
      //"substrate",
      "subsequence",
      //"subunit",
      "surface",
      "tail",
      "terminal",
      "termini",
      "terminus"
      //      "third"
  );
  private final static Set<String> prefixKeywords = Sets.newHashSet(
      "tyr"
  );
  private final static Set<String> suffixKeywords = Sets.newHashSet(
      "domain",
      "terminus"
  );

  private static boolean isKeyWord(String keyword, Set<String> keywords) {
    return keywords.contains(keyword);
  }

  private static boolean isPrefixKeyWord(String keyword) {
    for (String s : prefixKeywords) {
      if (keyword.startsWith(s)) {
        return true;
      }
    }
    return false;
  }

  private static boolean isSuffixKeyWord(String keyword) {
    for (String s : suffixKeywords) {
      if (keyword.endsWith(s)) {
        return true;
      }
    }
    return false;
  }

  /**
   * part > of > whole
   */
  private final static DependencyGraphPattern prepOfPattern = DependencyGraphPattern
      .of()
          // part
      .addVertex(0, v -> isKeyWord(v.getLemma(), keywords)
          || isPrefixKeyWord(v.getLemma())
          || isSuffixKeyWord(v.getLemma()))
          // whole
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of", "from", "in", "on"));

  private int addPrepOf() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = prepOfPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex part = matcher.get(0);
      DependencyGraphVertex whole = matcher.get(1);
      if (!whole.getNer().isProtein()
          && !isKeyWord(whole.getLemma(), keywords)
          && !isPrefixKeyWord(whole.getLemma())
          && !isSuffixKeyWord(whole.getLemma())
          && !graph.hasOutgoingEdgesOf(whole, Dependency.ofPredicate(DependencyName.PART_WHOLE))
          && !graph.hasOutgoingEdgesOf(whole, Dependency.ofPredicate(DependencyName.DESCRIPTIVE))) {
        continue;
      }
      edgeIndex = addIfNotExist(part, whole, DependencyName.PART_WHOLE, "prep", edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  /**
   * complex > of > part
   */
  private final static DependencyGraphPattern complexOfPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, complexPredicate)
          // part
      .addVertex(1, v -> v.getNer().isProtein())
      .addEdge(0, 1, Dependency.ofPrepPredicate("of").or(Dependency.ofPrepPredicate("with")));

  private int addComplexOf() {
    return add(complexOfPattern, 1, 0, DependencyName.PART_WHOLE, "of_complex");
  }

  /**
   * whole < nn < part
   */
  private final static DependencyGraphPattern nnPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, v -> v.getNer().isProtein())
          // part
      .addVertex(1, v -> isKeyWord(v.getLemma(), keywords))
      .addEdge(1, 0, Dependency.NnPredicate);

  private int addNN() {
    return add(nnPattern, 1, 0, DependencyName.PART_WHOLE, "nn");
  }

  /**
   * part < nn < complex
   */
  private final static DependencyGraphPattern nnComplexPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, complexPredicate)
          // part
      .addVertex(1, v -> v.getNer().isProtein())
      .addEdge(0, 1, Dependency.NnPredicate);

  private int addNNComplex() {
    return add(nnComplexPattern, 1, 0, DependencyName.PART_WHOLE, "nn_complex");
  }

  /**
   * part < nsubj < form > dobj > complex
   */
  private final static DependencyGraphPattern formComplexPattern = DependencyGraphPattern
      .of()
          // part
      .addVertex(0, v -> true)
          // form
      .addVertex(1, v -> v.getLemma().equals("form"))
          // whole
      .addVertex(2, complexPredicate)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addFormComplex() {
    return add(formComplexPattern, 0, 2, DependencyName.PART_WHOLE, "form_complex");
  }

  /**
   * complex < nsubjpass < form > by > part
   */
  private final static DependencyGraphPattern formedComplexPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, complexPredicate)
          // form
      .addVertex(1, v -> v.getLemma().equals("form"))
          // part
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 2, Dependency.ofPrepPredicate("by", "between"));

  private int addFormedComplex() {
    return add(formedComplexPattern, 2, 0, DependencyName.PART_WHOLE, "formed_complex");
  }

  /**
   * complex formation of part
   */
  private final static DependencyGraphPattern formationComplexPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, complexPredicate)
          // formation
      .addVertex(1, v -> v.getLemma().equals("formation"))
          // part
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NN))
      .addEdge(1, 2, Dependency.ofPrepPredicate("of", "with"));

  private int addFormationComplex() {
    return add(formationComplexPattern, 2, 0, DependencyName.PART_WHOLE, "formation_complex");
  }

  /**
   * whole > vmod > containing > dobj > part
   */
  private final static DependencyGraphPattern containingPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NN || complexPredicate.test(v))
          // containing
      .addVertex(1, v -> "containing".equals(v.getWord()))
          // part
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addContaining() {
    return add(containingPattern, 2, 0, DependencyName.PART_WHOLE, "contain");
  }

  /**
   * whole < nsubj < contains > dobj > part
   */
  private final static DependencyGraphPattern containsPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NN)
          // containing
      .addVertex(1, v -> "contain".equals(v.getLemma()))
          // part
      .addVertex(2, v -> true)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addContains() {
    return add(containsPattern, 2, 0, DependencyName.PART_WHOLE, "contain");
  }

  /**
   * whole < nsubj < has > dobj > part
   */
  private final static DependencyGraphPattern hasPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NN)
          // has
      .addVertex(1, v -> "have".equals(v.getLemma()))
          // part
      .addVertex(2, v -> isKeyWord(v.getLemma(), keywords))
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJ))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.DOBJ));

  private int addHas() {
    return add(hasPattern, 2, 0, DependencyName.PART_WHOLE, "have");
  }

  /**
   * <pre>
   *                 part-whole
   * +-------------------------------------+
   * |      nsubjpass              prep_of |
   * ++----------------------+    +-------+|
   * vv                      |    |       v|
   * whole is < auxpass < composed   of  part
   * </pre>
   */
  private final static DependencyGraphPattern composedOfPattern = DependencyGraphPattern
      .of()
          // whole
      .addVertex(0, v -> v.getTag() == PartOfSpeech.NN)
          // aux
      .addVertex(1, v -> v.getTag() == PartOfSpeech.NN)
          // composed
      .addVertex(2, v -> "composed".equals(v.getWord()))
          // part
      .addVertex(3, v -> true)
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.AUXPASS))
      .addEdge(2, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(2, 3, Dependency.ofPrepPredicate("of"));

  private int addComposedOf() {
    return add(composedOfPattern, 3, 0, DependencyName.PART_WHOLE, "composed");
  }
}
