package com.pengyifan.nlp.ssdg;

import java.util.Set;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;

class ReceptorAdder extends EdgeAdder {

  private static final Set<String> keywords = Sets.newHashSet(
      "agonist",
      "activator",
//      "coactivator",
//      "co-receptor",
      "inhibitor",
      "ligand",
      "substrate",
      "receptor"
      );

  private static boolean isKeyWord(String keyword) {
    for (String s : keywords) {
      if (keyword.endsWith(s)) {
        return true;
      }
    }
    return false;
  }

  public ReceptorAdder(DependencyGraph graph) {
    super(graph);
  }

  int add() {
    int numOfEdges = 0;
    numOfEdges += addReceptorProtein();
    numOfEdges += addNNReceptor();
    numOfEdges += addReceptorFor();
    return numOfEdges;
  }

  /**
   * <pre>
   *   nn/nmod
   * +---------+
   * v         |
   * N      receptor
   * ^      | |    ^
   * +------+ +----+
   *   arg0    arg1
   * </pre>
   */
  private static final DependencyGraphPattern nnReceptorPattern1 = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> v.getNer().isProtein())
      // receptor
      .addVertex(1, v -> isKeyWord(v.getLemma()))
      .addEdge(1, 0, Dependency.NnPredicate);
  private static final DependencyGraphPattern nnReceptorPattern2 = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> v.getNer().isProtein())
      // receptor
      .addVertex(1, v -> isKeyWord(v.getLemma()))
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.IS_A));
  
  private int addNNReceptor() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    
    DependencyGraphMatcher matcher = nnReceptorPattern1.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex receptor = matcher.get(1);
      DependencyGraphVertex protein = matcher.get(0);
      if (protein.getBeginPosition() < receptor.getBeginPosition()) {
        edgeIndex = addIfNotExist(receptor, protein, DependencyName.ARG0, "receptor", edgeIndex);
        edgeIndex = addIfNotExist(receptor, DependencyName.ARG1, "receptor", edgeIndex);
      }
    }
    
    matcher = nnReceptorPattern2.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex receptor = matcher.get(1);
      DependencyGraphVertex protein = matcher.get(0);
      if (protein.getBeginPosition() < receptor.getBeginPosition()) {
        edgeIndex = addIfNotExist(receptor, protein, DependencyName.ARG0, "receptor", edgeIndex);
        edgeIndex = addIfNotExist(receptor, DependencyName.ARG1, "receptor", edgeIndex);
      }
    }
    
    return edgeIndex - startIndex;
  }

  /**
   * receptor > prep_for > n
   */
  private static final DependencyGraphPattern receptorForPattern1 = DependencyGraphPattern
      .of()
      // receptor
      .addVertex(0, v -> isKeyWord(v.getLemma()))
      // n
      .addVertex(1, v -> v.getNer().isProtein())
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.PREP, "for"));
  /**
   * n < isa < receptor > prep_for > n
   */
  private static final DependencyGraphPattern receptorForPattern2 = DependencyGraphPattern
      .of()
      // n1
      .addVertex(0, v -> v.getNer().isProtein())
      // receptor
      .addVertex(1, v -> isKeyWord(v.getLemma()))
          // n
      .addVertex(2, v -> v.getNer().isProtein())
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.IS_A))
      .addEdge(1, 2, Dependency.ofPredicate(DependencyName.PREP, "for"));
  private int addReceptorFor() {
    int numOfEdges = 0;
    numOfEdges += add(receptorForPattern1, 0, 1, DependencyName.ARG1, "receptor");
    numOfEdges += add(receptorForPattern2, 1, 0, DependencyName.ARG0, "is-a");
    return numOfEdges;
  }

  /**
   * receptor < nn < protein
   */
  private static final DependencyGraphPattern receptorProteinPattern1 = DependencyGraphPattern
      .of()
      // receptor
      .addVertex(0, v -> isKeyWord(v.getLemma()))
      // n
      .addVertex(1, v -> v.getNer().isProtein())
      .addEdge(1, 0, Dependency.NnPredicate);
  /**
   * protein < nn < receptor < nn < protein
   */
  private static final DependencyGraphPattern receptorProteinPattern2 = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> v.getNer().isProtein())
      // receptor
      .addVertex(1, v -> isKeyWord(v.getLemma()))
      // n
      .addVertex(2, v -> v.getNer().isProtein())
      .addEdge(1, 0, Dependency.NnPredicate)
      .addEdge(1, 2, Dependency.NnPredicate);
  /**
   * n < poss < receptor < nn < protein
   */
  private static final DependencyGraphPattern receptorProteinPattern3 = DependencyGraphPattern
      .of()
      // n
      .addVertex(0, v -> v.getNer().isProtein())
      // receptor
      .addVertex(1, v -> isKeyWord(v.getLemma()))
      // n
      .addVertex(2, v -> v.getNer().isProtein())
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.POSS))
      .addEdge(1, 2, Dependency.NnPredicate);
  
  private int addReceptorProtein() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    
    DependencyGraphMatcher matcher = receptorProteinPattern1.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex protein = matcher.get(1);
      DependencyGraphVertex receptor = matcher.get(0);
      if (protein.getBeginPosition() > receptor.getBeginPosition()) {
        edgeIndex = addIfNotExist(receptor, protein, DependencyName.ARG1, "receptor", edgeIndex);
      }
    }
    
    matcher = receptorProteinPattern2.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex receptor = matcher.get(1);
      DependencyGraphVertex protein = matcher.get(0);
      if (protein.getBeginPosition() < receptor.getBeginPosition()) {
        edgeIndex = addIfNotExist(receptor, protein, DependencyName.ARG0, "receptor", edgeIndex);
      }
    }
    
    int n = edgeIndex - startIndex;
    n += add(receptorProteinPattern3, 1, 0, DependencyName.ARG0, "receptor");
    return n;
  }

  private int addIfNotExist(DependencyGraphVertex v,
      DependencyName name,
      String specific,
      int edgeIndex) {
    if (!graph.containsEdge(v, v, Dependency.ofPredicate(name, specific))) {
      graph.addSelfLoop(v, DependencyGraphEdge.of(edgeIndex, new Dependency(name, specific)));
      edgeIndex++;
    }
    return edgeIndex;
  }
}
