package com.pengyifan.nlp.ssdg;

import java.util.Optional;

import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphUtils;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;

class ReferentialLinkPropagation {

  private final DependencyGraph graph;

  ReferentialLinkPropagation(DependencyGraph graph) {
    this.graph = graph;
  }

  int propage() {
    int numOfEdges = 0;
    numOfEdges += propagatePartWhole();
    numOfEdges += propagateMemberCollection();
    numOfEdges += propagateIsA();
    numOfEdges += propagateCoreference();
    // numOfEdges += propagateAbbr();
    // numOfEdges += propagateAppositive();
    numOfEdges += propagateCTermFTerm();

    numOfEdges += propagateAnotherEvent();

    return numOfEdges;
  }

  @SuppressWarnings("unused")
  private int propagateAppositive() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      if (e.getDependency().getName() == DependencyName.IS_A
          && e.getDependency().getSpecific().equals(Optional.of("appos"))) {
        DependencyGraphVertex target = graph.getEdgeTarget(e);
        DependencyGraphVertex source = graph.getEdgeSource(e);
        numOfEdges += propagateFromSourceToTargetTarget(source, target);
        numOfEdges += propagateFromTargetSourceToSource(e, source, target);
        // graph.removeEdge(e);
      }
      // if (e.getDependency().getName() == DependencyName.APPOS) {
      // graph.removeEdge(e);
      // }
    }
    return numOfEdges;
  }

  private int propagateIsA() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      if (e.getDependency().getName() == DependencyName.IS_A) {
        DependencyGraphVertex target = graph.getEdgeTarget(e);
        DependencyGraphVertex source = graph.getEdgeSource(e);
        numOfEdges += propagateFromTargetSourceToSource(e, source, target);
        numOfEdges += propagateFromSourceSourceToTarget(source, target);
      }
    }
    return numOfEdges;
  }

  // private int propagateAbbr() {
  // int numOfEdges = 0;
  // for (DependencyGraphEdge e : graph.edgeSet()) {
  // if (e.getSemanticName() == SemanticName.ABBR) {
  // DependencyGraphVertex target = graph.getEdgeTarget(e);
  // DependencyGraphVertex source = graph.getEdgeSource(e);
  // numOfEdges += propagateFromTargetSourceToSource(
  // e,
  // source,
  // target);
  // }
  // }
  // return numOfEdges;
  // }

  private int propagateCoreference() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      if (e.getDependency().getName() == DependencyName.COREF) {
        DependencyGraphVertex target = graph.getEdgeTarget(e);
        DependencyGraphVertex source = graph.getEdgeSource(e);
        numOfEdges += propagateFromSourceSourceToTarget(source, target);
      }
    }
    return numOfEdges;
  }

  private int propagatePartWhole() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      DependencyGraphVertex target = graph.getEdgeTarget(e);
      DependencyGraphVertex source = graph.getEdgeSource(e);
      if (e.getDependency().getName() == DependencyName.PART_WHOLE) {
        if (PartWholeLinkAdder.complexPredicate.test(target)) {
          numOfEdges += propagateFromTargetSourceToSource(e, source, target);
        } else {
          numOfEdges += propagateFromSourceSourceToTargetWithoutIsA(graph, source, target);
        }
      }
    }
    return numOfEdges;
  }

  /**
   * <pre>
   *     (add)
   * +-----------+
   * |           v
   * a --> b --> c
   * </pre>
   */
  private int propagateFromSourceSourceToTargetWithoutIsA(
      DependencyGraph graph,
      DependencyGraphVertex b,
      DependencyGraphVertex c) {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge inEdgeOfB : graph.incomingEdgesOf(b)) {
      if (inEdgeOfB.getDependency().getName().isSyntactic()) {
        continue;
      }
      if (inEdgeOfB.getDependency().getName() == DependencyName.IS_A) {
        continue;
      }
      DependencyGraphVertex a = graph.getEdgeSource(inEdgeOfB);
      DependencyGraphEdge edge = new DependencyGraphEdge(edgeIndex, inEdgeOfB);
      if (a != c
          && !DependencyGraphUtils.containsEdge(graph, a, c, edge)) {
        graph.addEdge(a, c, edge);
        edgeIndex++;
      }
    }
    return edgeIndex - startIndex;
  }

  private int propagateCTermFTerm() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      if (e.getDependency().getName() == DependencyName.DESCRIPTIVE) {
        DependencyGraphVertex target = graph.getEdgeTarget(e);
        DependencyGraphVertex source = graph.getEdgeSource(e);
        numOfEdges += propagateFromSourceSourceToTarget(source, target);
      }
    }
    return numOfEdges;
  }

  private int propagateAnotherEvent() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      DependencyGraphVertex target = graph.getEdgeTarget(e);
      DependencyGraphVertex source = graph.getEdgeSource(e);
      if (e.getDependency().getName() == DependencyName.ARG0
          && source.getLemma().equals("activity")) {
        numOfEdges += propagateFromSourceSourceToTarget(source, target);
      }
    }
    return numOfEdges;
  }

  private int propagateMemberCollection() {
    int numOfEdges = 0;
    for (DependencyGraphEdge e : graph.edgeSet()) {
      if (e.getDependency().getName() == DependencyName.MEMBER_COLLECTION) {
        DependencyGraphVertex target = graph.getEdgeTarget(e);
        DependencyGraphVertex source = graph.getEdgeSource(e);
        numOfEdges += propagateFromTargetSourceToSource(e, source, target);
      }
    }
    return numOfEdges;
  }

  private int propagateFromSourceSourceToTarget(
      DependencyGraphVertex source,
      DependencyGraphVertex target) {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge incomingEdgeOfSource : graph.incomingEdgesOf(source)) {
      if (incomingEdgeOfSource.getDependency().getName().isSyntactic()) {
        continue;
      }
      if (incomingEdgeOfSource.getDependency().getName().isReferentialLink()) {
        continue;
      }
      DependencyGraphVertex sourceSource = graph.getEdgeSource(incomingEdgeOfSource);
      edgeIndex = addIfNotExsist(sourceSource, target, incomingEdgeOfSource, edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  /**
   * <pre>
   *             (add)
   *   +-------------------------+
   *   |                         |
   *   |                         v
   * source --> target --> target target
   * </pre>
   */
  private int propagateFromSourceToTargetTarget(
      DependencyGraphVertex source,
      DependencyGraphVertex target) {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge outgoingEdgeOfTarget : graph.outgoingEdgesOf(target)) {
      if (outgoingEdgeOfTarget.getDependency().getName().isSyntactic()) {
        continue;
      }
      DependencyGraphVertex targetTarget = graph.getEdgeTarget(outgoingEdgeOfTarget);
      edgeIndex = addIfNotExsist(source, targetTarget, outgoingEdgeOfTarget, edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  /**
   * <pre>
   *       +-----------------------------+
   *       |                             |
   *       |        (add)                v
   * target source -------> source ---> target
   * </pre>
   */
  private int propagateFromTargetSourceToSource(
      DependencyGraphEdge edge,
      DependencyGraphVertex source,
      DependencyGraphVertex target) {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphEdge inEdgeOfTarget : graph.incomingEdgesOf(target)) {
      if (!inEdgeOfTarget.getDependency().getName().isSyntactic()
          && inEdgeOfTarget.getDependency().getName() != edge.getDependency().getName()) {
        DependencyGraphVertex targetSource = graph.getEdgeSource(inEdgeOfTarget);
        edgeIndex = addIfNotExsist(targetSource, source, inEdgeOfTarget, edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  private int addIfNotExsist(DependencyGraphVertex source,
      DependencyGraphVertex target,
      DependencyGraphEdge edge,
      int edgeIndex) {
    if (source.getIndex() == target.getIndex()) {
      return edgeIndex;
    }
    if (DependencyGraphUtils.containsEdge(graph, source, target, edge)) {
      return edgeIndex;
    }
    if (edge.getDependency().getName().equals(DependencyName.ARG0, DependencyName.ARG1)
        && graph.containsEdge(
            source,
            target,
            d -> d.getName().equals(DependencyName.ARG0, DependencyName.ARG1))) {
      return edgeIndex;
    }
    DependencyGraphEdge newEdge = new DependencyGraphEdge(edgeIndex, edge);
    graph.addEdge(source, target, newEdge);
    edgeIndex++;
    return edgeIndex;
  }
}
