package com.pengyifan.nlp.ssdg;

import static com.google.common.base.Preconditions.checkNotNull;

import java.util.logging.Logger;

import com.pengyifan.brat.BratDocument;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.kernel.allpathgraph.singleprocess.DuplicatedLinkChecker;

public class SSDGBuilder {

  private static final Logger LOGGER = Logger.getAnonymousLogger();

  private final DependencyGraph graph;
  private final BratDocument corefDoc;
  private final BratDocument abbrDoc;

  private DependencyGraph ssdg;
  private String text;

  public SSDGBuilder(
      String text,
      DependencyGraph graph,
      BratDocument corefDoc,
      BratDocument abbrDoc) {
    this.text = text;
    this.graph = graph;
    this.corefDoc = corefDoc;
    this.abbrDoc = abbrDoc;
  }

  private int addSemanticLink() {
    LOGGER.finer("Start adding semantic links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new SemanticLinkAdder(ssdg).add();
  }

  private int addAbilityLink() {
    LOGGER.finer("Start adding ability links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new AbilityLinkAdder(ssdg).add();
  }

  private int addConjLink() {
    LOGGER.finer("Start adding conj links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new ConjLinkPropagation(ssdg).propagate();
  }

  private int addCoreference() {
    LOGGER.finer("Start adding coreference links");
    checkNotNull(ssdg, "Dependency graph is not set");
    checkNotNull(corefDoc, "Coreference is not set");
    return new CoreferenceAdder(ssdg, corefDoc).add();
  }

  private int addAbbrev() {
    LOGGER.finer("Start adding coreference links");
    checkNotNull(ssdg, "Dependency graph is not set");
    checkNotNull(abbrDoc, "Abbrev is not set");
    return new AbbrevAdder(ssdg, abbrDoc).add();
  }

  private int addIsALink() {
    LOGGER.finer("Start removing conj links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new IsALinkAdder(ssdg).add();
  }

  private int addMemberCollectionLink() {
    LOGGER.finer("Start adding member-collection links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new MemberCollectionLinkAdder(ssdg).add();
  }

  private int addFtermLink() {
    LOGGER.finer("Start adding nn links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new CTermFTermLinkAdder(ssdg).add();
  }

  private int addDescriptiveLink() {
    LOGGER.finer("Start adding nn links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new DescriptiveLinkAdder(ssdg).add();
  }

  private int addNullArgumentLink() {
    LOGGER.finer("Start adding ability links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new NullArgumentLinkAdder(ssdg).add();
  }

  private int addPartWholeLink() {
    LOGGER.finer("Start adding member-collection links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new PartWholeLinkAdder(ssdg).add();
  }

  private int addParsingError() {
    LOGGER.finer("Start adding parsing error links");
    checkNotNull(text, "Text is not set");
    checkNotNull(ssdg, "Dependency graph is not set");
    try {
      return new ParseError(text, ssdg).add();
    } catch (Exception e) {
      System.err.println(e.getMessage());
    }
    return 0;
  }

  private int addReceptorLink() {
    LOGGER.finer("Start adding receptor links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new ReceptorAdder(ssdg).add();
  }

  private int addVerbGroupLink() {
    LOGGER.finer("Start adding receptor links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new VerbGroupAdder(ssdg).add();
  }

  public int build() {

    ssdg = new DependencyGraph(graph);
    ssdg.checkHashCode();

    int numOfEdges = 0;
    numOfEdges += addParsingError();
    numOfEdges += addVerbGroupLink();

    numOfEdges += addCoreference();
    numOfEdges += addAbbrev();

    // numOfEdges += addSemanticLink();
    numOfEdges += addIsALink();
    for (int i = 0; i < 2; i++) {
      numOfEdges += addConjLink();
      numOfEdges += addNullArgumentLink();
      numOfEdges += addAbilityLink();
      numOfEdges += addPartWholeLink();
      numOfEdges += addFtermLink();
      numOfEdges += addDescriptiveLink();
      numOfEdges += addMemberCollectionLink();

      numOfEdges += addSemanticLink();
      numOfEdges += addReceptorLink();

//       System.out.println(ssdg);
      numOfEdges += propagateReference();
//       System.out.println(ssdg);
    }
    checkDuplicateLink();

    return numOfEdges;
  }

  private int propagateReference() {
    LOGGER.finer("Start propagating reference links");
    checkNotNull(ssdg, "Dependency graph is not set");
    return new ReferentialLinkPropagation(ssdg).propage();
  }

  private void checkDuplicateLink() {
    LOGGER.finer("Start removing negative links");
    checkNotNull(ssdg, "Dependency graph is not set");
    new DuplicatedLinkChecker(ssdg).check();
  }

  public DependencyGraph getSSDG() {
    return ssdg;
  }
}
