package com.pengyifan.nlp.ssdg;

import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphEdge;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.nlp.ling.verb.VerbUtils;

import java.util.List;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

class SemanticLinkAdder extends EdgeAdder {

  private static final Predicate<DependencyGraphVertex> verbWithPredicate =
      v -> v.getTag().isVerb() || VerbUtils.isVnorm(v.getLemma());

  private static final Predicate<DependencyGraphVertex> verbFromPredicate =
      v -> v.getLemma().equals("dissociate");

  private static final Predicate<DependencyGraphVertex> verbToPredicate =
      v -> (v.getTag().isActiveVerb()
          && (v.getLemma().equals("bind") || v.getLemma().equals("recruit")))
          || v.getWord().equals("bound")
          || v.getLemma().equals("binding")
          || v.getLemma().equals("cross-link");

  private static final Predicate<DependencyGraphVertex> verbOfPredicate =
      v -> (v.getTag().isNoun()
          && (VerbUtils.isVnorm(v.getLemma())
          || VerbUtils.isVion(v.getLemma())
          || v.getLemma().equals("activity")
          || v.getLemma().equals("cleavage")
          || v.getLemma().equals("recruitment")))
          || (v.getTag().isVerb() && v.getLemma().equals("compose"));

  private static final Predicate<DependencyGraphVertex> verbBetweenPredicate =
      v -> v.getTag().isNoun()
          && VerbUtils.isBetween(v.getLemma());

  private static final Predicate<DependencyGraphVertex> verbBindingSitePredicate =
      v -> "activated".equalsIgnoreCase(v.getWord())
          || "associated".equalsIgnoreCase(v.getWord())
          || "binding".equalsIgnoreCase(v.getWord())
          || "docking".equalsIgnoreCase(v.getWord())
          || "expressing".equalsIgnoreCase(v.getWord())
          || "interacting".equalsIgnoreCase(v.getWord())
          || "phosphorylated".equalsIgnoreCase(v.getWord());

  private static final Predicate<DependencyGraphVertex> bindingSitePredicate =
      v -> v.getLemma().equals("activity")
          || v.getLemma().equals("determinant")
          || v.getLemma().equals("domain")
          || v.getLemma().equals("epitope")
          || v.getLemma().equals("interface")
          || v.getLemma().equals("protein")
          || v.getLemma().equals("region")
          || v.getLemma().equals("site");

  SemanticLinkAdder(DependencyGraph graph) {
    super(graph);
  }

  int add() {
    int numOfEdges = 0;
    numOfEdges += addNSubjNSubjpass();
    numOfEdges += addPrepOfBy();
    numOfEdges += addPrepOfBy2();

    numOfEdges += addPrepBetween();
    numOfEdges += addPoss();
    numOfEdges += addVmod();
    numOfEdges += addJj();
    numOfEdges += addNnVnorm();
    numOfEdges += addXBindingSite();
    numOfEdges += addBindingSiteFor();

    numOfEdges += addPrepOf();
    numOfEdges += addPrepTo();
    numOfEdges += addPrepOn();
    numOfEdges += addPrepWith();
    numOfEdges += addPrepFrom();
    numOfEdges += addPrepBy();
    numOfEdges += addPrepVia();
    numOfEdges += addPrepToward();

    numOfEdges += addLike();
    return numOfEdges;
  }

  private static final DependencyGraphPattern nsubjPattern = DependencyGraphPattern
      .of()
          // source
      .addVertex(0, v -> v.getTag().isActiveVerb()
          || (v.getTag() == PartOfSpeech.VBN && v.getLemma().equals("cleave")))
          // target
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.NSUBJ, DependencyName.XSUBJ));
  private static final DependencyGraphPattern nsubjpassPattern = DependencyGraphPattern
      .of()
          // source
      .addVertex(0, v -> true)
          // target
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.NSUBJPASS));
  private static final DependencyGraphPattern dobjPattern = DependencyGraphPattern
      .of()
          // source
      .addVertex(0, v -> true)
          // target
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.DOBJ));
  private static final DependencyGraphPattern agentPattern = DependencyGraphPattern
      .of()
          // source
      .addVertex(0, v -> true)
          // target
      .addVertex(1, v -> true)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.AGENT));

  private int addNSubjNSubjpass() {
    int n = 0;
    n += add(nsubjPattern, 0, 1, DependencyName.ARG0, "nsubj");
    n += add(nsubjpassPattern, 0, 1, DependencyName.ARG1, "nsubjpass");
    n += add(dobjPattern, 0, 1, DependencyName.ARG1, "dobj");
    n += add(agentPattern, 0, 1, DependencyName.ARG0, "agent");
    return n;
  }

  private int addPrepX(String specific,
      Predicate<DependencyGraphVertex> predicate) {

    final DependencyGraphPattern prepXPattern = DependencyGraphPattern
        .of()
            // n1
        .addVertex(0, v -> true)
            // n2
        .addVertex(1, v -> true)
        .addEdge(0, 1, Dependency.ofPrepPredicate(specific));

    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = prepXPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex source = matcher.get(0);
      DependencyGraphVertex target = matcher.get(1);
      if (hasArg(source, target)) {
        continue;
      }
      if (predicate.test(source)) {
        DependencyName sname;
        if (canBeArg(source, target, DependencyName.ARG0)) {
          sname = DependencyName.ARG0;
        } else if (canBeArg(source, target, DependencyName.ARG1)) {
          sname = DependencyName.ARG1;
        } else {
          sname = DependencyName.ARG1;
        }
        edgeIndex = addIfNotExist(source, target, sname, specific, edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  private int addPrepBy() {
    return addPrepX("by", t -> t.getTag() == PartOfSpeech.VBN);
  }

  private int addPrepVia() {
    return addPrepX("via", t -> t.getTag() == PartOfSpeech.VBN);
  }

  private int addPrepWith() {
    return addPrepX("with", verbWithPredicate);
  }

  private int addPrepFrom() {
    return addPrepX("from", verbFromPredicate);
  }

  private int addPrepTo() {
    return addPrepX("to", verbToPredicate);
  }

  private int addPrepOn() {
    return addPrepX("on", verbToPredicate);
  }

  private int addPrepBetween() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphVertex v : graph.vertexSet()) {
      if (!verbBetweenPredicate.test(v)) {
        continue;
      }
      List<DependencyGraphVertex> betweens = graph
          .outgoingEdgeTargetsOf(v, Dependency.ofPredicate(DependencyName.PREP, "between"))
          .stream()
          .sorted((v1, v2) -> Integer.compare(v1.getBeginPosition(), v2.getBeginPosition()))
          .collect(Collectors.toList());
      if (betweens.isEmpty()) {
        continue;
      }
      edgeIndex = addIfNotExist(v, betweens.get(0), DependencyName.ARG0, "between", edgeIndex);
      for (int i = 1; i < betweens.size(); i++) {
        edgeIndex = addIfNotExist(v, betweens.get(i), DependencyName.ARG1, "between", edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * vion of n1 by n2
   */
  private static final DependencyGraphPattern prepOfByPattern = DependencyGraphPattern
      .of()
          // vion
      .addVertex(0, verbOfPredicate)
          // n1
      .addVertex(1, v -> true)
          // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"))
      .addEdge(0, 2, Dependency.ofPrepPredicate("by", "with", "to"));

  private int addPrepOfBy() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = prepOfByPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex vion = matcher.get(0);
      DependencyGraphVertex nOf = matcher.get(1);
      DependencyGraphVertex nBy = matcher.get(2);
      edgeIndex = addIfNotExist(vion, nOf, DependencyName.ARG1, "of", edgeIndex);
      edgeIndex = addIfNotExist(vion, nBy, DependencyName.ARG0, "by", edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  /**
   * vion of n1 by n2
   */
  private static final DependencyGraphPattern prepOfByPattern2 = DependencyGraphPattern
      .of()
          // vion
      .addVertex(0, verbOfPredicate)
          // n1
      .addVertex(1, v -> true)
          // n2
      .addVertex(2, v -> true)
      .addEdge(0, 1, Dependency.ofPrepPredicate("of"))
      .addEdge(1, 2, Dependency.ofPrepPredicate("by", "with", "to"));

  private int addPrepOfBy2() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = prepOfByPattern2.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex vion = matcher.get(0);
      DependencyGraphVertex nOf = matcher.get(1);
      DependencyGraphVertex nBy = matcher.get(2);
      edgeIndex = addIfNotExist(vion, nOf, DependencyName.ARG1, "of", edgeIndex);
      edgeIndex = addIfNotExist(vion, nBy, DependencyName.ARG0, "by", edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  private int addPrepToward() {
    return addPrepX("toward", t -> t.getLemma().equals("activity"));
  }

  private int addPrepOf() {
    return addPrepX("of", verbOfPredicate);
  }

  /**
   * its < poss < v-ion/activity/binding
   */
  private static final DependencyGraphPattern possPattern = DependencyGraphPattern
      .of()
          // its
      .addVertex(0, v -> true)
          // vion
      .addVertex(1, v -> v.getLemma().endsWith("ion")
          || v.getLemma().equals("activity")
          || v.getLemma().equals("binding"))
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.POSS));

  private int addPoss() {
    return add(possPattern, 1, 0, DependencyName.ARG0, "poss");
  }

  /**
   * <pre>
   *       arg1
   * +-------------+
   * v             |
   * N > vmod > VBN/VBD
   * </pre>
   */
  private static final DependencyGraphPattern vmodPattern1 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // vbn
      .addVertex(1, v -> v.getTag() == PartOfSpeech.VBN
          || (v.getTag() == PartOfSpeech.VBD && v.getLemma().equals("bind")))
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD));

  /**
   * <pre>
   *     arg0
   * +-----------+
   * v           |
   * N > vmod > VBG
   * </pre>
   */
  private static final DependencyGraphPattern vmodPattern2 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // vbg
      .addVertex(1, v -> v.getTag() == PartOfSpeech.VBG)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD));

  private int addVmod() {
    int n = 0;
    n += add(vmodPattern1, 1, 0, DependencyName.ARG1, "vmod");
    n += add(vmodPattern2, 1, 0, DependencyName.ARG0, "vmod");
    return n;
  }

  /**
   * <pre>
   *      arg1
   * +--------------+
   * |              v
   * JJ-ed < amod < NN
   */
  private static final DependencyGraphPattern jjPattern = DependencyGraphPattern
      .of()
          // jj
      .addVertex(0,
          v -> (v.getTag() == PartOfSpeech.JJ || v.getTag() == PartOfSpeech.VBN)
              && (v.getWord().endsWith("ed") || v.getWord().endsWith("ing")))
          // n
      .addVertex(1, v -> v.getTag().isNoun())
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.AMOD));

  private int addJj() {
    return add(jjPattern, 0, 1, DependencyName.ARG1, "jj");
  }

  /**
   * <pre>
   *          nn/amod
   * +---------------------+
   * |         +---------+ |
   * v         v         + +
   * N      binding       site
   * ^       + +           ^
   * +-------+ +-----------+
   *   arg0        arg1
   *
   * </pre>
   */
  private static final DependencyGraphPattern xBiningSitePattern1 = DependencyGraphPattern
      .of()
          // binding
      .addVertex(1, verbBindingSitePredicate)
          // site
      .addVertex(2, bindingSitePredicate)
      .addEdge(2, 1, e -> true);
  private static final DependencyGraphPattern xBiningSitePattern2 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> v.getNer().isProtein())
          // binding
      .addVertex(1, verbBindingSitePredicate)
          // site
      .addVertex(2, bindingSitePredicate)
      .addEdge(2, 1, e -> true)
      .addEdge(2, 0, Dependency.NnPredicate);

  private int addXBindingSite() {
    int n = 0;
    n += add(xBiningSitePattern1, 1, 2, DependencyName.ARG1, "binding-site");
    n += add(xBiningSitePattern2, 1, 0, DependencyName.ARG0, "binding-site");
    return n;
  }

  /**
   * <pre>
   * binding < xx < site/protein > prep_for > N
   * |                 ^                      ^
   * +-----------------+                      |
   * |      arg1                              |
   * +----------------------------------------+
   *                  arg0
   * </pre>
   */
  private static final DependencyGraphPattern bindingSiteForPattern = DependencyGraphPattern
      .of()
          // binding
      .addVertex(0, verbBindingSitePredicate)
          // site
      .addVertex(1, bindingSitePredicate)
          // n
      .addVertex(2, v -> true)
      .addEdge(1, 0, e -> true)
      .addEdge(1, 2, Dependency.ofPrepPredicate("for"));

  private int addBindingSiteFor() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = bindingSiteForPattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex binding = matcher.get(0);
      DependencyGraphVertex site = matcher.get(1);
      DependencyGraphVertex n = matcher.get(2);
      edgeIndex = addIfNotExist(binding, site, DependencyName.ARG1, "binding-site", edgeIndex);
      edgeIndex = addIfNotExist(binding, n, DependencyName.ARG0, "binding-site", edgeIndex);
    }
    return edgeIndex - startIndex;
  }

  /**
   * +--------------+
   * |              |
   * V              |
   * Like N, N < nsubj < V
   */
  private static final DependencyGraphPattern likePattern = DependencyGraphPattern
      .of()
          // n1
      .addVertex(0, v -> v.getNer().isProtein())
          // n2
      .addVertex(1, v -> true)
          // v
      .addVertex(2, v -> v.getTag().isVerb())
      .addEdge(2, 0, Dependency.ofPrepPredicate("like"))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.ARG0));
  private static final DependencyGraphPattern likePattern2 = DependencyGraphPattern
      .of()
          // n1
      .addVertex(0, v -> v.getNer().isProtein())
          // n2
      .addVertex(1, v -> true)
          // v
      .addVertex(2, v -> v.getTag().isVerb())
      .addEdge(2, 0, Dependency.ofPrepPredicate("like"))
      .addEdge(2, 1, Dependency.ofPredicate(DependencyName.ARG1));

  private int addLike() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    DependencyGraphMatcher matcher = likePattern.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex n1 = matcher.get(0);
      DependencyGraphVertex n2 = matcher.get(1);
      DependencyGraphVertex v = matcher.get(2);
      if (n1.getBeginPosition() < n2.getBeginPosition()) {
        edgeIndex = addIfNotExist(v, n1, DependencyName.ARG0, "like", edgeIndex);
      }
    }
    matcher = likePattern2.matcher(graph);
    while (matcher.find()) {
      DependencyGraphVertex n1 = matcher.get(0);
      DependencyGraphVertex n2 = matcher.get(1);
      DependencyGraphVertex v = matcher.get(2);
      if (n1.getBeginPosition() < n2.getBeginPosition()) {
        edgeIndex = addIfNotExist(v, n1, DependencyName.ARG1, "like", edgeIndex);
      }
    }
    return edgeIndex - startIndex;
  }

  /**
   * <pre>
   *        nn
   * +---------------+
   * v               |
   * N        phosphorylation
   * ^               |
   * +---------------+
   * arg0
   */
  private int addNnVnorm() {
    int startIndex = graph.getMaxIndex() + 1;
    int edgeIndex = startIndex;
    for (DependencyGraphVertex verb : graph.vertexSet()) {
      if (!verbOfPredicate.test(verb)) {
        continue;
      }
      List<DependencyGraphVertex> proteins = graph.outgoingEdgesOf(verb)
          .stream()
          .filter(e -> e.getDependency().getName().equals(
              DependencyName.NN,
              DependencyName.AMOD,
              DependencyName.NSUBJ,
              DependencyName.NUM))
          .map(graph::getEdgeTarget)
          .filter(v -> v.getNer().isProtein())
          .collect(Collectors.toList());
      edgeIndex = addProteins(proteins, verb, edgeIndex);
      boolean hasNN = !proteins.isEmpty();

      proteins = graph.incomingEdgesOf(verb)
          .stream()
          .filter(e -> e.getDependency().getName() == DependencyName.DEP)
          .map(graph::getEdgeSource)
          .filter(n -> n.getNer().isProtein())
          .collect(Collectors.toList());
      edgeIndex = addProteins(proteins, verb, edgeIndex);
      hasNN |= !proteins.isEmpty();

      // by
      if (hasNN) {
        for (DependencyGraphEdge edge : graph.outgoingEdgesOf(verb)) {
          if (edge.getDependency().equals(DependencyName.PREP, "by", "of", "upon", "with")) {
            DependencyGraphVertex target = graph.getEdgeTarget(edge);
            if (target.getNer().isProtein()
                || graph.hasOutgoingEdgesOf(target,
                Dependency.ofPredicate(DependencyName.PART_WHOLE))) {
              edgeIndex = addIfNotExist(verb, target, DependencyName.ARG1,
                  edge.getDependency().getSpecific().get(), edgeIndex);
            }
          }
        }
      }
    }
    return edgeIndex - startIndex;
  }

  private int addProteins(List<DependencyGraphVertex> proteins,
      DependencyGraphVertex source, int edgeIndex) {
    if (proteins.size() == 2) {
      DependencyGraphVertex p0 = proteins.get(0);
      DependencyGraphVertex p1 = proteins.get(1);
      if (graph.containsEdge(p0, p1, Dependency.ofPredicate(DependencyName.CONJ, "dash"))
          || graph.containsEdge(p1, p0, Dependency.ofPredicate(DependencyName.CONJ, "dash"))) {
        edgeIndex = addIfNotExist(source, p0, DependencyName.ARG0, "nn", edgeIndex);
        edgeIndex = addIfNotExist(source, p1, DependencyName.ARG1, "nn", edgeIndex);
        return edgeIndex;
      }
    }
    for (DependencyGraphVertex protein : proteins) {
      edgeIndex = addIfNotExist(source, protein, DependencyName.ARG0, "nn", edgeIndex);
    }
    return edgeIndex;
  }

  private boolean hasArg(DependencyGraphVertex source,
      DependencyGraphVertex target) {
    return graph.containsEdge(source, target,
        Dependency.ofPredicate(DependencyName.ARG0, DependencyName.ARG1));
  }

  private boolean canBeArg(DependencyGraphVertex source,
      DependencyGraphVertex target, DependencyName name) {
    Set<DependencyGraphVertex> set = graph.outgoingEdgeTargetsOf(
        source,
        Dependency.ofPredicate(name));
    set = set.stream().filter(v -> v.getEndPosition() != target.getEndPosition())
        .collect(Collectors.toSet());
    if (set.isEmpty()) {
      return true;
    }

    final Predicate<Dependency> conjPred = Dependency.ofPredicate(DependencyName.CONJ);
    for (DependencyGraphVertex v : set) {
      if (graph.containsEdge(v, target, conjPred)
          || graph.containsEdge(target, v, conjPred)) {
        return true;
      }
    }
    return false;
  }

}
