package com.pengyifan.nlp.ssdg;

import com.google.common.collect.Sets;
import com.pengyifan.commons.collections.dependencygraph.Dependency;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraph;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphMatcher;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphPattern;
import com.pengyifan.commons.collections.dependencygraph.DependencyGraphVertex;
import com.pengyifan.commons.collections.dependencygraph.DependencyName;
import com.pengyifan.commons.collections.dependencygraph.PartOfSpeech;
import com.pengyifan.nlp.ling.verb.VerbUtils;

import java.util.Set;
import java.util.function.Predicate;

class VerbGroupAdder extends EdgeAdder {

  public VerbGroupAdder(DependencyGraph graph) {
    super(graph);
  }

  public int add() {
    int numOfEdges = 0;
    numOfEdges += addProXcomp();
    numOfEdges += addProCcomp();
    return numOfEdges;
  }

  /**
   * <pre>
   *          nsubj
   * +---------------------------+
   * |                xcomp      |
   * |           +--------------+|
   * v           |              v|
   * N > amod > VBN  to < aux < VB
   * </pre>
   */
  private static final DependencyGraphPattern proXcompPattern1 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // vbn
      .addVertex(1, v -> v.getTag().equals(PartOfSpeech.VBN, PartOfSpeech.JJ))
          // to
      .addVertex(2, v -> true)
          // vb
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VB || v.getTag() == PartOfSpeech.VBP)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.AMOD))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.XCOMP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.AUX));
  /**
   * <pre>
   *          arg0
   * +---------------------------+
   * |                xcomp      |
   * |           +--------------+|
   * v           |              v|
   * N > vmod > known  to < aux < VB
   * </pre>
   */
  private static final DependencyGraphPattern proXcompPattern2 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // known
      .addVertex(1, v -> v.getWord().equals("known") || v.getWord().equals("shown"))
          // to
      .addVertex(2, v -> true)
          // vb
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VB || v.getTag() == PartOfSpeech.VBP)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.XCOMP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.AUX));
  /**
   * N > vmod > known to be VBN
   */
  private static final DependencyGraphPattern proXcompPattern3 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // known
      .addVertex(1, v -> v.getWord().equals("known") || v.getWord().equals("shown"))
          // to
      .addVertex(2, v -> true)
          // be
      .addVertex(3, v -> true)
          // vb
      .addVertex(4, v -> v.getTag() == PartOfSpeech.VBN)
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 4, Dependency.ofPredicate(DependencyName.XCOMP))
      .addEdge(4, 2, Dependency.ofPredicate(DependencyName.AUX))
      .addEdge(4, 3, Dependency.ofPredicate(DependencyName.AUXPASS));
  /**
   * N > vmod > known to be N
   */
  private static final DependencyGraphPattern proXcompPattern4 = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // known
      .addVertex(1, v -> v.getWord().equals("known") || v.getWord().equals("shown"))
          // to
      .addVertex(2, v -> true)
          // be
      .addVertex(3, v -> true)
          // vb
      .addVertex(4, v -> v.getTag().isNoun())
      .addEdge(0, 1, Dependency.ofPredicate(DependencyName.VMOD))
      .addEdge(1, 4, Dependency.ofPredicate(DependencyName.XCOMP))
      .addEdge(4, 2, Dependency.ofPredicate(DependencyName.AUX))
      .addEdge(4, 3, Dependency.ofPredicate(DependencyName.COP));
  private int addProXcomp() {
    int numOfEdge = 0;
    numOfEdge += add(proXcompPattern1, 3, 0, DependencyName.NSUBJ, "null_pro");
    numOfEdge += add(proXcompPattern2, 3, 0, DependencyName.NSUBJ, "null_pro");
    numOfEdge += add(proXcompPattern3, 4, 0, DependencyName.NSUBJPASS, "null_pro");
    numOfEdge += add(proXcompPattern4, 4, 0, DependencyName.NSUBJ, "null_pro");
    return numOfEdge;
  }

  /**
   * N < nsubjpass < shown > ccomp > to Vact
   */
  private static final DependencyGraphPattern proCcompPattern = DependencyGraphPattern
      .of()
          // n
      .addVertex(0, v -> true)
          // known
      .addVertex(1, v -> v.getWord().equals("known") || v.getWord().equals("shown"))
          // to
      .addVertex(2, v -> true)
          // vb
      .addVertex(3, v -> v.getTag() == PartOfSpeech.VB || v.getTag() == PartOfSpeech.VBP)
      .addEdge(1, 0, Dependency.ofPredicate(DependencyName.NSUBJPASS))
      .addEdge(1, 3, Dependency.ofPredicate(DependencyName.CCOMP))
      .addEdge(3, 2, Dependency.ofPredicate(DependencyName.AUX));
  private int addProCcomp() {
    int numOfEdge = 0;
    numOfEdge += add(proCcompPattern, 3, 0, DependencyName.NSUBJ, "null_pro");
    return numOfEdge;
  }
}
